import * as React from 'react';
import { css } from 'linaria';
import ConnectionStatus from './components/connection/ConnectionStatus';
// import LoggedInHome from './components/connection/LoggedInHome';
import ConnectionRequest from './components/connection/ConnectionRequest';
import AuthorizationRequest from './components/authorization/AuthorizationRequest';
import { ParentWalletLinkManager } from './serviceworker/ParentWalletLinkManager';
import {
  Pubsub,
  AppState,
  Disconnected,
  ParentRequestEthereumAccounts,
  Connected,
  RequestEthereumAccounts,
  ReloadParentRequestEthereumAccounts,
} from './util/Pubsub';
import { ampClient, getUserId } from './init/analytics';
import { setUserId } from "./init/bugsnag";

const parentWalletLinkManager = new ParentWalletLinkManager();

const baseStyle = css`
  background-color: var(--background);
`;

const getQuery = () => {
  if (typeof window !== 'undefined') {
    return new URLSearchParams(window.location.search);
  }
  return new URLSearchParams();
};

const getQueryBooleanVal = (key: string): boolean | null => {
  const strVal: string | null = getQuery().get(key);
  if (typeof strVal === 'undefined' || (strVal !== 'true' && strVal !== 'false')) {
    return null;
  }

  return strVal === 'true';
};

type AppProps = {
  initialAppState?: AppState;
};

const App = (props: AppProps) => {
  const [status, setStatus] = React.useState(props.initialAppState || new Disconnected());
  const isFullScreen = getQueryBooleanVal('fullScreen');

  React.useEffect(() => {
    let subscription = Pubsub.subscribe(s => {
      setStatus(s);

      if (s.constructor === ReloadParentRequestEthereumAccounts) {
        parentWalletLinkManager.resetAndRequestEthereumAccounts();
      }
    });

    const inPageRequest = getQueryBooleanVal('inPageRequest');

    if (!inPageRequest) {
      parentWalletLinkManager.requestEthereumAccounts(null);
    } else {
      const type = getQuery().get('type');
      const action = getQuery().get('action');
      const id = getQuery().get('id');

      if (!type || !action || !id) {
        return;
      }

      parentWalletLinkManager.handleEvent(type, action, id);
    }

    return () => subscription.unsubscribe()
  }, []);

  React.useEffect(() => {
    if (parentWalletLinkManager.selectedAddress) {
      getUserId(parentWalletLinkManager.selectedAddress).then(userId => {
        ampClient.setUserId(userId);
        setUserId(userId);
      });
    }
  }, [parentWalletLinkManager.selectedAddress]);

  const disconnect = () => {
    parentWalletLinkManager.disconnect();
  };

  const getCurrentComponent = (status: AppState) => {
    switch (status.constructor) {
      case ParentRequestEthereumAccounts:
        return (
          <ConnectionRequest
            options={status as ParentRequestEthereumAccounts}
            fullScreen={isFullScreen}
          />
        );
      case Connected:
        return (
          <ConnectionStatus
            userName={parentWalletLinkManager.walletUsername || ''}
            disconnect={disconnect}
          />
        );

        /*
        return (
          <LoggedInHome
            username={parentWalletLinkManager.walletUsername || ''}
          />
        );

         */

      case RequestEthereumAccounts:
        return (
          <AuthorizationRequest
            options={status as RequestEthereumAccounts}
            parentLink={parentWalletLinkManager}
            fullScreen={isFullScreen}
          />
        );
      default:
        return <></>;
    }
  };

  return (
    <div id="app-main" className={baseStyle}>
      {getCurrentComponent(status)}
    </div>
  );
};

export default App;
