import React, { FC, useState, useCallback } from 'react';
import { css, cx } from 'linaria';
import { useIntl } from 'react-intl';
import { Button, ButtonGroup } from '@cbhq/cds-web/buttons';
import { Icon } from '@cbhq/cds-web/icons/Icon';
import ListItemWithIcon from '../shared/ListItemIcon';
import LogoAuth from '../../assets/circleLogo.svg';
import { TextBodyGrey, TextTitle } from '../shared/typography/Text';
import { RequestEthereumAccounts, Pubsub } from '../../util/Pubsub';
import { ParentWalletLinkManager } from 'src/serviceworker/ParentWalletLinkManager';
import { sendEvent } from '../../init/analytics';
import { EVENTS } from '../../init/events';
import log from '../../util/log';
import AuthorizationError from './AuthorizationError';

const container = css`
  overflow: hidden;
`;

const header = css`
  display: flex;
  justify-content: center;
  align-items: center;
  margin: 23px;
  img {
    width: 40px;
  }
`;

const body = css`
  display: flex;
  flex-flow: column nowrap;
  background-color: var(--background);
  box-shadow: 0px 0px 8px rgba(0, 0, 0, 0.04), 0px 16px 24px rgba(0, 0, 0, 0.06);
  border-radius: 16px;
`;

const bodyExit = css`
  transform: translateY(120%);
  transition: transform 0.3s ease-out;
`;

const authRequest = css`
  margin: 24px;
`;

const authButtons = css`
  margin: 24px;
  display: flex;
  flex-flow: column nowrap;
`;

const leftLogo = css`
  z-index: 1;
  position: relative;
  left: 5px;
`;

const rightLogo = css`
  background-color: rgb(var(--gray0));
  box-shadow: 0px 3px 6px rgb(0 0 0 / 7%);
  width: 40px;
  height: 40px;
  border-radius: 100px;
  position: relative;
  left: -5px;
  & > img {
    width: 36px;
    display: block;
    margin: 2px 2px;
    border-radius: 100px;
  }
`;

const authRequestHeader = css`
  margin-right: 87px;
`;

const authRequestBody = css`
  margin-top: 24px;
  justify-content: space-between;
`;

const leftLogoAnimating = css`
  @keyframes logoAnimation {
    0% {
    }
    40% {
      transform: scale(2.5) translate(-20px, 83px);
    }
    60% {
      transform: scale(2.5) translate(-20px, 83px);
      opacity: 1;
    }
    90% {
      transform: scale(2.5) translate(8px, 83px);
    }
    100% {
      transform: scale(2.5) translate(8px, 83px);
      opacity: 0;
    }
  }
  animation: logoAnimation 1s;
  animation-timing-function: cubic-bezier(0.17, 0.97, 0.56, 1.1);
  animation-fill-mode: forwards;
`;

const rightLogoAnimating = css`
  @keyframes logoAnimation {
    0% {
    }
    40% {
      transform: scale(2.5) translate(20px, 83px);
    }
    60% {
      transform: scale(2.5) translate(20px, 83px);
      opacity: 1;
    }
    90% {
      transform: scale(2.5) translate(-4px, 83px);
    }
    100% {
      transform: scale(2.5) translate(-4px, 83px);
      opacity: 0;
    }
  }
  animation: logoAnimation 1s;
  animation-timing-function: cubic-bezier(0.17, 0.97, 0.56, 1.1);
  animation-fill-mode: forwards;
`;

const checkmarkCircle = css`
  background-color: rgb(var(--blue50));
  width: 102px;
  height: 102px;
  position: absolute;
  top: calc(50% - 60px);
  left: (50% - 51px);
  border-radius: 100px;
  z-index: 2;
  opacity: 0;
  display: flex;
`;

const checkmarkCircleAnimating = css`
  @keyframes logoAnimation {
    0% {
      opacity: 0;
    }
    65% {
      opacity: 0;
      transform: scale(1.2);
    }
    100% {
      opacity: 1;
    }
  }
  animation: logoAnimation 1s;
  animation-timing-function: cubic-bezier(0.25, 1.67, 0.89, 1.05);
  animation-fill-mode: forwards;
`;

const connectedText = css`
  opacity: 0;
  position: absolute;
  top: calc(50% - 60px + 120px);
  left: (50% - 51px);
  font-size: 16px;
  letter-spacing: 0.4px;
  font-weight: 500;
`;

const connectedTextAnimating = css`
  @keyframes textAnimation {
    0% {
      opacity: 0;
    }
    55% {
      opacity: 0;
    }
    100% {
      opacity: 1;
    }
  }
  animation: textAnimation 1s cubic-bezier(0.17, 0.97, 0.56, 1.1) 0.7s forwards;
`;

const checkmark = css`
  width: 80px;
  height: 80px;
  border-radius: 50%;
  display: block;
  stroke-width: 3;
  stroke: white;
  stroke-miterlimit: 10;
  box-shadow: inset 0px 0px 0px $green;
  margin: auto;
  opacity: 0;
`;

const checkAnimating = css`
  @keyframes stroke {
    0% {
      opacity: 0;
    }
    50% {
      transform: scale(1.3);
    }
    100% {
      opacity: 1;
    }
  }
  animation: stroke 1s cubic-bezier(0.17, 0.97, 0.56, 1.1) 0.7s forwards;
`;

const check = css`
  transform-origin: 50% 50%;
`;

type AuthorizationRequestProps = {
  options: RequestEthereumAccounts;
  parentLink: ParentWalletLinkManager;
  fullScreen: boolean | null;
};

const AuthorizationRequest: FC<AuthorizationRequestProps> = props => {
  const [authed, setAuthed] = useState(false);
  const [authorizing, setAuthorizing] = useState(false);
  const [error, setError] = useState(null);
  const resetError = useCallback(() => setError(null), [setError]);
  const intl = useIntl();

  React.useEffect(() => {
    let subscription = Pubsub.subscribe(s => {
      if (s.constructor === RequestEthereumAccounts) {
        setAuthed(false);
      }
    });

    return () => subscription.unsubscribe();
  }, []);

  const denyAuthorize = () => {
    props.parentLink.cancel(props.options).then(() => {
      sendEvent(EVENTS.AUTHORIZE_DENY, { dapp: window.location.host });
      setAuthed(false);
      window.close();
    });
  };

  const allowAuthorize = () => {
    setAuthorizing(true);
    props.parentLink
      .connectChild(props.options)
      .then(() => {
        setAuthed(true);
        setAuthorizing(false);
        if (props.parentLink.hasNextChildRequest()) {
          setTimeout(() => {
            props.parentLink.handleNextChildRequest();
          }, 2000);
        } else {
          setTimeout(() => {
            window.close();
          }, 2000);
        }
      })
      .catch(error => {
        log('Had error, need design ' + error);
        setAuthorizing(false);
        setError(error || true); // if no error object exists, we still want to enter error state
      });
  };

  //@ts-ignore
  const setDappLogoURLAlt = ev => {
    props.options.dappLogoURL = props.options.dappLogoURLAlt;
    ev.target.src = props.options.dappLogoURLAlt;
  };

  if (error) {
    return <AuthorizationError resetError={resetError} />;
  }

  return (
    <div className={container}>
      <div className={header}>
        <div className={cx(leftLogo, authed && leftLogoAnimating)}>
          <img src={LogoAuth} />
        </div>
        <div className={cx(rightLogo, authed && rightLogoAnimating)}>
          <img src={props.options.dappLogoURL} onError={setDappLogoURLAlt} />
        </div>
        <div className={cx(checkmarkCircle, authed && checkmarkCircleAnimating)}>
          <svg className={cx(checkmark, authed && checkAnimating)} viewBox="0 0 52 52">
            <path className={cx(check)} fill="none" d="M14.1 27.2l7.1 7.2 16.7-16.8" />
          </svg>
        </div>
        <div className={cx(connectedText, authed && connectedTextAnimating)}>
          {intl.formatMessage({
            id: 'connectedStatusMessage',
            description: 'Connected status message',
            defaultMessage: 'Connected',
          })}
        </div>
      </div>
      <div className={cx(body, authed && bodyExit)}>
        <div className={authRequest}>
          <div className={authRequestHeader}>
            <TextTitle
              text={intl.formatMessage({
                id: 'allowWebsiteStatusMessage',
                description: 'Status message header',
                defaultMessage: 'Allow website',
              })}
            />
            <TextBodyGrey text={new URL(props.options.dappURL).hostname} />
          </div>
          <div className={authRequestBody} data-testid="authorization-request-permission-list">
            <ListItemWithIcon
              text={intl.formatMessage({
                id: 'seeWalletActivityPermission',
                description: 'Permission prompt',
                defaultMessage: 'To see your wallet balance and activity',
              })}
            >
              <Icon name="bank" size="m" color="primary" />
            </ListItemWithIcon>
            <ListItemWithIcon
              text={intl.formatMessage({
                id: 'sendRequestsForTransactionsPermission',
                description: 'Permission prompt',
                defaultMessage: 'To send you requests for transactions',
              })}
            >
              <Icon name="circleCheckmark" size="m" color="primary" />
            </ListItemWithIcon>
            <ListItemWithIcon
              text={intl.formatMessage({
                id: 'cannotMoveFundsWithoutYourPermission',
                description: 'Permission prompt',
                defaultMessage: 'Cannot move funds without your permission',
              })}
            >
              <Icon name="circleCross" size="m" color="negative" />
            </ListItemWithIcon>
          </div>
        </div>
        <div className={authButtons}>
          <ButtonGroup
            block
            vertical
            accessibilityLabel={intl.formatMessage({
              id: 'authorizationButtonGroup',
              description: 'Button group accessibility label',
              defaultMessage: 'Action buttons',
            })}
          >
            <Button
              type="button"
              variant="primary"
              onPress={allowAuthorize}
              loading={authorizing}
              data-testid="allow-authorize-button"
            >
              {intl.formatMessage({
                id: 'allowActionButton',
                description: 'Action button text',
                defaultMessage: 'Allow',
              })}
            </Button>
            <Button
              type="button"
              variant="secondary"
              onPress={denyAuthorize}
              data-testid="deny-authorize-button"
            >
              {intl.formatMessage({
                id: 'denyActionButton',
                description: 'Action button text',
                defaultMessage: 'Deny',
              })}
            </Button>
          </ButtonGroup>
        </div>
      </div>
    </div>
  );
};

export default AuthorizationRequest;
