import React from 'react';
import { act, render, screen, waitFor } from '@testing-library/react';
import userEvent from '@testing-library/user-event';
import AuthorizationRequest from '../AuthorizationRequest';
import { RequestEthereumAccounts } from '../../../util/Pubsub';
import { ParentWalletLinkManager } from '../../../serviceworker/ParentWalletLinkManager';

const mockCancel = jest.fn().mockResolvedValue(undefined);
const mockConnectChild = jest.fn().mockResolvedValue(undefined);
const mockHasNextChildRequest = jest.fn().mockReturnValue(true);
const mockHandleNextChildRequest = jest.fn();

jest.mock('../../../serviceworker/ParentWalletLinkManager', () => {
  return {
    ParentWalletLinkManager: () => {
      return {
        cancel: mockCancel,
        connectChild: mockConnectChild,
        hasNextChildRequest: mockHasNextChildRequest,
        handleNextChildRequest: mockHandleNextChildRequest
      };
    }
  };
});

describe('Authorization Request', () => {
  const options = new RequestEthereumAccounts(
    '1234',
    'dappName',
    'dappLogoURL',
    'dappLogoURLAlt',
    'https://dappURL',
    'sessionid',
    'sessionsecret'
  );

  const mockParentWalletLinkManager = new ParentWalletLinkManager();

  beforeEach(() => {
    mockCancel.mockClear();
    mockConnectChild.mockClear();
    mockHasNextChildRequest.mockClear();
    mockHandleNextChildRequest.mockClear();
  });

  it('Should render', () => {
    render (
      <AuthorizationRequest
        options={options}
        parentLink={mockParentWalletLinkManager}
        fullScreen={false}
      />
    );

    expect(screen.getByTestId('authorization-request-permission-list')).toBeInTheDocument();
    expect(screen.getByTestId('allow-authorize-button')).toBeInTheDocument();
    expect(screen.getByTestId('deny-authorize-button')).toBeInTheDocument();
  });

  it('It should cancel if Deny is pressed', async () => {
    // The code calls window.close which will throw a DOM error in the test, so we Mock it to do nothing
    window.close = jest.fn(() => {});

    render (
      <AuthorizationRequest
        options={options}
        parentLink={mockParentWalletLinkManager}
        fullScreen={false}
      />
    );

    act(() => {
      userEvent.click(screen.getByTestId('deny-authorize-button'));
    });
    expect(mockParentWalletLinkManager.cancel).toBeCalled();
  });

  it('It should approve if Authorize is pressed', async () => {
    // Mock setTimeout timer
    jest.useFakeTimers();

    render (
      <AuthorizationRequest
        options={options}
        parentLink={mockParentWalletLinkManager}
        fullScreen={false}
      />
    );

    await act(async () => {
      userEvent.click(screen.getByTestId('allow-authorize-button'));
    });
    expect(mockParentWalletLinkManager.connectChild).toBeCalled();

    act(() => {
      jest.advanceTimersByTime(2001);
    });
    expect(mockParentWalletLinkManager.handleNextChildRequest).toBeCalled();

    // Reset timers
    jest.runOnlyPendingTimers();
    jest.useRealTimers();
  });

  it('It should close window if no next child request', async () => {
    // Mock window.close
    window.close = jest.fn(() => {});

    // Mock setTimeout timer
    jest.useFakeTimers();

    mockParentWalletLinkManager.hasNextChildRequest = jest.fn().mockReturnValue(false);

    render (
      <AuthorizationRequest
        options={options}
        parentLink={mockParentWalletLinkManager}
        fullScreen={false}
      />
    );

    await act(async () => {
      userEvent.click(screen.getByTestId('allow-authorize-button'));
    });
    expect(mockParentWalletLinkManager.connectChild).toBeCalled();

    act(() => {
      jest.advanceTimersByTime(2001);
    });
    expect(window.close).toBeCalled();

    // Reset timers
    jest.runOnlyPendingTimers();
    jest.useRealTimers();
  });

  it('It should render error if connectChild() fails', async () => {
    // Mock window.close
    window.close = jest.fn(() => {});

    // Mock setTimeout timer
    jest.useFakeTimers();

    mockParentWalletLinkManager.connectChild = jest.fn().mockRejectedValue(true);

    render (
      <AuthorizationRequest
        options={options}
        parentLink={mockParentWalletLinkManager}
        fullScreen={false}
      />
    );

    await act(async () => {
      userEvent.click(screen.getByTestId('allow-authorize-button'));
    });
    expect(mockParentWalletLinkManager.connectChild).toBeCalled();
    expect(screen.getByTestId('try-again')).toBeInTheDocument();

    // Reset timers
    jest.runOnlyPendingTimers();
    jest.useRealTimers();
  });
});
