// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

import React, { useState, FC, useEffect } from 'react';
import { useIntl } from 'react-intl';
import { Icon } from '@cbhq/cds-web/icons/Icon';
import QRCode from './QRCode';
import { css } from 'linaria';
import FooterLogo from '../../assets/circleLogo.svg';
import Vector from '../../assets/Vector.svg';
import { ParentRequestEthereumAccounts } from 'src/util/Pubsub';
import { TextBody1, TextBodyGrey } from '../shared/typography/Text';
import { sendEvent } from '../../init/analytics';
import { EVENTS } from '../../init/events';

const connectionRequestWrapper = css`
  position: relative;
`;

const qrContainer = css`
  padding: 24px;
  &:hover {
    padding: 16px;
  }
  transition: padding 200ms;
  height: 300px;
`;

const QRCodeFrame = css`
  align-items: center;
  justify-content: center;

  /* White */
  background: #ffffff;
  /* Grey 30% */

  border: 1px solid #eaedf0;
  box-sizing: border-box;
  /* Shadow level 2 */

  box-shadow: 0px 8px 24px rgba(7, 20, 34, 0.12);
  border-radius: 16px;
  display: flex;
`;

const bodyContainer = css`
  display: flex;
  flex-flow: column nowrap;
  padding: 8px 24px;
`;

const iconContainer = css`
  margin-left: 10px;
`;

const toolTip = css`
  position: absolute;
  z-index: 1;
  visibility: hidden;
  text-align: center;
  opacity: 0;
  transition: opacity 0.3s;
  background-color: rgb(var(--gray100));
  border-radius: 8px;
  color: rgb(var(--gray0));
  font-size: 13px;
  top: 18px;
  left: -26px;
  width: 90px;
`;

const connectionTitle = css`
  /* Scan QR code */
  font-size: 20px;
  font-weight: 500;
  line-height: 28px;
  letter-spacing: 0.03em;
  text-align: left;
  color: rgb(var(--gray100));
  margin-bottom: 8px;
`;

const toolTipContainer = css`
  display: inline-block;
  position: relative;
  cursor: pointer;

  :hover div {
    visibility: visible;
    opacity: 1;
  }
`;

const lineStyle = css`
  /* Grey 10 - Lines */
  border: 1px solid rgb(var(--gray5));
  margin-top: 60px;
`;

const vectorDiv = css`
  position: relative;
  left: 60px;
`;

const textWidthLimiter = css`
  width: 150px;
`;

const footerContainer = css`
  display: flex;
  flex-flow: row nowrap;
  align-items: center;
  padding-left: 15px;
  padding-bottom: 17px;
  padding-top: 14px;
  cursor: pointer;
  &:hover .${textWidthLimiter} > div {
    color: rgb(var(--blue70));
  }

  &:hover .${vectorDiv} > img {
    filter: brightness(0) saturate(100%) invert(25%) sepia(96%) saturate(2012%) hue-rotate(217deg)
      brightness(91%) contrast(109%);
  }
`;

const logo = css`
  padding-right: 13px;
  img {
    width: 34px;
  }
`;

const errorMessage = css`
  color: rgb(var(--red50));
  position: absolute;
  bottom: 90px;
`;

type ConnectionRequestProps = {
  options: ParentRequestEthereumAccounts;
  fullScreen: boolean | null;
};
const ConnectionRequest: FC<ConnectionRequestProps> = (props: {
  options: ParentRequestEthereumAccounts;
}) => {
  const [showError, setShowError] = useState(false);
  const intl = useIntl();
  const qrUrl = `${props.options.walletLinkUrl}/#/link?parent-id=${props.options.sessionId}&secret=${props.options.sessionSecret}&server=${props.options.walletLinkUrl}&v=1`;
  const toWalletLink = () => {
    sendEvent(EVENTS.REDIRECT);
    chrome.tabs.create({ url: 'home.html' });
  };
  useEffect(() => {
    setTimeout(() => {
      setShowError(true);
    }, 60000);
  }, []);

  return (
    <div className={connectionRequestWrapper}>
      <div className={qrContainer} data-testid="qr-container">
        <div className={QRCodeFrame}>
          <QRCode content={qrUrl} width={248} height={248} fgColor="#000" bgColor="transparent" />
        </div>
      </div>
      <input type="hidden" value={qrUrl} />
      <div className={bodyContainer}>
        <div className={connectionTitle}>
          {intl.formatMessage({
            id: 'scanQrCode',
            description: 'Prompt user to scan a QR code',
            defaultMessage: 'Scan QR code',
          })}
          <div className={toolTipContainer}>
            <div className={iconContainer} onClick={toWalletLink} data-testid="tooltip-link">
              <Icon name="questionMark" size="s" color="foreground" />
            </div>
            <div className={toolTip}>
              {intl.formatMessage({
                id: 'learnMoreTooltip',
                description: 'Tooltip on hover of icon',
                defaultMessage: 'Learn more',
              })}
            </div>
          </div>
        </div>
        <TextBody1
          text={intl.formatMessage({
            id: 'scanQrCodeWithMobileApp',
            description: 'Scan QR code prompt',
            defaultMessage: 'Scan the QR code with your Wallet mobile app to connect',
          })}
        />
        {showError && (
          <div className={errorMessage} data-testid="connection-timeout-prompt">
            {intl.formatMessage({
              id: 'cantConnectWallet',
              description: "Can't connect prompt",
              defaultMessage:
                "Can't connect? Make sure you're using the latest version of the Wallet mobile app.",
            })}
          </div>
        )}
      </div>
      <div>
        <hr className={lineStyle} />
      </div>
      <div className={footerContainer} onClick={toWalletLink} data-testid="footer-link">
        <div className={logo}>
          <img src={FooterLogo} />
        </div>
        <div>
          <div className={textWidthLimiter}>
            <TextBodyGrey
              text={intl.formatMessage({
                id: 'dontHaveWalletMobile',
                description: 'Link to Wallet mobile app download page',
                defaultMessage: "Don't have the Wallet mobile app?",
              })}
            />
          </div>
        </div>
        <div className={vectorDiv}>
          <img src={Vector} />
        </div>
      </div>
    </div>
  );
};

export default ConnectionRequest;
