import React from 'react';
import { css, cx } from 'linaria';
import { useIntl } from 'react-intl';
import { Button, ButtonGroup } from '@cbhq/cds-web/buttons';
import connectionLogo from '../../assets/circleLogo.svg';
import { sendEvent } from '../../init/analytics';
import { EVENTS } from '../../init/events';

type ConnectionStatusProps = {
  userName: string;
  disconnect: () => void;
};

const connectionStatusContainer = css`
  background-color: var(--background-alternate);
`;

const logoSection = css`
  height: 320px;
  box-shadow: 0px 4px 35px rgba(0, 0, 0, 0.02);
  display: flex;
  justify-content: center;
  align-items: center;
  position: relative;
  overflow: hidden;
`;
const ripple = css`
  /* Ellipse 58 */
  box-shadow: 0px 4px 35px rgba(0, 0, 0, 0.07);
  border-radius: 50%;
  position: absolute;
  display: flex;
  align-items: center;
  justify-content: center;
  width: calc(100% + 60px);
  height: calc(100% + 32px);
  opacity: 0;
  z-index: 1;
  @keyframes ripple {
    0% {
      transform: scale(0);
      opacity: 0;
      background-color: rgb(var(--gray0));
    }
    20% {
      transform: scale(1.5);
      opacity: 1;
      background-color: var(--background-alternate);
    }
    100% {
      transform: scale(1.5);
      opacity: 1;
      background-color: var(--background-alternate);
    }
  }
  animation: ripple 5s linear;
  animation-iteration-count: infinite;
`;
const outerCircle = css`
  animation-delay: 100ms;
`;
const innerCircle = css`
  animation-delay: 400ms;
`;
const logo = css`
  filter: drop-shadow(0px 0px 8px rgba(0, 0, 0, 0.04)),
    drop-shadow(0px 16px 24px rgba(0, 0, 0, 0.06));
  background: #ffffff;
  /* Z4 */
  box-shadow: 0px 0px 8px rgba(0, 0, 0, 0.04), 0px 16px 24px rgba(0, 0, 0, 0.06);
  border-radius: 100px;
  width: 133px;
`;
const logoDiv = css`
  background: #ffffff;
  box-shadow: 0px 0px 8px rgba(0, 0, 0, 0.04), 0px 16px 24px rgba(0, 0, 0, 0.06);
  border-radius: 100px;
  animation: none;
  opacity: 1;
  margin: 87px;
  z-index: 2;
  @keyframes ripple {
    0% {
      transform: scale(0.9);
    }
    5% {
      transform: scale(0.9);
    }
    8% {
      transform: scale(1);
    }
    90% {
      transform: scale(1);
    }
    100% {
      transform: scale(0.9);
    }
  }
  animation: ripple 5s linear;
  animation-iteration-count: infinite;
`;
const connectionSection = css`
  position: relative;
  padding: 24px;
  border-top-left-radius: 16px;
  border-top-right-radius: 16px;
  background-color: rgb(var(--gray0));
`;
const userNameText = css`
  order: 1;
  font-style: normal;
  font-weight: bold;
  font-size: 13px;
  line-height: 16px;
  color: rgb(var(--gray30));
`;
const statusText = css`
  order: 2;
  margin-bottom: 8px;
  margin-top: 1px;
  margin-right: 77px;
  font-style: normal;
  font-weight: 600;
  font-size: 20px;
  line-height: 29px;
  color: rgb(var(--gray100));
`;
const description = css`
  order: 3;
`;
const connectionText = css`
  margin-bottom: 24px;
  font-style: normal;
  font-weight: normal;
  font-size: 13px;
  line-height: 143.2%;
  color: rgb(var(--gray100));
`;

const ConnectionStatus: React.FC<ConnectionStatusProps> = React.memo(props => {
  const intl = useIntl();

  const disconnect = () => {
    sendEvent(EVENTS.DISCONNECTED);
    props.disconnect();
  };

  return (
    <div className={connectionStatusContainer}>
      <div className={logoSection}>
        <div className={cx(ripple, outerCircle)} />
        <div className={cx(ripple, innerCircle)} />
        <div className={logoDiv}>
          <img className={logo} src={connectionLogo} />
        </div>
      </div>
      <div className={connectionSection}>
        <div className={connectionText}>
          <div className={userNameText}>@{props.userName}</div>
          <div className={statusText} data-testid="wallet-connected-status">
            {intl.formatMessage({
              id: 'walletConnectedLabel',
              description: 'Wallet connected label',
              defaultMessage: 'Wallet connected'
            })}
          </div>
          <div className={description}>
            {intl.formatMessage({
              id: 'youCanNowConnectCryptoApps',
              description: 'Status text description',
              defaultMessage: 'You can now connect to crypto apps'
            })}
          </div>
        </div>

        <ButtonGroup
          block
          accessibilityLabel={intl.formatMessage({
            id: 'disconnectButtonGroup',
            description: 'Button group accessibility label',
            defaultMessage: 'Action buttons'
          })}
        >
          <Button
            type="button"
            variant="negative"
            onPress={disconnect}
            data-testid="disconnect-button"
          >
            {intl.formatMessage({
              id: 'disconnectButtonAction',
              description: 'Action button label',
              defaultMessage: 'Disconnect'
            })}
          </Button>
        </ButtonGroup>
      </div>
    </div>
  );
});
export default ConnectionStatus;
