import SafeEventEmitter from '@metamask/safe-event-emitter';
import { Web3Provider, RequestArguments } from 'walletlink/build/npm/dist/provider/Web3Provider';
import {
  JSONRPCRequest,
  JSONRPCResponse,
  JSONRPCMethod,
} from 'walletlink/build/npm/dist/provider/JSONRPC';
import { Callback, AddressString } from 'walletlink/build/npm/dist/types';

export class MultiProvider extends SafeEventEmitter {
  private providers = new Array<Web3Provider & SafeEventEmitter>();
  private selectedProvider: (Web3Provider & SafeEventEmitter) | null = null;
  private overrideIsMetaMask: boolean = false;

  public constructor() {
    super();
    this.enable = this.enable.bind(this);
    this.send = this.send.bind(this);
    this.sendAsync = this.sendAsync.bind(this);
    this.request = this.request.bind(this);
  }

  public async enable(): Promise<AddressString[]> {
    if (this.selectedProvider != null) {
      //@ts-ignore
      return this.selectedProvider.enable();
    }

    return this.request({ method: JSONRPCMethod.eth_requestAccounts });
  }

  public send(request: JSONRPCRequest<any[]>): JSONRPCResponse<any, any>;
  public send(request: JSONRPCRequest<any[]>[]): JSONRPCResponse<any, any>[];
  public send(request: JSONRPCRequest<any[]>, callback: Callback<JSONRPCResponse<any, any>>): void;
  public send(
    request: JSONRPCRequest<any[]>[],
    callback: Callback<JSONRPCResponse<any, any>[]>
  ): void;
  public send<T = any>(method: string, params?: any): Promise<T>;
  public send(
    requestOrMethod: JSONRPCRequest | JSONRPCRequest[] | string,
    callbackOrParams?: Callback<JSONRPCResponse> | Callback<JSONRPCResponse[]> | any[] | any
  ): JSONRPCResponse | JSONRPCResponse[] | void | Promise<any> {
    if (this.selectedProvider) {
      //@ts-ignore
      return this.selectedProvider.send(requestOrMethod, callbackOrParams);
    }

    throw new Error('Not implemented send ' + requestOrMethod);
  }

  public sendAsync(
    request: JSONRPCRequest<any[]>,
    callback: Callback<JSONRPCResponse<any, any>>
  ): void;
  public sendAsync(
    request: JSONRPCRequest<any[]>[],
    callback: Callback<JSONRPCResponse<any, any>[]>
  ): void;
  public sendAsync(request: any, callback: any) {
    if (this.selectedProvider != null) {
      return this.selectedProvider.sendAsync(request, callback);
    }

    if (typeof callback !== 'function') {
      throw new Error('callback is required');
    }

    if (
      request.method === JSONRPCMethod.net_version ||
      request.method === JSONRPCMethod.eth_accounts ||
      request.method === JSONRPCMethod.eth_requestAccounts
    ) {
      const cb = callback as Callback<JSONRPCResponse>;
      this.request({ method: request.method })
        .then(result => {
          cb(null, {
            jsonrpc: '2.0',
            id: request.id,
            result: result,
          });
        })
        .catch(err => cb(err, null));
      return;
    }
    throw new Error('Not implemented sendAsync ' + request);
  }

  public get isMetaMask(): boolean {
    return this.overrideIsMetaMask;
  }

  public get chainId(): string {
    if (this.selectedProvider) {
      return this.selectedProvider.chainId;
    }

    return '0x1';
  }

  public get networkVersion(): string {
    if (this.selectedProvider) {
      //@ts-ignore
      return this.selectedProvider.networkVersion;
    }

    return '1';
  }

  public request<T>(args: RequestArguments): Promise<T> {
    if (this.selectedProvider && this.shouldSelectProvider(this.selectedProvider)) {
      return this.selectedProvider.request(args);
    } else {
      this.selectedProvider = null;
    }

    if (args.method === JSONRPCMethod.net_version) {
      for (var i = 0; i < this.providers.length; i++) {
        let provider = this.providers[i];
        if (provider.chainId != '0x1') {
          this.setSelectedProvider(provider);
          return provider.request(args);
        }
      }

      //@ts-ignore
      return Promise.resolve('1');
    }

    if (args.method === JSONRPCMethod.eth_accounts) {
      for (var i = 0; i < this.providers.length; i++) {
        let provider = this.providers[i];
        if (this.hasSelectedAddress(provider)) {
          this.setSelectedProvider(provider);
          //@ts-ignore
          return Promise.resolve([provider.selectedAddress]);
        }
      }

      //@ts-ignore
      return Promise.resolve([]);
    }

    if (args.method === JSONRPCMethod.eth_requestAccounts) {
      return new Promise((resolve, reject) => {
        var numResponses = 0;
        var numErrors = 0;
        this.providers.forEach(provider => {
          provider
            .request(args)
            .then(result => {
              numResponses++;
              if (this.selectedProvider) {
                //Provider already selected
                return;
              }

              this.setSelectedProvider(provider);

              //@ts-ignore
              if (provider.isWalletLink) {
                window.dispatchEvent(
                  //@ts-ignore
                  new CustomEvent('walletlink:addresses', { detail: result })
                );
              }

              //@ts-ignore
              this.emit('accountsChanged', result);
              //@ts-ignore
              resolve(result);
            })
            .catch(error => {
              numErrors++;
              if (numErrors === this.providers.length) {
                reject(error);
              }
            });
        });
      });
    }

    return Promise.reject(new Error('No provider selected for request ' + args.method));
  }

  public registerProviders(providers: [Web3Provider & SafeEventEmitter]) {
    for (var i = 0; i < providers.length; i++) {
      let provider = providers[i];
      if (this.selectedProvider == null && this.shouldSelectProvider(provider)) {
        //@ts-ignore
        this.setSelectedProvider(provider);
      }

      //@ts-ignore
      if (provider.isMetaMask) {
        this.overrideIsMetaMask = true;
      }

      this.providers.push(provider);
    }

    let chainId = this.chainId;
    this.emit('connect', { chainId });
  }

  private setSelectedProvider(provider: Web3Provider & SafeEventEmitter) {
    this.selectedProvider = provider;
    this.selectedProvider.on('chainChanged', chain => {
      if (this.selectedProvider === provider && !this.shouldSelectProvider(provider)) {
        this.selectedProvider = null;
      }

      this.emit('chainChanged', chain);
    });
    this.selectedProvider.on('accountsChanged', accounts => {
      if (this.selectedProvider === provider && !this.shouldSelectProvider(provider)) {
        this.selectedProvider = null;
      }

      this.emit('accountsChanged', accounts);
    });
    this.selectedProvider.on('connect', chain => this.emit('connect', chain));
    this.selectedProvider.on('message', message => this.emit('message', message));
  }

  private shouldSelectProvider(provider: Web3Provider & SafeEventEmitter): boolean {
    if (provider.chainId != null && provider.chainId != '0x1') {
      // If the user has set a provider with a non-mainnet chain, select that provider.
      return true;
    }

    return this.hasSelectedAddress(provider);
  }

  private hasSelectedAddress(provider: Web3Provider & SafeEventEmitter): boolean {
    return (
      //@ts-ignore
      typeof provider.selectedAddress !== 'undefined' &&
      //@ts-ignore
      provider.selectedAddress != null
    );
  }
}
