import { WalletLinkUI, WalletLinkUIOptions } from 'walletlink/build/npm/dist/provider/WalletLinkUI';
import { Snackbar, SnackbarInstanceProps } from 'walletlink/build/npm/dist/components/Snackbar';
import { AddressString } from 'walletlink/build/npm/dist/types';
import { injectCssReset } from 'walletlink/build/npm/dist/lib/cssReset';
import { prepend0x, getFavicon } from 'walletlink/build/npm/dist/util';
import crypto from 'crypto';
import {
  WalletLinkExtensionConnectionRequest,
  WalletLinkExtensionConnectionResponse,
} from './WalletLinkExtensionConnectionRequest';
import log from '../util/log';

export interface WalletLinkInPageUIOptions extends WalletLinkUIOptions {
  resetConnectionCallback: () => void;
}

export class WalletLinkInPageExtensionUI extends WalletLinkUI {
  private readonly snackbar: Snackbar;
  private attached = false;
  private _nextRequestId = 0;
  private _idBase: string;
  private resetConnectionCallback: () => void;
  private cancelLinkCallbacks = new Map<string, () => void>();
  private onAccountsCallbacks = new Map<string, (accounts: [AddressString]) => void>();

  constructor(options: Readonly<WalletLinkInPageUIOptions>) {
    super(options);

    this.snackbar = new Snackbar({
      darkMode: options.darkMode,
    });

    this._idBase = crypto.randomBytes(16).toString('hex');
    this.resetConnectionCallback = options.resetConnectionCallback;
  }

  attach(): void {
    if (this.attached) {
      throw new Error('WalletLinkUI is already attached');
    }

    let el = document.documentElement;
    const container = document.createElement('div');
    container.className = '-walletlink-css-reset';
    el.appendChild(container);

    this.snackbar.attach(container);
    this.attached = true;
    injectCssReset();
    window.addEventListener('message', event => {
      if (event.data.type != 'extensionUIResponse') return;
      log("In page got ui response from popup")
      log(event)
      if (event.data.data.action === WalletLinkExtensionConnectionResponse.parentDisconnected) {
        this.resetConnectionCallback();
        return;
      }

      if (typeof event.data.data.id !== 'string') return;

      let id = event.data.data.id;
      switch (event.data.data.action) {
        case WalletLinkExtensionConnectionResponse.requestEthereumAccountsCancel:
          this.cancelLinkCallbacks.get(id)?.();
          this.cancelLinkCallbacks.delete(id);
          break;
        case WalletLinkExtensionConnectionResponse.requestEthereumAccountsResponse:
          this.onAccountsCallbacks.get(id)?.(event.data.data.addresses);
          this.onAccountsCallbacks.delete(id);
          break;
      }
    });
  }

  requestEthereumAccounts(options: {
    onCancel: () => void;
    onAccounts: (accounts: [AddressString]) => void;
  }): void {
    let id = this.makeLinkRequestId();
    this.cancelLinkCallbacks.set(id, options.onCancel);
    this.onAccountsCallbacks.set(id, options.onAccounts);
    window.postMessage(
      {
        type: 'extensionUIRequest',
        data: {
          id: id,
          action: WalletLinkExtensionConnectionRequest.requestEthereumAccounts,
          dappInfo: {
            dappLogoURL: getFavicon()
          }
        },
      },
      '*'
    );
  }

  hideRequestEthereumAccounts(): void {
    //No-op this is handled by the extension itself already
  }

  showConnecting(options: { onCancel: () => void; onResetConnection: () => void }): () => void {
    const snackbarProps: SnackbarInstanceProps = {
      message: 'Confirm on phone',
      menuItems: [
        {
          isRed: true,
          info: 'Cancel transaction',
          svgWidth: '11',
          svgHeight: '11',
          path:
            'M10.3711 1.52346L9.21775 0.370117L5.37109 4.21022L1.52444 0.370117L0.371094 1.52346L4.2112 5.37012L0.371094 9.21677L1.52444 10.3701L5.37109 6.53001L9.21775 10.3701L10.3711 9.21677L6.53099 5.37012L10.3711 1.52346Z',
          defaultFillRule: 'inherit',
          defaultClipRule: 'inherit',
          onClick: options.onCancel,
        },
        {
          isRed: false,
          info: 'Reset connection',
          svgWidth: '10',
          svgHeight: '11',
          path:
            'M5.00008 0.96875C6.73133 0.96875 8.23758 1.94375 9.00008 3.375L10.0001 2.375V5.5H9.53133H7.96883H6.87508L7.80633 4.56875C7.41258 3.3875 6.31258 2.53125 5.00008 2.53125C3.76258 2.53125 2.70633 3.2875 2.25633 4.36875L0.812576 3.76875C1.50008 2.125 3.11258 0.96875 5.00008 0.96875ZM2.19375 6.43125C2.5875 7.6125 3.6875 8.46875 5 8.46875C6.2375 8.46875 7.29375 7.7125 7.74375 6.63125L9.1875 7.23125C8.5 8.875 6.8875 10.0312 5 10.0312C3.26875 10.0312 1.7625 9.05625 1 7.625L0 8.625V5.5H0.46875H2.03125H3.125L2.19375 6.43125Z',
          defaultFillRule: 'evenodd',
          defaultClipRule: 'evenodd',
          onClick: options.onResetConnection,
        },
      ],
    };

    return this.snackbar.presentItem(snackbarProps);
  }

  reloadUI(): void {
    document.location.reload();
  }

  inlineAccountsResponse(): boolean {
    return true;
  }

  public makeLinkRequestId(): string {
    // max nextId == max int32 for compatibility with mobile
    this._nextRequestId = (this._nextRequestId + 1) % 0x7fffffff;
    const id = this._nextRequestId;
    const idStr = this._idBase + prepend0x(id.toString(16));
    // unlikely that this will ever be an issue, but just to be safe
    const callback = this.cancelLinkCallbacks.get(idStr);
    if (callback) {
      // TODO - how to handle this case
      this.cancelLinkCallbacks.delete(idStr);
    }
    return idStr;
  }
}
