import {
  WalletLinkExtensionConnectionRequest,
  WalletLinkExtensionConnectionResponse,
} from './WalletLinkExtensionConnectionRequest';
import { ScopedLocalStorage } from 'walletlink/build/npm/dist/lib/ScopedLocalStorage';
import log from '../util/log';
import { ServiceWorkerStorageKeys } from '../serviceworker/ServiceWorkerStorageKeys';
class WalletLinkContent {
  private walletLinkOrigin = 'https://www.walletlink.org';
  private storage: ScopedLocalStorage;
  private hasMultipleProviders: boolean = false;

  constructor() {
    this.storage = new ScopedLocalStorage(`-walletlink:${this.walletLinkOrigin}`);
  }

  start() {
    if (location.protocol == 'chrome-extension:') {
      // Running in the extension's popup process return early as we don't inject walletlink in this case
      log('In content script for popup');
      return;
    } else {
      this.injectWalletLinkProvider();

      //Listen for extension UI events
      //@ts-ignore
      const listener = event => {
        this.handleExtensionUIRequestEvent(event.data);
      };
      window.addEventListener('message', listener);

      this.connectServiceWorker();

      chrome.runtime.onMessage.addListener((message, sender, sendResponse) => {
        switch (message.type) {
          case 'extensionUIResponse':
            return this.handlePopupResponseEvent(message);
          default:
            log('Unknown message ' + message.type);
            return false;
        }
      });
    }
  }

  /**
   * Service worker and its events tends to die after 5 minutes (limitation of Chrome extension v3)
   * We always need the service worker to be listening for events.  The solution here is to reconnect
   * to the service worker whenever it disconnects.
   */
  private connectServiceWorker() {
    let content = this;
    chrome.runtime.connect({ name: 'extensionUIRequest' }).onDisconnect.addListener(function () {
      log('Extension disconnected');
      content.connectServiceWorker();
    });
  }

  /**
   * Inject a thin wrapper that implements WalletlinkProvider request ethereum accounts interface and errors
   * out on any other request
   */
  private injectWalletLinkProvider() {
    log('Injecting walletlink');
    //Inject walletlink
    const script = document.createElement('script');
    script.type = 'text/javascript';
    script.setAttribute('async', 'false');
    script.src = chrome.runtime.getURL('requestProvider.js');
    const container = document.head || document.documentElement;
    container.insertBefore(script, container.children[0]);
    container.removeChild(script);
  }

  /**
   * Inject the full wallet link relay + UI and set it as window.walletLinkRelay
   */
  private injectWalletLinkRelay() {
    // Inject WalletLinkRelay
    const script = document.createElement('script');
    script.type = 'text/javascript';
    script.src = chrome.runtime.getURL('requestRelay.js');
    const content = this;
    script.onload = script.onload = function () {
      window.postMessage(
        {
          type: 'extensionUIResponse',
          data: {
            action: 'loadedWalletLinkRelay',
          },
        },
        '*'
      );

      content.listenForStorageUpdate()
    };
    const container = document.head || document.documentElement;
    container.appendChild(script);
  }

  private handlePopupResponseEvent(event: any): void {
    log('In content script sending response back to page');
    log(event);

    if (typeof event.data.id === 'undefined') {
      log('Undefined id in service worker response handler');
      return;
    }

    window.postMessage(event, '*');
  }

  private handleExtensionUIRequestEvent(event: any) {
    if (event.type !== 'extensionUIRequest') {
      log('In content script, skipping event');
      log(event);
      return;
    }

    if (typeof event.data.id === 'undefined') {
      log('Undefined id in content script');
      return;
    }
    if (typeof event.data.action === 'undefined') {
      log('Undefined action in content script');
      return;
    }

    switch (event.data.action) {
      case WalletLinkExtensionConnectionRequest.requestEthereumAccounts:
        this.createEthereumAccountsRequest(event);
        break;
      case 'loadWalletLinkRelay':
        this.injectWalletLinkRelay();
        break;
      case 'hasMultipleProviders':
        this.hasMultipleProviders = true
        break;
      default:
        log(`Got unknown action type ${event.data.action}`);
    }
  }

  private createEthereumAccountsRequest(event: any) {
    const secret = this.storage.getItem('session:secret');
    const id = this.storage.getItem('session:id');
    const requestEthereumAccountsRequest = {
      type: event.type,
      data: {
        id: event.data.id,
        action: event.data.action,
        dappInfo: {
          dappLogoURL: event.data.dappInfo.dappLogoURL
        },
        childSession: {
          id: id,
          secret: secret,
        },
        hasMultipleProviders: this.hasMultipleProviders
      },
    };
    chrome.runtime.sendMessage(requestEthereumAccountsRequest);
  }

  private listenForStorageUpdate() {
    chrome.storage.onChanged.addListener(changes => {
      let changedValue = changes[ServiceWorkerStorageKeys.parentConnected];
      let hasAddresses = this.storage.getItem("Addresses") != null
      if (typeof changedValue !== 'undefined' && !changedValue.newValue && hasAddresses) {
        window.postMessage(
          {
            type: 'extensionUIResponse',
            data: {
              action: WalletLinkExtensionConnectionResponse.parentDisconnected,
            },
          },
          '*'
        );
      }
    });

    chrome.storage.local.get([ServiceWorkerStorageKeys.parentConnected], storageItem => {
      let hasAddresses = this.storage.getItem("Addresses") != null
      if (typeof storageItem !== 'undefined' && !storageItem.parentConnected && hasAddresses) {
        window.postMessage(
          {
            type: 'extensionUIResponse',
            data: {
              action: WalletLinkExtensionConnectionResponse.parentDisconnected,
            },
          },
          '*'
        );
      }
    });
  }
}

new WalletLinkContent().start();
