import { WalletLinkProvider } from 'walletlink/build/npm/dist/provider/WalletLinkProvider';
import { WalletLinkRelayEventManager } from 'walletlink/build/npm/dist/relay/WalletLinkRelayEventManager';
import { ScopedLocalStorage } from 'walletlink/build/npm/dist/lib/ScopedLocalStorage';
import { WalletLinkRelayAbstract } from 'walletlink/build/npm/dist/relay/WalletLinkRelayAbstract';
import { MultiProvider } from './MultiProvider';
import log from '../util/log';
import { SelectedChain } from '../datalayer/ethereum/models/EthereumChain';

const WALLETLINK_URL = 'https://www.walletlink.org';
const JSON_RPC_URL = SelectedChain.nodeUrl;

/**
 * The flow for loading the request provider and lazy loading the walletlink relay + in-page UI:
 * Content script -> Inject requestProvider
 * In Page -> Instantiate RequestProvider and set as window.walletLinkExtension and window.ethereum
 * In Page -> Dapp calls window.ethereum.sendAsync({"eth_requestAccounts"})
 * In Page -> new Promise to resolve WalletLinkRelay
 * In Page -> Request Provider window.sendMessage loadWalletLinkRelay
 * Content script -> get message loadWalletLinkRelay
 * Content script -> inject requestRelay.js
 *     -> on load of requestRelay.js emit message loadedWalletLinkRelay
 * In Page -> Handle loadedWalletLinkRelay message and resolve Promise
 */
export class RequestProvider {
  public relayEventManager: WalletLinkRelayEventManager;
  public storage: ScopedLocalStorage;

  constructor() {
    this.relayEventManager = new WalletLinkRelayEventManager();

    this.storage = new ScopedLocalStorage(`-walletlink:${WALLETLINK_URL}`);
  }

  provide(): WalletLinkProvider {
    const relayProvider: () => Promise<WalletLinkRelayAbstract> = () => {
      return new Promise((resolve, reject) => {
        const id = new Uint8Array(16);
        window.crypto.getRandomValues(id);

        window.postMessage(
          {
            type: 'extensionUIRequest',
            data: {
              id: id.toString(),
              action: 'loadWalletLinkRelay',
            },
          },
          '*'
        );

        window.addEventListener('message', event => {
          if (event.data.type != 'extensionUIResponse') return;

          log('In request provider got response from content script');
          log(event);

          if (event.data.data.action === 'loadedWalletLinkRelay') {
            //@ts-ignore
            resolve(window.walletLinkRelay);
          }
        });
      });
    };

    return new WalletLinkProvider({
      relayProvider: relayProvider,
      relayEventManager: this.relayEventManager,
      storage: this.storage,
      jsonRpcUrl: JSON_RPC_URL,
      chainId: Number(SelectedChain.chainId),
      overrideIsMetaMask: false,
    });
  }
}

//@ts-ignore
window.walletLinkRequestProvider = new RequestProvider();

//@ts-ignore
window.walletLinkExtension = window.walletLinkRequestProvider.provide();

if (typeof window.ethereum !== 'undefined') {
  let multiProvider = new MultiProvider();
  //@ts-ignore
  multiProvider.registerProviders([window.walletLinkExtension, window.ethereum]);

  const id = new Uint8Array(16);
  window.crypto.getRandomValues(id);
  window.postMessage(
    {
      type: 'extensionUIRequest',
      data: {
        id: id.toString(),
        action: 'hasMultipleProviders',
      },
    },
    '*'
  );

  //@ts-ignore
  window.ethereum = multiProvider;
} else {
  //@ts-ignore
  window.ethereum = window.walletLinkExtension;
}

/**
 * Returns true for `provider.isToshi()`
 * This makes Dapps that integrate with Web3Modal show "Coinbase Wallet"
 * instead of showing a generic "Web3 wallet"
 */
(WalletLinkProvider.prototype as any).isToshi = function () {
  return true;
};
