import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { filterNullish } from '@core/extensions/Array';
import { Startable } from '@core/interfaces/Startable';
import { Observable, of, zip } from 'rxjs';
import { flatMap, map } from 'rxjs/operators';
import { CompoundFinanceAPI } from '../apis/CompoundFinanceAPI';
import { CompoundTokenDAO } from '../daos/CompoundTokenDAO';
import { CompoundExchangeRateResult } from '../models/CompoundExchangeRateResult';
import { CompoundToken } from '../models/CompoundToken';

export class CompoundFinanceRepository implements Startable {
  compoundTokenDAO: CompoundTokenDAO;
  compoundFinanceAPI: CompoundFinanceAPI;

  constructor(compoundTokenDAO: CompoundTokenDAO, compoundFinanceAPI: CompoundFinanceAPI) {
    this.compoundTokenDAO = compoundTokenDAO;
    this.compoundFinanceAPI = compoundFinanceAPI;
  }

  private tokens = new Map<string, CompoundToken>();

  supportedTokens(): Observable<CompoundToken[]> {
    return of([...this.tokens.values()]);
  }

  /**
   * Get a list of supported Compound Tokens
   *
   * @return A Single wrapping a list of supported Compound Tokens
   */
  getTokens(): Observable<CompoundToken[]> {
    return of(this.tokens).pipe(
      flatMap(tokens => {
        if (tokens.size != 0) {
          return of([...tokens.values()]);
        }

        return zip(
          this.compoundFinanceAPI.getTokens(),
          this.compoundFinanceAPI.getCryptoCurrencies()
        ).pipe(
          map(([cTokensInfo, allTokens]) => {
            const contractAddressToDecimalsMap = new Map(
              allTokens
                .filter(token => token.contractAddress !== undefined)
                .map(token => [token.contractAddress, token.decimals])
            );
            const ethDecimals =
              allTokens.find(
                it => it.code === CurrencyCode.ETH.rawValue && it.contractAddress === undefined
              )?.decimals ?? BigInt(18);

            return filterNullish(
              cTokensInfo.map(cTokenInfo => {
                let underlyingDecimals = contractAddressToDecimalsMap.get(
                  cTokenInfo.underlying_address
                );
                if (underlyingDecimals === undefined) {
                  if (
                    cTokenInfo.underlying_symbol == CurrencyCode.ETH.rawValue &&
                    cTokenInfo.underlying_address === undefined
                  ) {
                    underlyingDecimals = ethDecimals;
                  } else {
                    return undefined;
                  }
                }

                return new CompoundToken(
                  cTokenInfo.name,
                  new CurrencyCode(cTokenInfo.symbol),
                  BigInt(8),
                  cTokenInfo.token_address,
                  new CurrencyCode(cTokenInfo.underlying_symbol),
                  underlyingDecimals,
                  cTokenInfo.underlying_address
                );
              })
            );
          })
        );
      })
    );
  }

  /**
   * Get list compound token info for given currency codes
   *
   * @param contractAddresses List of cToken contract addresses
   *
   * @return A single wrapping a list of token info
   */
  getTokensWithContractAddresses(contractAddresses: string[]): CompoundToken[] {
    return filterNullish(contractAddresses.map(address => this.tokens.get(address)));
  }

  getExchangeRates(contractAddresses: string[]): Observable<CompoundExchangeRateResult[]> {
    return zip(
      ...this.getTokensWithContractAddresses(contractAddresses).map(token =>
        this.compoundFinanceAPI
          .getExchangeRate(token.contractAddress)
          .pipe(map(rate => new CompoundExchangeRateResult(token, rate)))
      )
    );
  }

  start(): Observable<any> {
    return this.getTokens()
      .pipe(flatMap(token => this.compoundTokenDAO.saveCompoundTokens(token)))
      .pipe(
        map(savedTokens => {
          this.tokens = new Map(
            savedTokens.map(savedToken => [savedToken.contractAddress, savedToken])
          );
        })
      );
  }
}
