import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { CompoundFinanceAPI } from '@compoundfinance/apis/CompoundFinanceAPI';
import { CompoundTokenDAO } from '@compoundfinance/daos/CompoundTokenDAO';
import { CompoundToken } from '@compoundfinance/models/CompoundToken';
import { WalletsDatabase } from '@wallets/db/WalletsDatabase';
import axios from 'axios';
import Dexie from 'dexie';
import { CompoundFinanceRepository } from '../CompoundFinanceRepository';
import MockAdapter from 'axios-mock-adapter';

Dexie.dependencies.indexedDB = require('fake-indexeddb');
Dexie.dependencies.IDBKeyRange = require('fake-indexeddb/lib/FDBKeyRange');

const mockedAxios = new MockAdapter(axios);

const compoundTokens = [
  {
    name: 'Compound Sai',
    symbol: 'cSAI',
    token_address: '0xf5dce57282a584d2746faf1593d3121fcac444dc',
    underlying_symbol: 'SAI',
    underlying_address: '0x89d24a6b4ccb1b6faa2625fe562bdd9a23260359',
  },
  {
    name: 'Compound Wrapped BTC',
    symbol: 'cWBTC',
    token_address: '0xc11b1268c1a384e55c48c2391d8d480264a3a7f4',
    underlying_symbol: 'WBTC',
    underlying_address: '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599',
  },
];

const cryptoCurrencies = [
  {
    name: 'Compound Wrapped BTC',
    code: 'CWBTC',
    imageUrl: 'https://1.png',
    blockchain: 'eth',
    decimals: 8,
    marketCap: 758901722568.4563,
    contractAddress: '0xc11b1268c1a384e55c48c2391d8d480264a3a7f4',
  },
  {
    name: 'Wrapped Bitcoin',
    code: 'WBTC',
    imageUrl:
      'https://wallet-api-production.s3.amazonaws.com/uploads/tokens/0x2260fac5e5542a773aa44fbcfedf7c193bc2c599_288.png',
    blockchain: 'eth',
    decimals: 8,
    marketCap: 7401978010.076866,
    contractAddress: '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599',
  },
  {
    name: 'Ethereum',
    code: 'ETH',
    imageUrl: 'https://2.png',
    blockchain: 'eth',
    decimals: 18,
    marketCap: 300249833368.14075,
  },
];

describe('Compound Finance repository', () => {
  it('should return cTokens', () => {
    const expectedTokens = [
      new CompoundToken(
        'Compound Wrapped BTC',
        new CurrencyCode('cWBTC'),
        8n,
        '0xc11b1268c1a384e55c48c2391d8d480264a3a7f4',
        new CurrencyCode('WBTC'),
        8n,
        '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599'
      ),
    ];

    const walletsDatabase = new WalletsDatabase();
    if (walletsDatabase.compoundTokenTable == null) {
      throw new Error('Undefined');
    }

    const compoundRepository = new CompoundFinanceRepository(
      new CompoundTokenDAO(walletsDatabase.compoundTokenTable),
      new CompoundFinanceAPI('infuraKey', 'https://api.wallet.coinbase.com')
    );

    mockedAxios
      .onGet('https://compound.wallet.coinbase.com/api/v2/ctoken')
      .reply(200, { cToken: compoundTokens });
    mockedAxios
      .onGet('https://api.wallet.coinbase.com/rpc/v2/getCryptoCurrencies')
      .reply(200, { result: { cryptoCurrencies } });

    return compoundRepository
      .start()
      .toPromise()
      .then(_ => {
        return compoundRepository
          .supportedTokens()
          .toPromise()
          .then(data => {
            expect(data.length).toEqual(expectedTokens.length);
            data.forEach((res, index) => {
              const resStr = JSON.stringify(res, (key, value) => value.toString());
              const actualStr = JSON.stringify(expectedTokens[index], (key, value) =>
                value.toString()
              );
              expect(resStr).toEqual(actualStr);
            });
          })
          .catch(error => {
            expect(error).toBeNull();
          });
      });
  });

  it('should return ExchangeRates', () => {
    const expectedRate = '0x1234';

    const walletsDatabase = new WalletsDatabase();
    if (walletsDatabase.compoundTokenTable == null) {
      throw new Error('Undefined');
    }

    const compoundRepository = new CompoundFinanceRepository(
      new CompoundTokenDAO(walletsDatabase.compoundTokenTable),
      new CompoundFinanceAPI('infuraKey', 'https://api.wallet.coinbase.com')
    );

    mockedAxios
      .onGet('https://compound.wallet.coinbase.com/api/v2/ctoken')
      .reply(200, { cToken: compoundTokens });
    mockedAxios
      .onGet('https://api.wallet.coinbase.com/rpc/v2/getCryptoCurrencies')
      .reply(200, { result: { cryptoCurrencies } });
    mockedAxios
      .onPost('https://mainnet-infura.wallet.coinbase.com/v3/infuraKey')
      .reply(200, { jsonrpc: 'jsonrpc', id: 0, result: expectedRate });

    return compoundRepository
      .start()
      .toPromise()
      .then(_ => {
        return compoundRepository
          .getExchangeRates(['0xc11b1268c1a384e55c48c2391d8d480264a3a7f4'])
          .toPromise()
          .then(data => {
            expect(data.length).toEqual(1);
            expect(data[0].rate).toEqual(BigInt(expectedRate));
            expect(data[0].token.name).toEqual('Compound Wrapped BTC');
          })
          .catch(error => {
            expect(error).toBeNull();
          });
      });
  });
});
