import { CipherCore } from 'cipher-core';
import { KeyPair } from 'cipher-core/dist/CipherCore/types';
import { Observable } from 'rxjs';
import { flatMap, map } from 'rxjs/operators';

declare global {
  interface CipherCore {
    getEthereumWalletKeyPair(index: bigint, mnemonic: string): Observable<KeyPair>;
    keyPair(derivationPath: string, mnemonic: string): Observable<KeyPair>;
    address(derivationPath: string, mnemonic: string): Observable<String>;
  }
}
/**
 * Create Ethereum wallet key pair based on current active mnemonic and index
 *
 * @param index: Index of wallet
 * @param mnemonic: The mnemonic used to generate the key pair
 *
 * @return A single wrapping ethereum wallet `KeyPair` or an error is thrown
 */
export function getEthereumWalletKeyPair(thiz: CipherCore, index: bigint, mnemonic: string): Observable<KeyPair> {
  let derivationPath = ethereumAddressDerivationPath((index = index));

  return keyPair(thiz, derivationPath, mnemonic);
}

/**
 * Returns the derivation path for the ethereum address at the given index
 *
 * @param index Index of the wallet
 *
 * @return The derivation path
 */
function ethereumAddressDerivationPath(index: bigint): string {
  return `m/44'/60'/0'/0/${index}`;
}

/**
 * Returns the Ethereum address for the wallet at the given index
 *
 * @param index Index of the wallet
 * @param mnemonic The mnemonic from which to derive the address
 *
 * @return A Single wrapping the ethereum address
 */
function ethereumAddress(thiz: CipherCore, index: bigint, mnemonic: string): Observable<string> {
  return address(thiz, ethereumAddressDerivationPath((index = index)), mnemonic);
}

/**
 * Get ethereum address at given derivation path
 *
 * @param derivationPath Path used to generate address
 * @param mnemonic The mnemonic from which to derive the address
 *
 * @return Ethereum address at the given derivation path
 */
function address(thiz: CipherCore, derivationPath: string, mnemonic: string): Observable<string> {
  return keyPair(thiz, derivationPath, mnemonic).pipe(
    flatMap(keyPair => thiz.ethereumAddressFromPublicKey(keyPair.publicKey))
  );
}

function keyPair(thiz: CipherCore, derivationPath: string, mnemonic: string): Observable<KeyPair> {
  return new Observable<Buffer>(observer => {
    thiz.seedFromRecoveryPhrase(mnemonic)
      .then(seed => {
        observer.next(seed);
        observer.complete();
      })
      .catch(error => observer.error(error));
  }).pipe(map(seed => thiz.deriveKeyPair(seed, derivationPath)));
}
