import { HTTPService } from './models/HTTPService';
import { Observable } from 'rxjs';
import axios, { AxiosRequestConfig } from 'axios';
import { HTTPResponse } from './models/HTTPResponse';
import { HTTPRequestFailed } from './errors/HTTPError';

export class HTTP {
  static get<T>(
    service: HTTPService,
    path: string,
    parameters?: Map<string, any>,
    headers?: Map<string, string>
  ): Observable<HTTPResponse<T>> {
    const request = this.buildGetRequest(service, path, parameters, headers);

    return new Observable(observer => {
      axios
        .get(request.url, request.config)
        .then(response => {
          const parsedResponse: T = response.data as T;
          observer.next(new HTTPResponse(response.headers, parsedResponse));
          observer.complete();
        })
        .catch(error => {
          if (error.response && error.response.data) {
            let err = new HTTPRequestFailed(error.response, error.response.data);
            observer.error(err);
            return;
          }
          observer.error(error);
        });
    });
  }

  static post<T>(
    service: HTTPService,
    path: string,
    parameters?: Map<string, any>,
    headers?: Map<string, string>
  ): Observable<HTTPResponse<T>> {
    const request = this.buildPostRequest(service, path, parameters, headers);

    return new Observable(observer => {
      axios
        .post(request.url, request.requestBody, request.config)
        .then(response => {
          const parsedResponse: T = response.data as T;
          observer.next(new HTTPResponse(response.headers, parsedResponse));
          observer.complete();
        })
        .catch(error => {
          if (error.response && error.response.data) {
            let err = new HTTPRequestFailed(error.response, error.response.data);
            observer.error(err);
            return;
          }
          observer.error(error);
        });
    });
  }

  private static buildGetRequest(
    service: HTTPService,
    path: string,
    parameters?: Map<string, string>,
    headers?: Map<string, string>
  ): Request {
    let url = `${service.url}${path}`;
    let config: AxiosRequestConfig;
    if (typeof headers === 'undefined') {
      config = {};
    } else {
      config = { headers: Object.fromEntries(headers) };
    }

    if (typeof parameters !== 'undefined') {
      config.params = Object.fromEntries(parameters);
    }

    return { url, config };
  }

  private static buildPostRequest(
    service: HTTPService,
    path: string,
    parameters?: Map<string, any>,
    headers?: Map<string, string>
  ): Request {
    let url = `${service.url}${path}`;
    let config: AxiosRequestConfig;
    if (typeof headers === 'undefined') {
      config = {};
    } else {
      config = { headers: Object.fromEntries(headers) };
    }

    let requestBody: object | null;
    if (typeof parameters !== 'undefined') {
      requestBody = Object.fromEntries(parameters);
    } else {
      requestBody = null;
    }

    return { url, config, requestBody };
  }
}

interface Request {
  url: string;
  config?: AxiosRequestConfig;
  requestBody?: any;
}
