import { Observable } from 'rxjs';
import { catchError, flatMap, map, retryWhen } from 'rxjs/operators';
import { HTTP } from './HTTP';
import { RPCResponse } from './models/RPCResponse';
import { HTTPService } from './models/HTTPService';
import { HTTPError } from './errors/HTTPError';
import { AuthenticationRepository } from '@user/repositories/AuthenticationRepository';
import { AccessTokenError } from '@user/exceptions/AccessTokenError';
import { RPCError } from './errors/RPCError';

export class RPC {
  private static maxRetryAttempts = 3;
  private static authenticationRepository: AuthenticationRepository;

  /**
   * Set up with authenticationRepository
   */
  static setup(authenticationRepository: AuthenticationRepository) {
    RPC.authenticationRepository = authenticationRepository;
  }

  /**
   * Perform a GET RPC call
   *
   * @param path: The path portion of the URL
   * @param parameters: The parameters of the call
   * @param headers: Optional map of headers
   * @param isAuthRequired: If true, the call will be made with an access token
   * Including the result type here allows for cleaner syntax by callers
   *
   * @returns An Observable wrapping the result
   */
  static get<T>(
    service: HTTPService,
    path: string,
    isAuthRequired = false,
    parameters?: Map<string, string>,
    headers?: Map<string, string>
  ): Observable<T> {
    if (isAuthRequired) {
      return RPC.getProtected<T>(service, path, parameters, headers)
    }

    return HTTP.get<RPCResponse<T>>(service, path, parameters, headers).pipe(
      map(response => response.body.result!), //Always defined, not sure why it's saying undefined
      catchError(error => {
        if (error instanceof HTTPError) {
          throw error.asRPCError ?? error;
        }
        throw error;
      })
    );
  }

  /**
   * Perform a POST RPC call
   *
   * @param path: The path portion of the URL
   * @param parameters: The parameters of the call
   * @param isAuthRequired: If true, the call will be made with an access token
   * Including the result type here allows for cleaner syntax by callers
   *
   * @returns An Observable wrapping the result
   */
  static post<T>(
    service: HTTPService,
    path: string,
    isAuthRequired = false,
    parameters?: Map<string, any>
  ): Observable<T> {
    if (isAuthRequired) {
      return RPC.postProtected(service, path, parameters)
    }

    return HTTP.post<RPCResponse<T>>(service, path, parameters).pipe(
      map(response => response.body.result!),
      catchError(error => {
        if (error instanceof HTTPError) {
          throw error.asRPCError ?? error;
        }
        throw error;
      })
    );
  }

  private static getProtected<T>(
    service: HTTPService,
    path: string,
    parameters?: Map<string, string>,
    extraHeaders?: Map<string, string>
  ): Observable<T> {
    if (!RPC.authenticationRepository) {
      return new Observable(observable =>
        observable.error(AccessTokenError.accessTokenManagerNotConfigured)
      );
    }

    var token: string | undefined;

    return RPC.authenticationRepository.getAccessToken().pipe(
      flatMap((accessToken: string) => {
        token = accessToken;
        var headers: Map<string, string> = new Map([['Authorization', `Bearer ${accessToken}`]]);

        if (extraHeaders) {
          // Merge extra headers with existing headers, override existing headers
          headers = new Map([
            ...Array.from(headers.entries()),
            ...Array.from(extraHeaders.entries()),
          ]);
        }

        return HTTP.get<RPCResponse<T>>(service, path, parameters, headers);
      }),
      map(response => response.body.result!),
      retryWhen(errors => {
        return errors.pipe(
          flatMap((error, attempt) => {
            if (attempt >= RPC.maxRetryAttempts || !token) {
              return new Observable(observer => observer.error(error));
            }

            return RPC.handleProtectedApiError(error, token);
          })
        );
      })
    );
  }

  private static postProtected<T>(service: HTTPService, path: string, parameters?: Map<string, any>): Observable<T> {
    if (!RPC.authenticationRepository) {
      return new Observable(observable =>
        observable.error(AccessTokenError.accessTokenManagerNotConfigured)
      );
    }

    var token: string | undefined;

    return RPC.authenticationRepository.getAccessToken().pipe(
      flatMap((accessToken: string) => {
        token = accessToken;
        var headers: Map<string, string> = new Map([['Authorization', `Bearer ${accessToken}`]]);

        return HTTP.post<RPCResponse<T>>(service, path, parameters, headers);
      }),
      map(response => response.body.result!),
      retryWhen(errors => {
        return errors.pipe(
          flatMap((error, attempt) => {
            if (attempt >= RPC.maxRetryAttempts || !token) {
              return new Observable(observer => observer.error(error));
            }

            return RPC.handleProtectedApiError(error, token);
          })
        );
      })
    );
  }

  private static handleProtectedApiError(error: Error, token: string): Observable<void> {
    var rpcError: RPCError | undefined;
    if (error instanceof HTTPError) {
      rpcError = error.asRPCError;
    }

    if (!RPC.authenticationRepository || !rpcError) {
      return new Observable(observer => observer.error(error));
    }

    if (rpcError.code !== 'invalidAuthenticationToken') {
      return new Observable(observer => observer.error(rpcError));
    }

    //TODO there's logic on the native clients to only invalidate the access token
    //once we're online.  We don't have a way to check for that yet in the
    //data layer
    RPC.authenticationRepository.invalidateAccessToken(token);
    return new Observable(observer => observer.complete())
  }
}
