import { BehaviorSubject, Observable } from 'rxjs';
import { StoreError } from './exceptions/StoreError';
import { Storage } from './interfaces/Storage';
import { StoreKey } from './models/StoreKey';
import { StoreKind } from './models/StoreKind';
import { LocalStorage } from './storages/LocalStorage';
import { MemoryStorage } from './storages/MemoryStorage';

export class Store {
  private localStorage: LocalStorage;
  private memoryStorage: MemoryStorage;
  private changeObservers = new Map<String, BehaviorSubject<any>>();
  private isDestroyed: boolean = false;
  constructor(localStorage: LocalStorage, memoryStorage: MemoryStorage) {
    this.localStorage = localStorage;
    this.memoryStorage = memoryStorage;
  }

  set<T>(key: StoreKey<T>, value: T | undefined) {
    let hasObserver = this.changeObservers.get(key.name) != null;
    if (this.isDestroyed) {
      if (hasObserver) {
        this.observer(key).error(StoreError.storeDestroyed);
      }
      return;
    }

    this.storageForKey(key).set(key, value);
    this.observer(key).next(value);
  }

  get<T>(key: StoreKey<T>): T | undefined {
    if (this.isDestroyed) return undefined;

    return this.storageForKey(key).get(key);
  }

  has<T>(key: StoreKey<T>): boolean {
    if (this.isDestroyed) return false;

    return this.get(key) != null;
  }

  observe<T>(key: StoreKey<T>): Observable<T | undefined> {
    if (this.isDestroyed) {
      return new Observable(observer => observer.error(StoreError.storeDestroyed));
    }

    return this.observer(key);
  }

  destroy() {
    if (this.isDestroyed) {
      return;
    }

    this.isDestroyed = true;
    this.deleteAllEntries([StoreKind.LocalStorage, StoreKind.Memory]);
  }

  removeAll(kinds: StoreKind[]) {
    if (this.isDestroyed) {
      return;
    }

    this.deleteAllEntries(kinds);
    this.changeObservers.forEach((key, _) => {
      key.next(null);
    });
  }

  private deleteAllEntries(kinds: StoreKind[]) {
    kinds.forEach(kind => {
      switch (kind) {
        case StoreKind.LocalStorage:
          this.localStorage.destroy();
        case StoreKind.Memory:
          this.memoryStorage.destroy();
      }
    });
  }

  private storageForKey<T>(key: StoreKey<T>): Storage {
    switch (key.kind) {
      case StoreKind.LocalStorage:
        return this.localStorage;
      case StoreKind.Memory:
        return this.memoryStorage;
    }
  }

  private observer<T>(key: StoreKey<T>): BehaviorSubject<T | undefined> {
    let currentObserver: BehaviorSubject<T | undefined> | undefined = this.changeObservers.get(
      key.name
    );

    if (typeof currentObserver !== 'undefined') {
      return currentObserver;
    }

    let value = this.get(key);
    let newObserver = new BehaviorSubject(value);

    this.changeObservers.set(key.name, newObserver);
    return newObserver;
  }
}
