import { StoreError } from '@store/exceptions/StoreError';
import { LocalStorageStoreKey } from '@store/models/LocalStorageStoreKey';
import { MemoryStorageKey } from '@store/models/MemoryStorageStoreKey';
import { LocalStorage } from '@store/storages/LocalStorage';
import { MemoryStorage } from '@store/storages/MemoryStorage';
import { Store } from '@store/Store';

class Bar {
  rawValue: string;

  constructor(value: string) {
    this.rawValue = value;
  }
}

class Foo {
  id: String;
  bar: Bar;
  private bazStr: String;

  constructor(bar: Bar, baz: bigint) {
    this.id = bar.rawValue;
    this.bar = bar;
    this.bazStr = baz.toString();
  }

  get baz(): bigint {
    return BigInt(this.bazStr);
  }
}


const StoreKey_fooMemory = new MemoryStorageKey<Foo>('fooKey');
const StoreKey_fooLocal = new LocalStorageStoreKey<Foo>('fooKey');

LocalStorage.mappedPrototypes.set(StoreKey_fooLocal.name, Foo.prototype)

describe('memory storage', () => {
  it('should set value', () => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    let expected = new Foo(new Bar('bar'), 18n);
    store.set(StoreKey_fooMemory, expected);
    expect(store.get(StoreKey_fooMemory)).toEqual(expected);
  });

  it('should observe default value', () => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    let expected = new Foo(new Bar('bar'), 18n);
    store.set(StoreKey_fooMemory, expected);

    return new Promise<Foo>((resolve, reject) => {
      store.observe<Foo>(StoreKey_fooMemory).subscribe(result => resolve(result!));
    }).then(result => {
      expect(result!.bar).toEqual(expected.bar);
      expect(result!.baz.toString()).toEqual(expected.baz.toString());
      expect(result!.id).toEqual(expected.id);
    });
  });
});

describe('local storage', () => {
  it('should set value', () => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    let expected = new Foo(new Bar('bar'), 18n);

    store.set(StoreKey_fooLocal, expected);
    let result = store.get<Foo>(StoreKey_fooLocal);
    expect(result!.bar).toEqual(expected.bar);
    expect(result!.baz.toString()).toEqual(expected.baz.toString());
    expect(result!.id).toEqual(expected.id);
  });

  it('should observe default value', () => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    let expected = new Foo(new Bar('bar'), 18n);
    store.set(StoreKey_fooLocal, expected);

    return new Promise<Foo>((resolve, reject) => {
      store.observe<Foo>(StoreKey_fooLocal).subscribe(result => resolve(result!));
    }).then(result => {
      expect(result!.bar).toEqual(expected.bar);
      expect(result!.baz.toString()).toEqual(expected.baz.toString());
      expect(result!.id).toEqual(expected.id);
    });
  });
});

describe('store', () => {
  it('should return null when destroyed', () => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    let expected = new Foo(new Bar('bar'), 18n);
    store.set(StoreKey_fooMemory, expected);

    store.destroy();

    expect(store.get(StoreKey_fooMemory)).toBeUndefined();
  });

  it('observer should error out when destroyed', () => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    let expected = new Foo(new Bar('bar'), 18n);
    store.set(StoreKey_fooMemory, expected);

    store.destroy();

    return store
      .observe(StoreKey_fooMemory)
      .toPromise()
      .then(value => fail('Got value when it should have failed ' + value))
      .catch(error => expect(error).toEqual(StoreError.storeDestroyed));
  });
});
