import { HTTPService } from '@http/models/HTTPService';
import { RPC } from '@http/RPC';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import { Signature } from '../repositories/AuthenticationRepository';

export class AuthenticationAPI {
  private service: HTTPService;
  constructor(service: HTTPService) {
    this.service = service;
  }

  /**
   * Create a nonce for an API request
   *
   * @return An observable wrapping a nonce
   */
  createNonce(): Observable<string> {
    return RPC.post<CreateNonceResponseDTO>(this.service, '/rpc/v2/createNonce').pipe(
      map((nonceResponse: CreateNonceResponseDTO) => nonceResponse.nonce)
    );
  }

  /**
   * Create a new user
   *
   * @param username: The username for the new user
   * @param isPublic: If true, the user will be searchable by other Wallet users
   * @param nonce: The nonce for the request
   * @param signatures: The user's ethereum addresses and corresponding signatures
   *
   * @return An Observable wrapping an AccessTokenResult
   */
  createUser(
    username: string,
    isPublic: boolean,
    nonce: string,
    signatures: [Signature]
  ): Observable<AccessTokenResult> {
    let parameters = new Map<string, any>([
      ['nonce', nonce],
      ['username', username],
      ['isPublic', isPublic],
      ['ethereumAddresses', signatures],
    ]);

    return RPC.post(this.service, '/rpc/v2/createUser', false, parameters);
  }

  /**
   * Sign in an existing user
   *
   * @param nonce: The nonce for the request
   * @param address: The user's ethereum payment address
   * @param signature: The user's signature
   *
   * @return An Observable wrapping an AccessTokenResult
   */
  createAccessToken(nonce: string, address: string, signature: string): Observable<AccessTokenResult> {
    let parameters = new Map<string, any>([
      ['nonce', nonce],
      ['address', address],
      ['signature', signature],
    ]);

    return RPC.post(this.service, '/rpc/v2/createAccessToken', false, parameters);
  }

  /**
   * Refresh the user's access token
   *
   * @param Parameter refreshToken: The user's refresh token
   *
   * @return An Observable wrapping an AccessTokenResult
   */
  refreshAccessToken(refreshToken: string): Observable<AccessTokenResult> {
    let parameters = new Map<string, any>([['refreshToken', refreshToken]]);

    return RPC.post(this.service, '/rpc/v2/refreshAccessToken', false, parameters);
  }
}
