import { Store } from '@store/Store';
import { AuthenticationAPI } from '../apis/AuthenticationAPI';
import { CipherCore } from 'cipher-core';
import { KeyPair } from 'cipher-core/dist/CipherCore/types';
import { Observable, of, onErrorResumeNext } from 'rxjs';
import { RPCError } from '@http/errors/RPCError';
import { catchError, flatMap, map } from 'rxjs/operators';
import { StoreKeys_accessToken, StoreKeys_refreshToken } from '../extensions/StoreKeys+User';
import { AccessTokenError } from '../exceptions/AccessTokenError';
import { getEthereumWalletKeyPair } from '@ethereum/extensions/CipherCore+Ethereum';
import { WalletLinkProvider } from 'walletlink';

export interface Signature {
  address: string;
  signature: string;
}

export class AuthenticationRepository {
  private api: AuthenticationAPI;
  private store: Store;
  private cipherCore: CipherCore;
  private walletLinkProvider: WalletLinkProvider;
  private hostname = 'api.wallet.coinbase.com';

  constructor(
    api: AuthenticationAPI,
    store: Store,
    cipherCore: CipherCore,
    walletLinkProvider: WalletLinkProvider
  ) {
    this.api = api;
    this.store = store;
    this.cipherCore = cipherCore;
    this.walletLinkProvider = walletLinkProvider;
  }

  getAccessToken(): Observable<string> {
    let accessToken = this.store.get<string>(StoreKeys_accessToken);
    if (accessToken) {
      return new Observable<string>(observable => {
        observable.next(accessToken);
        observable.complete();
      });
    }

    let refreshToken = this.store.get<string>(StoreKeys_refreshToken);
    if (refreshToken) {
      return this.api.refreshAccessToken(refreshToken).pipe(
        catchError(error => {
          if (error instanceof RPCError && error === RPCError.invalidRefreshToken) {
            throw new Error('Api told us refresh token is invalid, currently no way to refresh');
          }
          throw error;
        }),
        map(token => {
          this.setAccessToken(token.token, token.refreshToken);
          return token.token;
        })
      );
    } else {
      //Note: On mobile here we will re-sign the jwt
      //and get a new refresh token.
      throw new Error('Refresh token is invalid, currently no way to refresh');
    }
  }

  createUser(username: string, isPublic: boolean, mnemonic: string): Observable<void> {
    throw Error('Not implemented yet');
  }

  invalidateAccessToken(token: string) {
    this.store.set(StoreKeys_accessToken, null)
  }

  signIn(mnemonic: string): Observable<AccessTokenResult> {
    return this.api.createNonce().pipe(
      flatMap(nonce => {
        return this.signature('Coinbase Wallet Sign In', nonce, mnemonic, 0n).pipe<
          [string, Signature]
        >(
          map(signature => {
            return [nonce, signature];
          })
        );
      }),
      flatMap(([nonce, signature]: [string, Signature]) => {
        return this.api.createAccessToken(nonce, signature.address, signature.signature);
      }),
      map(token => {
        this.setAccessToken(token.token, token.refreshToken);
        return token;
      })
    );
  }

  signInWithWalletlink(address: string): Observable<AccessTokenResult> {
    return this.api.createNonce().pipe(
      flatMap(nonce => {
        return this.signatureWithWalletlink('Coinbase Wallet Sign In', nonce, address).pipe<
          [string, Signature]
        >(
          map(signature => {
            return [nonce, signature];
          })
        );
      }),
      //Not sure why this doesn't compile
      //   onErrorResumeNext(error => {
      //       return new Observable(observer => observer.error<[string, Signature]>(AccessTokenError.unableToCreateAccessToken))
      //   }),
      flatMap(([nonce, signature]: [string, Signature]) => {
        return this.api.createAccessToken(nonce, signature.address, signature.signature);
      }),
      map(token => {
        this.setAccessToken(token.token, token.refreshToken);
        return token;
      })
    );
  }

  private signature(
    header: string,
    nonce: string,
    mnemonic: string,
    index: bigint
  ): Observable<Signature> {
    return getEthereumWalletKeyPair(this.cipherCore, index, mnemonic).pipe(
      map((keyPair: KeyPair) => {
        let address = this.cipherCore.ethereumAddressFromPublicKey(keyPair.publicKey);
        let message = `${header}\n\nAddress: ${address}\nNonce: ${nonce}\nHost: ${this.hostname}`;
        let messageData = Buffer.from(message);
        let data = this.cipherCore.keccak256Hash(messageData);
        let signature = this.cipherCore.signEthereumMessage(data, keyPair.privateKey, false);
        return { address, signature: '0x' + signature.toString('hex') };
      })
    );
  }

  private signatureWithWalletlink(
    header: string,
    nonce: string,
    address: string
  ): Observable<Signature> {
    let message = `${header}\n\nAddress: ${address}\nNonce: ${nonce}\nHost: ${this.hostname}`;
    let messageData = Buffer.from(message);
    let data = this.cipherCore.keccak256Hash(messageData);
    return new Observable(observer => {
      this.walletLinkProvider
        .request<string>({ method: 'eth_sign', params: [address, data] })
        .then(result => {
          observer.next({ address, signature: result });
          observer.complete();
        })
        .catch(error => observer.error(error));
    });
  }

  private setAccessToken(accessToken: string, refreshToken: string) {
    this.store.set(StoreKeys_accessToken, accessToken);
    this.store.set(StoreKeys_refreshToken, refreshToken);
  }
}
