import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { HTTPService } from '@http/models/HTTPService';
import { RPC } from '@http/RPC';
import {
  cryptoExchangeRatesDTOToExchangeRates,
  GetCryptoExchangeRatesDTO,
} from '@wallets/dtos/GetCryptoExchangeRatesResultDTO';
import {
  GetFiatExchangeRatesDTO,
  fiatExchangeRatesDTOToExchangeRates,
} from '@wallets/dtos/GetFiatExchangeRatesDTO';
import { ExchangeRate } from '@wallets/models/ExchangeRate';
import { Observable } from 'rxjs';
import { delay, map, retryWhen, take } from 'rxjs/operators';
import { ContractAddress } from '@ethereum/models/ContractAddress';
import { Decimal } from 'decimal.js';
import { filterNullish } from '@core/extensions/Array';
import { HTTP } from '@http/HTTP';
import { ConsumerSpotPricesDTO } from '@wallets/dtos/ConsumerSpotPricesDTO';

export class ExchangeRateAPI {
  private service: HTTPService;
  constructor(service: HTTPService) {
    this.service = service;
  }

  /**
   * Get exchange rates for *fiat currencies*, based on USD
   *
   * @param currencyCodes The fiat currencies to get exchange rates for, always based on USD
   *
   * @return A [Single] wrapping a [List] of [ExchangeRate]s
   */
  getFiatExchangeRates(currencyCodes: CurrencyCode[]): Observable<ExchangeRate[]> {
    const encodedSymbols = this.encodeCurrencyCodes(currencyCodes);
    return RPC.get<GetFiatExchangeRatesDTO>(
      this.service,
      '/rpc/v2/getExchangeRates',
      false,
      new Map(
        Object.entries({
          symbols: encodedSymbols,
        })
      )
    )
      .pipe(retryWhen(errors => errors.pipe(delay(500), take(3))))
      .pipe(map(response => fiatExchangeRatesDTOToExchangeRates(response)));
  }

  /**
   * Get exchange rates for *crypto currencies*, based on USD
   *
   * @param currencyCodes The fiat currencies to get exchange rates for, always based on USD
   * @param contractAddresses List of contract addresses of ERC20 tokens to get exchange rates for
   * @param chainId Ethereum chainId, server uses main-net by default, unless provided different
   *
   * @return A [Single] wrapping a [List] of [ExchangeRate]s
   */
  getCryptoExchangeRates(
    currencyCodes: CurrencyCode[],
    contractAddresses: ContractAddress[],
    chainId?: bigint
  ): Observable<ExchangeRate[]> {
    const params = new Map(
      Object.entries({
        symbols: this.encodeCurrencyCodes(currencyCodes),
        contractAddresses: this.encodeContractAddresses(contractAddresses),
      })
    );

    if (chainId !== undefined) {
      params.set('chainId', chainId.toString());
    }

    return RPC.get<GetCryptoExchangeRatesDTO>(this.service, '/rpc/v2/getAssetPrices', false, params)
      .pipe(retryWhen(errors => errors.pipe(delay(500), take(3))))
      .pipe(map(response => cryptoExchangeRatesDTOToExchangeRates(response, chainId)));
  }

  /**
   * Make a request to get the spot prices from the Coinbase Api
   *
   * @return A [Single] wrapping the [List] of [ExchangeRate]
   */
  getConsumerSpotPrices(): Observable<ExchangeRate[]> {
    return HTTP.get<ConsumerSpotPricesDTO>(
      new HTTPService('https://api.coinbase.com'),
      '/v2/prices/USD/spot'
    ).pipe(
      map(response =>
        filterNullish(
          response.body.data.map(
            it => new ExchangeRate(new CurrencyCode(it.base), new Decimal(it.amount))
          )
        )
      )
    );
  }

  /**
   * Encode a list of currency codes into the query string format expected by the endpoint
   *
   * @param input The list of currency codes to encode
   *
   * @return The encoded currency codes
   */
  private encodeCurrencyCodes(input: CurrencyCode[]): string {
    return encodeURI(input.map(it => it.rawValue).join(','));
  }

  /**
   * Encode a list of contract addresses the query string format expected by the endpoint
   *
   * @param input The list of contract addresses to encode
   *
   * @return The encoded contract addresses
   */
  private encodeContractAddresses(input: ContractAddress[]): string {
    return encodeURI(input.map(it => it.rawValue).join(','));
  }
}
