import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { ExchangeRate } from '@wallets/models/ExchangeRate';
import Decimal from 'decimal.js';

/**
 * Method to get BigDecimal exchange rate (or null if not found) value for provided token params
 *
 * @param currencyCode token currency code, like 'BTC'
 * @param contractAddress optional contract address for ERC20 tokens
 * @param name optional human readable name value, like 'Bitcoin'
 *
 * @return Optional BigDecimal value for exchange rate
 */
export function rateFor(
  thiz: ExchangeRate[],
  currencyCode: CurrencyCode,
  contractAddress?: string,
  name?: string
): Decimal | undefined {
  /**
   * Special handling for MATIC due to MATIC being ERC20 on mainnet and base currency for Polygon chain
   * there will be legit 2 entries in the DB, one with contract address - that will be correctly resolved by
   * 2nd check, and one without - resolved by special case check
   */
  if (currencyCode.code === 'MATIC' && contractAddress === undefined) {
    return thiz.find(it => it.currencyCode.code === 'MATIC' && it.contractAddress === undefined)
      ?.rate;
  }

  /**
   * First check is with currency code, which will return a unique value in 99% of the cases. Currency code has
   * priority, because it's more reliable from backend perspective, same token might return rate when requested by
   * currency code, but return nothing when requested based on contract address, so filtering by contract address
   * first might result in an empty array, when we do actually have the rate for the token based on the currency code
   */
  const matchingRates = thiz.filter(it => it.currencyCode.rawValue === currencyCode.rawValue);
  if (matchingRates.length === 1) return matchingRates[0].rate;

  /**
   * If currency code returns more than 1 value, attempt to determine correct token using contract address
   */
  if (contractAddress !== undefined) {
    return matchingRates.find(
      it => it.contractAddress?.toLowerCase() === contractAddress.toLowerCase()
    )?.rate;
  }

  /**
   * If currency code is not unique, and there's no contract address (can potentially happen with consumer flows)
   * as a last resort attempt to find correct rate based on human readable token name, which is available in
   * consumer account model
   */
  if (matchingRates.length > 0) {
    const unwrappedName = name ?? currencyCode.code;
    return matchingRates.find(it => it.name?.toLowerCase() === unwrappedName.toLowerCase())?.rate;
  } else {
    return undefined;
  }
}
