import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { Store } from '@store/Store';
import { rateFor } from '@wallets/extensions/List+Wallets';
import { ExchangeRate } from '@wallets/models/ExchangeRate';
import Decimal from 'decimal.js';
import { Observable, of } from 'rxjs';
import { StoreKeys_activeCurrency } from '@wallets/extensions/StoreKeys+Wallets';
import { FiatCurrency } from '@wallets/models/FiatCurrency';

Decimal.set({ precision: 100 });

export class CurrencyFormatter {
  static defaultMaxDecimalDisplay = 8n;

  private store: Store;
  private exchangeRateObservable: Observable<ExchangeRate[]>;
  private exchangeRates: ExchangeRate[] = [];

  constructor(store: Store, exchangeRateObservable: Observable<ExchangeRate[]>) {
    this.store = store;
    this.exchangeRateObservable = exchangeRateObservable;
  }

  /**
   * Format a [BigInteger] representation of a cryptocurrency amount to a fiat format using the
   * given [fromCurrencyCode]
   *
   * @param fromCurrencyCode Crypto currency code
   * @param contractAddress ERC20 token contract address
   * @param decimals Crypto currency decimals count
   * @param value Value to format
   * @param isAbbreviated If true, returns a shorthand representation of the fiat amount i.e. "21.5M" or "1.2B"
   * @param name Human readable name, like `Bitcoin`
   * @return
   */
  formatToFiat(
    fromCurrencyCode: CurrencyCode,
    decimals: bigint,
    value: bigint,
    isAbbreviated = false,
    contractAddress?: string,
    name?: string
  ): string | undefined {
    const fiatValue = this.fiatValue(fromCurrencyCode, decimals, value, contractAddress, name);
    if (fiatValue === undefined) return undefined;

    if (isAbbreviated) {
      const formatter = Intl.NumberFormat(undefined, {
        style: 'decimal',
        notation: 'compact',
        compactDisplay: 'short',
        maximumFractionDigits: 1,
      });
      return formatter.format(fiatValue.toNumber());
    } else {
      const formatter = Intl.NumberFormat(undefined, {
        style: 'currency',
        currency: this.getActiveCurrency(),
      });
      return this.formatDecimal(formatter, fiatValue);
    }
  }

  /**
   * Format [BigInteger] to a crypto format
   *
   * @param currencyCode Currency code
   * @param decimals Currency decimals count
   * @param value Value to format
   * @param includeCode Whether we should show the currency suffix i.e. "1,000 ETH" vs `1,000`
   * @param maxDisplayDecimals Max number of decimals displayed in the formatted string.
   *
   * @return A Single wrapping the formatted crypto
   */
  formatToCrypto(
    currencyCode: CurrencyCode,
    decimals: bigint,
    value: bigint | Decimal,
    includeCode = true,
    maxDisplayDecimals: bigint = CurrencyFormatter.defaultMaxDecimalDisplay,
    minDisplayDecimals = 0n
  ): string {
    const decimalValue = this.toDecimal(value, decimals);

    const isDustAmount = decimalValue.lessThan(new Decimal(10).pow(-Number(maxDisplayDecimals)));
    const numOfDecimalsToDisplay = !isDustAmount ? maxDisplayDecimals : decimals;
    const formatter = Intl.NumberFormat('en-US', {
      maximumFractionDigits: Number(numOfDecimalsToDisplay),
      minimumFractionDigits: Number(minDisplayDecimals),
      minimumIntegerDigits: 1,
    });
    const roundedValue = decimalValue.toDecimalPlaces(
      Number(numOfDecimalsToDisplay),
      Decimal.ROUND_FLOOR
    );
    const formattedValue = this.formatDecimal(formatter, roundedValue);

    return `${formattedValue} ${includeCode ? currencyCode.rawValue : ''}`.trim();
  }

  /**
   * Returns true if we can represent a given crypto in fiat
   *
   * @param currencyCode The currency code of the crypto
   * @param contractAddress ERC20 token contract address
   *
   * @return true if an exchange rate is available
   */
  hasFiatRepresentation(currencyCode: CurrencyCode, contractAddress?: string): boolean {
    return rateFor(this.exchangeRates, currencyCode, contractAddress) != null;
  }

  fiatValue(
    currencyCode: CurrencyCode,
    decimals: bigint,
    value: bigint,
    contractAddress?: string,
    name?: string
  ): Decimal | undefined {
    const exchangeRate = rateFor(this.exchangeRates, currencyCode, contractAddress, name);
    if (exchangeRate === undefined) return undefined;

    const decimalValue = this.toDecimal(value, decimals);
    return decimalValue.mul(exchangeRate);
  }

  start(): Observable<any> {
    this.exchangeRateObservable.subscribe(it => {
      this.exchangeRates = it;
    });

    return of();
  }

  private getActiveCurrency(): string {
    const fiatCurrency: FiatCurrency | undefined = this.store.get(StoreKeys_activeCurrency);
    return (fiatCurrency?.code ?? FiatCurrency.USD.code).rawValue;
  }

  private toDecimal(value: bigint | Decimal, decimals: bigint) {
    const decimalValue = typeof value === 'bigint' ? new Decimal(value.toString()) : value;
    const denom = new Decimal(10).pow(Number(decimals));
    return decimalValue.dividedBy(denom);
  }

  private formatDecimal(formatter: Intl.NumberFormat, value: Decimal): string {
    if (value.isInteger()) {
      return formatter.format(BigInt(value.toFixed()));
    }

    const formattedZero = formatter.format(0);
    const decimalSeparator = formattedZero.charAt(formattedZero.indexOf('0') + 1);

    const integer = value.truncated();
    let formattedInteger = formatter.format(BigInt(integer.toFixed()));
    const decimalIndex = formattedInteger.indexOf(decimalSeparator);
    if (decimalIndex > 0) {
      formattedInteger = formattedInteger.substring(0, decimalIndex);
    }

    const fraction = value.sub(integer);
    let formattedFraction = formatter.format(fraction.toNumber());
    formattedFraction = formattedFraction.substring(
      formattedFraction.indexOf(decimalSeparator) + 1
    );

    return formattedInteger + decimalSeparator + formattedFraction;
  }
}
