import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { LocalStorage } from '@store/storages/LocalStorage';
import { MemoryStorage } from '@store/storages/MemoryStorage';
import { Store } from '@store/Store';
import { StoreKeys_activeCurrency } from '@wallets/extensions/StoreKeys+Wallets';
import { ExchangeRate } from '@wallets/models/ExchangeRate';
import { FiatCurrency } from '@wallets/models/FiatCurrency';
import Decimal from 'decimal.js';
import { of } from 'rxjs';
import { CurrencyFormatter } from '../CurrencyFormatter';

describe('Currency Formatter', () => {
  const jpyExchangeRate = new Decimal(0.0088);
  const expectedZrxRate = new Decimal(3);
  const expectedEthRate = new Decimal(200);
  const expectedBatRate = new Decimal(12);

  const store: Store = (() => {
    const store = new Store(new LocalStorage(), new MemoryStorage());
    store.set(StoreKeys_activeCurrency, FiatCurrency.USD);
    return store;
  })();

  const exchangeRateObservable = of([
    new ExchangeRate(CurrencyCode.ETH, expectedEthRate),
    new ExchangeRate(new CurrencyCode('ZRX'), expectedZrxRate),
    new ExchangeRate(new CurrencyCode('BAT'), expectedBatRate),
    new ExchangeRate(CurrencyCode.USD, new Decimal(1)),
    new ExchangeRate(new CurrencyCode('JPY'), jpyExchangeRate),
  ]);

  let currencyFormatter: CurrencyFormatter;

  beforeAll(() => {
    currencyFormatter = new CurrencyFormatter(store, exchangeRateObservable);
    currencyFormatter.start().subscribe();
  });

  it('FormatToCrypto', () => {
    store.set(StoreKeys_activeCurrency, new FiatCurrency(new CurrencyCode('JPY'), 'Yen', 2n));
    const value = 100000000000023000000000000000000n;
    const expected = '100,000,000,000,023 ZRX';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('ZRX'), 18n, value, true, 10n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToCryptoSuperLargeNumber', () => {
    store.set(StoreKeys_activeCurrency, new FiatCurrency(new CurrencyCode('JPY'), 'Yen', 2n));
    const value = 12345678901234567890123456789012345678901234567890123456789012345678901234567800n;
    const expected =
      '12,345,678,901,234,567,890,123,456,789,012,345,678,901,234,567,890' +
      ',123,456,789,012.3456789012345678 ZRX';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('ZRX'), 18n, value, true, 20n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToCryptoSmallNumber', () => {
    store.set(StoreKeys_activeCurrency, new FiatCurrency(new CurrencyCode('JPY'), 'Yen', 2n));
    const value = 123000000000000000000n;
    const expected = '123 BAT';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('BAT'), 18n, value, true, 2n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToFiat', () => {
    store.set(StoreKeys_activeCurrency, FiatCurrency.USD);
    const value = 1234123456789012345678n;

    const groupingSeparator = ',';
    const decimalSeparator = '.';
    const symbol = '$';

    const expected =
      `${symbol}2${groupingSeparator}468${groupingSeparator}246` +
      `${groupingSeparator}913${groupingSeparator}578${groupingSeparator}024` +
      `${groupingSeparator}691${groupingSeparator}356${decimalSeparator}00`;

    const actual = currencyFormatter.formatToFiat(CurrencyCode.ETH, 2n, value);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToFiat2', () => {
    store.set(StoreKeys_activeCurrency, FiatCurrency.USD);
    const value = 1234123456789012345678n;

    const groupingSeparator = ',';
    const decimalSeparator = '.';
    const symbol = '$';

    const expected = `${symbol}246${groupingSeparator}824${decimalSeparator}69`;

    const actual = currencyFormatter.formatToFiat(CurrencyCode.ETH, 18n, value);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToFiatNoKnownExchangeRate', () => {
    store.set(StoreKeys_activeCurrency, FiatCurrency.USD);
    const value = 1234123456789012345678n;

    const actual = currencyFormatter.formatToFiat(new CurrencyCode('XYZ'), 18n, value);

    expect(actual).toBeUndefined();
  });

  it('SmallCryptoFormatting', () => {
    const value = 1n;
    const expected = '0.000000000000000001 ZRX';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('ZRX'), 18n, value, true, 4n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('SmallNonOneCryptoFormatting', () => {
    const value = 36n;
    const expected = '0.000000000000000036 ZRX';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('ZRX'), 18n, value, true, 4n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('SmallNonOne2CryptoFormatting', () => {
    const value = 36n;
    const expected = '0.000000003 ZRX';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('ZRX'), 10n, value, true, 9n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('DisplayDecimalLargerThanCoinFormatting', () => {
    const value = 36n;
    const expected = '0.36 ZRX';

    const actual = currencyFormatter.formatToCrypto(new CurrencyCode('ZRX'), 2n, value, true, 9n);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToAbbreviatedFiatSmallAmount', () => {
    const value = 1200n;
    const expected = '12';

    const actual = currencyFormatter.formatToFiat(CurrencyCode.USD, 2n, value, true);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToAbbreviatedFiatThousandAmount', () => {
    const value = 1234500n;
    const expected = '12.3K';

    const actual = currencyFormatter.formatToFiat(CurrencyCode.USD, 2n, value, true);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToAbbreviatedFiatMillionAmount', () => {
    const value = 1234567800n;
    const expected = '12.3M';

    const actual = currencyFormatter.formatToFiat(CurrencyCode.USD, 2n, value, true);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToAbbreviatedFiatBillionAmount', () => {
    const value = 1234567899900n;
    const expected = '12.3B';

    const actual = currencyFormatter.formatToFiat(CurrencyCode.USD, 2n, value, true);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });

  it('FormatToAbbreviatedFiatTrillionAmount', () => {
    const value = 1234567899999900n;
    const expected = '12.3T';

    const actual = currencyFormatter.formatToFiat(CurrencyCode.USD, 2n, value, true);

    console.log(`
    value:   ${value}
    expected:${expected}
    actual:  ${actual}
    `);
    expect(actual).toEqual(expected);
  });
});
