import {
  BehaviorSubject,
  combineLatest,
  Observable,
  of,
  ReplaySubject,
  Subscription,
  zip,
} from 'rxjs';
import {
  concatMap,
  distinctUntilChanged,
  filter,
  flatMap,
  map,
  single,
  throttleTime,
} from 'rxjs/operators';
import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { CompoundFinanceRepository } from '@compoundfinance/repositories/CompoundFinanceRepository';
import { ExchangeRateAPI } from '../apis/ExchangeRateAPI';
import { FiatExchangeRateDAO } from '../daos/FiatExchangeRateDAO';
import { ExchangeRate } from '../models/ExchangeRate';
import { FiatCurrency } from '../models/FiatCurrency';
import { IWalletRepository } from './WalletRepository';
import { ContractAddress } from '@ethereum/models/ContractAddress';
import { CompoundExchangeRateResult } from '@compoundfinance/models/CompoundExchangeRateResult';
import { Decimal } from 'decimal.js';
import { distinct, distinctBy, filterNullish } from '@core/extensions/Array';
import { ExchangeRateError } from '@wallets/errors/ExchangeRateError';

export class ExchangeRateRepository {
  private activeFiatCurrencyObservable: Observable<FiatCurrency>;
  private exchangeRateApi: ExchangeRateAPI;
  private compoundFinanceRepository: CompoundFinanceRepository;
  private exchangeRateDAO: FiatExchangeRateDAO;
  private walletRepository: IWalletRepository;

  private isReadySubject: ReplaySubject<boolean>;
  private exchangeRateSubject: BehaviorSubject<ExchangeRate[]>;
  private cryptoCurrencyCodesObservable: Observable<CurrencyCode[]>;

  constructor(
    activeFiatCurrencyObservable: Observable<FiatCurrency>,
    exchangeRateApi: ExchangeRateAPI,
    compoundFinanceRepository: CompoundFinanceRepository,
    exchangeRateDAO: FiatExchangeRateDAO,
    walletRepository: IWalletRepository
  ) {
    this.activeFiatCurrencyObservable = activeFiatCurrencyObservable;
    this.exchangeRateApi = exchangeRateApi;
    this.compoundFinanceRepository = compoundFinanceRepository;
    this.exchangeRateDAO = exchangeRateDAO;
    this.walletRepository = walletRepository;

    this.isReadySubject = new ReplaySubject<boolean>(1);
    this.exchangeRateSubject = new BehaviorSubject<ExchangeRate[]>([]);
    this.cryptoCurrencyCodesObservable = this.walletRepository
      .observeWallets(false)
      .pipe(map(wallets => wallets.map(it => it.currencyCode)));

    this.reloadExchangeRatesFromDatabase();
  }

  get exchangeRateObservable(): Observable<ExchangeRate[]> {
    return this.exchangeRateSubject.pipe(distinctUntilChanged());
  }

  refresh(isForced: boolean): Observable<any> {
    // fetch latest fiat exchange rates
    const fiatExchangeRatesSingles = this.activeFiatCurrencyObservable.pipe(
      single(),
      flatMap(it => {
        const currencyCodes = distinctBy([it.code, CurrencyCode.USD], it => it.rawValue);
        return this.exchangeRateApi.getFiatExchangeRates(currencyCodes);
      })
    );

    // fetch latest crypto exchange rates in fiat
    const cryptoExchangeRatesSingles = zip(
      this.walletRepository.observeWallets(false).pipe(single()),
      this.exchangeRateApi.getConsumerSpotPrices()
    ).pipe(
      flatMap(([wallets, consumerExchangeRates]) => {
        const addresses = filterNullish(wallets.map(it => it.contractAddress));
        const codes = [
          ...wallets.map(it => it.currencyCode),
          ...consumerExchangeRates.map(it => it.currencyCode),
        ];

        // get all underlying erc20 codes for the user's compound finance tokens
        const underlyingContractAddresses = filterNullish(
          this.compoundFinanceRepository
            .getTokensWithContractAddresses(addresses)
            .map(it => it.underlyingContractAddress)
        );

        const allContractAddresses = distinct([...addresses, ...underlyingContractAddresses]).map(
          it => new ContractAddress(it)
        );
        const getCryptoExchangeRates = this.exchangeRateApi.getCryptoExchangeRates(
          distinctBy(codes, it => it.rawValue),
          allContractAddresses
        );

        return zip(getCryptoExchangeRates, of(consumerExchangeRates));
      }),
      flatMap(([exchangeRates, consumerExchangeRates]) => {
        const mutableExchangeRates = distinctBy(exchangeRates, it => [
          it.currencyCode.rawValue,
          it.contractAddress?.toLowerCase(),
        ]);

        // we take consumer exchange rate over crypto exchange rate where possible
        // if there are 2 tokens with same currency code - because consumer doesn't have contract address
        // we do nothing, to avoid replacing the wrong one
        consumerExchangeRates.forEach(consumerRate => {
          const matchingExchangeRates = mutableExchangeRates.filter(
            exchangeRate => exchangeRate.currencyCode.rawValue == consumerRate.currencyCode.rawValue
          );
          if (matchingExchangeRates.length == 1) {
            const matchingRate = matchingExchangeRates[0];
            const index = mutableExchangeRates.indexOf(matchingRate);
            const newRate = matchingRate.copy();
            newRate.rate = consumerRate.rate;
            mutableExchangeRates[index] = newRate;
          }
        });

        return this.calculateExchangeRate(mutableExchangeRates).pipe(
          map(it => [...mutableExchangeRates, ...it])
        );
      })
    );

    return zip(cryptoExchangeRatesSingles, fiatExchangeRatesSingles).pipe(
      flatMap(it => this.exchangeRateDAO.saveExchangeRates([...it[0], ...it[1]])),
      map(() => this.reloadExchangeRatesFromDatabase())
    );
  }

  getFiatExchangeRate(fiatCurrencyCode: CurrencyCode): Observable<Decimal> {
    return this.exchangeRateApi.getFiatExchangeRates([fiatCurrencyCode]).pipe(
      map(items => {
        if (items.length > 0) {
          return items[0].rate;
        } else {
          throw ExchangeRateError.unableToGetExchangeRate;
        }
      })
    );
  }

  getCryptoExchangeRate(
    cryptoCurrencyCode: CurrencyCode,
    contractAddress?: string,
    chainId?: bigint
  ): Observable<Decimal> {
    const contractAddresses =
      contractAddress !== undefined ? [new ContractAddress(contractAddress)] : [];

    return this.exchangeRateApi
      .getCryptoExchangeRates([cryptoCurrencyCode], contractAddresses, chainId)
      .pipe(
        map((rates: ExchangeRate[]) => {
          const rateValue =
            contractAddress !== undefined
              ? rates.find(it => it.contractAddress === contractAddress)
              : rates.find(it => it.currencyCode.rawValue === cryptoCurrencyCode.rawValue);

          if (rateValue !== undefined) {
            return rateValue.rate;
          } else {
            throw ExchangeRateError.unableToGetExchangeRate;
          }
        })
      );
  }

  start(): Observable<any> {
    return combineLatest(
      this.cryptoCurrencyCodesObservable,
      this.activeFiatCurrencyObservable
    ).pipe(
      throttleTime(500),
      distinctUntilChanged((lhs, rhs) => lhs[0] == rhs[0] && lhs[1] == rhs[1]),
      concatMap(() => {
        this.isReadySubject.next(true);
        return this.refresh(false);
      })
    );
  }

  destroy(): Observable<boolean> {
    this.isReadySubject.next(false);
    this.exchangeRateSubject.next([]);

    return of(true);
  }

  private reloadExchangeRatesFromDatabase(): Subscription {
    return zip(
      this.exchangeRateDAO.getExchangeRates(),
      this.activeFiatCurrencyObservable.pipe(single()),
      this.isReadySubject.pipe(filter(it => it))
    )
      .pipe(
        map(([exchangeRates, activateFiatCurrency]) => {
          const quoteCurrency = exchangeRates.find(
            it => it.currencyCode.rawValue == activateFiatCurrency.code.rawValue
          );

          if (quoteCurrency === undefined) return;

          if (quoteCurrency.rate.equals(1)) {
            this.exchangeRateSubject.next(exchangeRates);
          } else {
            // go through each base currency and convert them against our quote currency
            // e.g if our current base currency is USD at the rate of 1 USD/USD
            // and our currency quote currency is CAD at the rate of 0.77 CAD/USD
            // map[USD] = 1/0.77 = 1.298 USD/CAD
            const correctedExchangeRates: ExchangeRate[] = [];
            exchangeRates.forEach(baseRate => {
              const correctedRate = baseRate.copy();
              correctedRate.rate = baseRate.rate.dividedBy(quoteCurrency.rate);
              correctedExchangeRates.push(correctedRate);
            });
            this.exchangeRateSubject.next(correctedExchangeRates);
          }
        })
      )
      .subscribe();
  }

  /**
   * Calculate the final cToken exchange rates based on the underlying ERC20 exchange rate
   *
   * @param erc20ExchangeRates List of underlying ERC20 exchange rates
   *
   * @return A single wrapping list of cToken calculated exchange rates
   */
  private calculateExchangeRate(erc20ExchangeRates: ExchangeRate[]): Observable<ExchangeRate[]> {
    return this.compoundFinanceRepository.supportedTokens().pipe(
      flatMap(supportedTokens =>
        this.compoundFinanceRepository.getExchangeRates(
          supportedTokens.map(it => it.contractAddress)
        )
      ),
      map((cTokenExchangeRates: CompoundExchangeRateResult[]) =>
        filterNullish(
          cTokenExchangeRates.map(rateResult => {
            const rate = rateResult.rate;
            const underlyingDecimals = rateResult.token.underlyingDecimals;
            let erc20BackingRate: ExchangeRate | undefined;
            if (rateResult.token.underlyingContractAddress === undefined) {
              erc20BackingRate = erc20ExchangeRates.find(
                exchangeRate =>
                  exchangeRate.currencyCode.rawValue == CurrencyCode.ETH.rawValue &&
                  exchangeRate.contractAddress == null
              );
            } else {
              erc20BackingRate = erc20ExchangeRates.find(
                exchangeRate =>
                  exchangeRate.contractAddress == rateResult.token.underlyingContractAddress
              );
            }

            if (erc20BackingRate === undefined) return undefined;

            const ten = new Decimal(10);
            const denom = new Decimal(100)
              .mul(ten.pow(underlyingDecimals.toString()))
              .mul(ten.pow(rateResult.token.decimals.toString()));
            const compoundExchangeRate = new Decimal(rate.toString())
              .mul(erc20BackingRate.rate)
              .dividedBy(denom);

            return new ExchangeRate(
              rateResult.token.code,
              compoundExchangeRate,
              rateResult.token.contractAddress,
              undefined,
              rateResult.token.name
            );
          })
        )
      )
    );
  }
}
