import { Observable, of } from 'rxjs';
import { CryptoCurrency } from '@blockchains/models/CryptoCurrency';
import { CryptoCurrencyDAO } from '@wallets/daos/CryptoCurrencyDAO';
import { CryptoCurrencyAPI } from '@wallets/apis/CryptoCurrencyAPI';
import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { map, catchError, flatMap } from 'rxjs/operators';
import { from } from 'rxjs';
import { Startable } from '@core/interfaces/Startable';
import { Wallet } from '@blockchains/models/Wallet';

export interface IWalletRepository {
  /**
   * Observe wallet updates
   *
   * @param onlyUsable Usable coins are coins that are fully synced and have a balance greater than zero
   *
   * @return An observable wrapping list of [Wallet]
   */
  observeWallets(onlyUsable: boolean): Observable<Wallet[]>;

  /**
   * Get list of crypto currencies ordered by market cap
   *
   * @param currencyCodes A [List] of currency codes
   *
   * @return A [List] of fetched [CryptoCurrency]
   */
  getCryptoCurrencies(currencyCodes?: CurrencyCode[]): Observable<CryptoCurrency[]>;
}

export class WalletRepository implements Startable, IWalletRepository {
  cryptoCurrencyDAO: CryptoCurrencyDAO;
  cryptoCurrencyAPI: CryptoCurrencyAPI;

  constructor(cryptoCurrencyDAO: CryptoCurrencyDAO, cryptoCurrencyAPI: CryptoCurrencyAPI) {
    this.cryptoCurrencyDAO = cryptoCurrencyDAO;
    this.cryptoCurrencyAPI = cryptoCurrencyAPI;
  }

  observeWallets(onlyUsable: boolean): Observable<Wallet[]> {
    // throw new Error('Method not implemented.');
    return of([]);
  }

  // Overriding Startable
  start(): Observable<any> {
    this.cryptoCurrencyAPI
      .getCryptoCurrencies(BigInt(1))
      .pipe(flatMap(currencies => this.cryptoCurrencyDAO.saveCryptoCurrencies(currencies)))
      .subscribe();

    return from([]);
  }

  getCryptoCurrencies(currencyCodes?: CurrencyCode[]): Observable<CryptoCurrency[]> {
    let getCurrencies: Observable<CryptoCurrency[]>;
    if (currencyCodes == null) {
      getCurrencies = this.cryptoCurrencyDAO.getCryptoCurrencies();
    } else {
      getCurrencies = this.cryptoCurrencyDAO.getSelectCryptoCurrencies(currencyCodes);
    }

    return getCurrencies.pipe(
      //.logError() - Analytics logging
      map(currencies => {
        if (currencies.length == 0) {
          //TODO add custom error type
          throw Error('Foo');
        } else {
          return currencies;
        }
      }),
      catchError((error, caught) => {
        if (currencyCodes == null) {
          return this.getUserOwnedCryptoCurrencies();
        } else {
          const distinctCurrencyCodes = currencyCodes
            .map(code => code.rawValue)
            .filter((value, index, self) => self.indexOf(value) === index);

          return this.getUserOwnedCryptoCurrencies().pipe(
            map(currencies => {
              return currencies.filter(
                value => distinctCurrencyCodes.indexOf(value.code.rawValue) >= 0
              );
            })
          );
        }
      })
    );
  }

  private getUserOwnedCryptoCurrencies(): Observable<CryptoCurrency[]> {
    // TODO fill this in
    return from([]);
  }
}
