import axios from 'axios';
import Dexie from 'dexie';
import MockAdapter from 'axios-mock-adapter';
import { WalletsDatabase } from '@wallets/db/WalletsDatabase';
import { CompoundFinanceRepository } from '@compoundfinance/repositories/CompoundFinanceRepository';
import { BehaviorSubject, Observable, of, ReplaySubject } from 'rxjs';
import { ExchangeRateRepository } from '../ExchangeRateRepository';
import { ExchangeRateAPI } from '@wallets/apis/ExchangeRateAPI';
import { HTTPService } from '@http/models/HTTPService';
import { FiatExchangeRateDAO } from '@wallets/daos/FiatExchangeRateDAO';
import { FiatCurrency } from '@wallets/models/FiatCurrency';
import { CompoundTokenDAO } from '@compoundfinance/daos/CompoundTokenDAO';
import { CompoundFinanceAPI } from '@compoundfinance/apis/CompoundFinanceAPI';
import { Wallet } from '@blockchains/models/Wallet';
import { WalletAddress } from '@blockchains/models/WalletAddress';
import { AddressType } from '@blockchains/models/AddressType';
import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { Blockchain } from '@blockchains/models/Blockchain';
import { Network } from '@blockchains/models/Network';
import { CompoundToken } from '@compoundfinance/models/CompoundToken';
import { CompoundExchangeRateResult } from '@compoundfinance/models/CompoundExchangeRateResult';
import { filter } from 'rxjs/operators';
import Decimal from 'decimal.js';

Dexie.dependencies.indexedDB = require('fake-indexeddb');
Dexie.dependencies.IDBKeyRange = require('fake-indexeddb/lib/FDBKeyRange');

const mockedCToken = new CompoundToken(
  'Compound Wrapped BTC',
  new CurrencyCode('cWBTC'),
  8n,
  '0xc11b1268c1a384e55c48c2391d8d480264a3a7f4',
  new CurrencyCode('WBTC'),
  8n,
  '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599'
);

const supportedTokens = (): Observable<CompoundToken[]> => of([mockedCToken]);

const getExchangeRates = (x: string[]): Observable<CompoundExchangeRateResult[]> =>
  of([new CompoundExchangeRateResult(mockedCToken, 100n)]);

const getTokensWithContractAddresses = (x: string[]): CompoundToken[] => [mockedCToken];

jest.mock('@compoundfinance/repositories/CompoundFinanceRepository', () => {
  return {
    CompoundFinanceRepository: jest.fn().mockImplementation(() => {
      return {
        supportedTokens: supportedTokens,
        getExchangeRates: getExchangeRates,
        getTokensWithContractAddresses: getTokensWithContractAddresses,
      };
    }),
  };
});

const walletsDatabase = new WalletsDatabase();

const mockedCompoundRepo = new CompoundFinanceRepository(
  new CompoundTokenDAO(walletsDatabase.compoundTokenTable!),
  new CompoundFinanceAPI('', '')
);

const mockedWalletRepo = {
  observeWallets: (x: boolean) =>
    of([
      new Wallet(
        'primaryAddress',
        [new WalletAddress(new AddressType('type'), 'address', 0n)],
        'displayName',
        new CurrencyCode('WBTC'),
        100n,
        100n,
        new Blockchain('blockchain'),
        Network.createTestnetNetwork(true)
      ),
    ]),
  getCryptoCurrencies: jest.fn(),
};

const mockedAxios = new MockAdapter(axios);

describe('Exchange rate repo', () => {
  it('should return exchange rates', () => {
    const expectedKRWRate = 0.001;
    const expectedUSDRateInKRW = 1000;
    const expectedWBTCPrice = 54321.98765;
    const expectedWBTCPriceInKRW = 54321987.65;

    mockedAxios.onGet('https://api.coinbase.com/v2/prices/USD/spot').reply(200, {
      data: [
        { base: 'WBTC', currency: 'USD', amount: expectedWBTCPrice.toString() },
        { base: 'ETH', currency: 'USD', amount: '2018.5' },
      ],
    });

    mockedAxios.onGet('https://api.wallet.coinbase.com/rpc/v2/getExchangeRates').reply(200, {
      result: { rates: { USD: '1', KRW: expectedKRWRate.toString() } },
    });

    mockedAxios.onGet('https://api.wallet.coinbase.com/rpc/v2/getAssetPrices').reply(200, {
      result: {
        rates: {
          WBTC: {
            price: expectedWBTCPrice,
            percentageChange1d: 3.2716801,
            percentageChange7d: -15.10298095,
            lastUpdated: 1624421467000,
            name: 'Wrapped Bitcoin',
            contractAddress: '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599',
            symbol: 'WBTC',
          },
        },
      },
    });

    const walletsDatabase = new WalletsDatabase();
    if (walletsDatabase.exchangeRateTable == null) {
      throw new Error('Undefined');
    }

    const dao = new FiatExchangeRateDAO(walletsDatabase.exchangeRateTable);

    const exchangeRateRepo = new ExchangeRateRepository(
      of(new FiatCurrency(new CurrencyCode('KRW'), 'Korean Won', 4n)),
      new ExchangeRateAPI(new HTTPService('https://api.wallet.coinbase.com')),
      mockedCompoundRepo,
      dao,
      mockedWalletRepo
    );

    const isTestFinishedSubject = new BehaviorSubject(false);

    exchangeRateRepo.start().subscribe(() => {
      exchangeRateRepo.exchangeRateObservable.pipe(filter(data => data.length > 0)).subscribe(
        data => {
          expect(data.length).toBeGreaterThan(0);

          const usd = data.find(it => it.currencyCodeStr === CurrencyCode.USD.rawValue);
          expect(usd).toBeDefined();
          expect(usd?.rate).toEqual(new Decimal(expectedUSDRateInKRW));

          const wbtc = data.find(it => it.currencyCodeStr === 'WBTC');
          expect(wbtc).toBeDefined();
          expect(wbtc?.rate).toEqual(new Decimal(expectedWBTCPriceInKRW));

          isTestFinishedSubject.next(true);
          isTestFinishedSubject.complete();
        },
        error => {
          console.error(error);
        }
      );
    });

    return isTestFinishedSubject.toPromise().then(finished => {
      expect(finished).toBeTruthy();
    });
  });
});
