import { HTTPService } from '@http/models/HTTPService';
import axios from 'axios';
import { CryptoCurrency } from '@blockchains/models/CryptoCurrency';
import { CurrencyCode } from '@blockchains/models/CurrencyCode';
import { Blockchain } from '@blockchains/models/Blockchain';
import { CryptoCurrencyAPI } from '@wallets/apis/CryptoCurrencyAPI';
import { CryptoCurrencyDAO } from '@wallets/daos/CryptoCurrencyDAO';
import { WalletsDatabase } from '@wallets/db/WalletsDatabase';
import { WalletRepository } from '../WalletRepository';
import Dexie from 'dexie';
Dexie.dependencies.indexedDB = require('fake-indexeddb');
Dexie.dependencies.IDBKeyRange = require('fake-indexeddb/lib/FDBKeyRange');

jest.mock('axios');
const mockedAxios = axios as jest.Mocked<typeof axios>;

const cryptoCurrencies = [
  {
    name: 'Bitcoin',
    code: 'BTC',
    imageUrl: 'https://1.png',
    blockchain: 'btc',
    decimals: 8,
    marketCap: 758901722568.4563,
    swappable: false,
  },
  {
    name: 'Ethereum',
    code: 'ETH',
    imageUrl: 'https://2.png',
    blockchain: 'eth',
    decimals: 18,
    marketCap: 300249833368.14075,
    swappable: false,
  },
  {
    name: 'WETH',
    code: 'WETH',
    imageUrl: 'https://3.png',
    blockchain: 'eth',
    decimals: 18,
    marketCap: 300249833368.14075,
    swappable: true,
    contractAddress: '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2',
  },
];

describe('Wallet repository', () => {
  it('Should return all crypto currencies', () => {
    let expectedCryptoCurrencies = [
      new CryptoCurrency(
        new CurrencyCode('BTC'),
        'Bitcoin',
        8n,
        new Blockchain('btc'),
        0n,
        'https://1.png'
      ),
      new CryptoCurrency(
        new CurrencyCode('ETH'),
        'Ethereum',
        18n,
        new Blockchain('eth'),
        1n,
        'https://2.png'
      ),
      new CryptoCurrency(
        new CurrencyCode('WETH'),
        'WETH',
        18n,
        new Blockchain('eth'),
        2n,
        'https://3.png'
      ),
    ];

    const walletsDatabase = new WalletsDatabase();
    if (walletsDatabase.cryptoCurrencyTable == null) {
      throw new Error('Undefined');
    }

    const walletRepository = new WalletRepository(
      new CryptoCurrencyDAO(walletsDatabase.cryptoCurrencyTable),
      new CryptoCurrencyAPI(new HTTPService('https://api.wallet.coinbase.com'))
    );

    mockedAxios.get.mockResolvedValue({ data: { result: { cryptoCurrencies } } });

    return walletRepository
      .start()
      .toPromise()
      .then(_ => {
        return walletRepository
          .getCryptoCurrencies()
          .toPromise()
          .then(data => {
            expect(data.length).toEqual(expectedCryptoCurrencies.length);
            data.forEach((res, index) => {
              let resStr = JSON.stringify(res, (key, value) => {
                return typeof value === 'bigint' ? value.toString() : value;
              });

              let actualStr = JSON.stringify(expectedCryptoCurrencies[index], (key, value) => {
                return typeof value === 'bigint' ? value.toString() : value;
              });

              expect(resStr).toEqual(actualStr);
            });
          });
      });
  });

  it('Should return selected crypto currencies', () => {
    let expectedCryptoCurrencies = [
      new CryptoCurrency(
        new CurrencyCode('ETH'),
        'Ethereum',
        18n,
        new Blockchain('eth'),
        1n,
        'https://2.png'
      ),
      new CryptoCurrency(
        new CurrencyCode('WETH'),
        'WETH',
        18n,
        new Blockchain('eth'),
        2n,
        'https://3.png'
      ),
    ];

    const walletsDatabase = new WalletsDatabase();
    if (walletsDatabase.cryptoCurrencyTable == null) {
      throw new Error('Undefined');
    }

    const walletRepository = new WalletRepository(
      new CryptoCurrencyDAO(walletsDatabase.cryptoCurrencyTable),
      new CryptoCurrencyAPI(new HTTPService('https://api.wallet.coinbase.com'))
    );

    mockedAxios.get.mockResolvedValue({ data: { result: { cryptoCurrencies } } });

    return walletRepository
      .start()
      .toPromise()
      .then(_ => {
        return walletRepository
          .getCryptoCurrencies([new CurrencyCode('ETH'), new CurrencyCode('WETH')])
          .toPromise()
          .then(data => {
            expect(data.length).toEqual(expectedCryptoCurrencies.length);
            data.forEach((res, index) => {
              let resStr = JSON.stringify(res, (key, value) => {
                return typeof value === 'bigint' ? value.toString() : value;
              });

              let actualStr = JSON.stringify(expectedCryptoCurrencies[index], (key, value) => {
                return typeof value === 'bigint' ? value.toString() : value;
              });

              expect(resStr).toEqual(actualStr);
            });
          });
      });
  });
});
