import { css } from 'linaria';
import * as React from 'react';
import walletLogo from '../../assets/logo-round-no-border.svg';
import { WalletRepository } from '@wallets/repositories/WalletRepository';
import { WalletsDatabase } from '@wallets/db/WalletsDatabase';
import { CryptoCurrencyDAO } from '@wallets/daos/CryptoCurrencyDAO';
import { CryptoCurrencyAPI } from '@wallets/apis/CryptoCurrencyAPI';
import { HTTPService } from '@http/models/HTTPService';
import { initDefaultErrorHandler } from '@core/extensions/Observable+Core';
import { CompoundFinanceRepository } from '@compoundfinance/repositories/CompoundFinanceRepository';
import { CompoundTokenDAO } from '@compoundfinance/daos/CompoundTokenDAO';
import { CompoundFinanceAPI } from '@compoundfinance/apis/CompoundFinanceAPI';
import { LocalStorage } from '@store/storages/LocalStorage';
import { StoreKeys_activeCurrency } from '@wallets/extensions/StoreKeys+Wallets';
import { FiatCurrency } from '@wallets/models/FiatCurrency';
import { ParentWalletLinkManager } from '../../../src/serviceworker/ParentWalletLinkManager';
import { AuthenticationRepository } from '@user/repositories/AuthenticationRepository';
import { Store } from '@store/Store';
import { MemoryStorage } from '@store/storages/MemoryStorage';
import { CipherCore } from 'cipher-core';
import { AuthenticationAPI } from '@user/apis/AuthenticationAPI';
import { RPC } from '@http/RPC';
import { StoreKeys_accessToken } from '@user/extensions/StoreKeys+User';
import { RPCResponse } from '@http/models/RPCResponse';
import { ExchangeRateRepository } from '@wallets/repositories/ExchangeRateRepository';
import { of } from 'rxjs';
import { ExchangeRateAPI } from '@wallets/apis/ExchangeRateAPI';
import { FiatExchangeRateDAO } from '@wallets/daos/FiatExchangeRateDAO';
import { CurrencyCode } from '@blockchains/models/CurrencyCode';

const contentWidth = 430;
const contentPadding = 32;

const parentContainer = css`
  margin: auto;
  padding-left: ${contentPadding}px;
  padding-right: ${contentPadding}px;
`;

const container = css`
  --width: ${contentWidth}px;
  --lmargin: calc((100vw / 3) - (var(--width) / 2));
  width: var(--width);
  margin-left: max(0px, var(--lmargin));

  @media (max-width: ${contentWidth + 2 * contentPadding + 2 * 8 /*body padding*/}px) {
    width: 100%;
  }
`;

const logo = css`
  width: 56px;
  height: 56px;
  margin-top: 60px;
`;

const titleText = css`
  /* Title 1 */
  font-size: 28px;
  color: #000000;
  font-weight: 500;

  margin-top: 16px;
  margin-bottom: 0;
`;

const subtitleText = css`
  /* Title 3 */
  font-size: 20px;
  line-height: 28px;
  font-weight: 500;

  /* Grey - dark */
  color: #587088;

  margin-top: 0;
`;

const descriptionText = css`
  /* Body */
  font-size: 16px;
  line-height: 24px;
  color: #000000;

  margin-top: 26px;
  margin-bottom: 48px;
`;

const backgroundContainer = css`
  width: 100vw;
  height: 100vh;
  overflow: hidden;
  position: fixed;
  top: 0;
  left: 0;
  z-index: -1;
`;

const circleBackground = css`
  --lpos: calc((100vw / 3) - ${contentWidth}px);
  height: max(200vh, 175vw);
  width: max(200vh, 175vw);
  top: 10vh;
  left: max(calc(${contentWidth}px / 2), var(--lpos));
  border-radius: 50%;
  filter: blur(20px);
  z-index: -1;
  display: block;
  position: absolute;

  /* Blue - Normal */
  background: #1652f0;
`;

export const debugButton = css`
  /* CTA/full/active */
  /* Coinbase Blue */
  background: #1652f0;
  border-radius: 16px;

  /* Default bold - 16 medium */
  font-style: normal;
  font-weight: 500;
  font-size: 16px;
  line-height: 23px;

  /* identical to box height */
  display: flex;
  align-items: center safe;
  justify-content: center;
  text-align: center;
  color: #ffffff;
  padding: 16px;

  user-select: none;
  cursor: pointer;
  transition: transform 0.3s;

  &:hover {
    transform: scale(1.05);
  }
`;

initDefaultErrorHandler();

/**
 * Init store prototype mappings
 */
LocalStorage.mappedPrototypes.set(StoreKeys_activeCurrency.name, FiatCurrency.prototype);

const walletsDatabase = new WalletsDatabase();
if (walletsDatabase.cryptoCurrencyTable == null) {
  throw new Error('Undefined');
}

if (walletsDatabase.compoundTokenTable == null) {
  throw new Error('Undefined');
}

if (walletsDatabase.exchangeRateTable == null) {
  throw new Error('Undefined');
}

const walletRepository = new WalletRepository(
  new CryptoCurrencyDAO(walletsDatabase.cryptoCurrencyTable),
  new CryptoCurrencyAPI(new HTTPService('https://api.wallet.coinbase.com'))
);

const compoundRepository = new CompoundFinanceRepository(
  new CompoundTokenDAO(walletsDatabase.compoundTokenTable),
  new CompoundFinanceAPI(`${process.env.CB_WALLET_INFURA_KEY}`, 'https://api.wallet.coinbase.com')
);

const exchangeRateRepo = new ExchangeRateRepository(
  of(FiatCurrency.USD),
  new ExchangeRateAPI(new HTTPService('https://api.wallet.coinbase.com')),
  compoundRepository,
  new FiatExchangeRateDAO(walletsDatabase.exchangeRateTable),
  walletRepository
);

const walletLinkProvider = new ParentWalletLinkManager().provider;
const store = new Store(new LocalStorage(), new MemoryStorage());
const cipherCore = new CipherCore({
  ethereum: {
    networks: {},
  },
});

const authRepository = new AuthenticationRepository(
  new AuthenticationAPI(new HTTPService('https://api.wallet.coinbase.com')),
  store,
  cipherCore,
  walletLinkProvider
);

RPC.setup(authRepository);

// TODO move this into its own app start service
walletRepository.start().subscribe();
compoundRepository.start().subscribe();
exchangeRateRepo.start().subscribe();

//@ts-ignore
window.walletRepository = walletRepository;

const showCryptoCurrencies = () => {
  walletRepository.getCryptoCurrencies().subscribe(result => {
    let str = '';
    result.forEach(res => (str += `${res.blockchain} ${res.code}\r\n`));
    alert(str);
  });
};

const showCompoundTokens = () => {
  compoundRepository.supportedTokens().subscribe(result => {
    let str = '';
    result.forEach(res => (str += `${res.name} ${res.code}\r\n`));
    alert(str);
  });
};

const signInWithWalletlink = () => {
  authRepository.signInWithWalletlink(walletLinkProvider.selectedAddress!!).subscribe();
};

const signInWithMnemonic = () => {
  authRepository
    .signIn('mansion turn clock casual share weird thought only parent equip abuse track')
    .subscribe();
};

const makeAuthenticatedApiCall = () => {
  const selectedAddress = walletLinkProvider.selectedAddress;
  if (!selectedAddress) {
    throw Error('Need to have connected walletlink to make authenticated api call');
  }

  if (!store.has(StoreKeys_accessToken)) {
    throw Error('Need to have signed in using walletlink to make authenticated api call');
  }

  RPC.get<RPCResponse<any>>(
    new HTTPService('https://api.wallet.coinbase.com'),
    '/rpc/v2/getERC20TokensOwnedByAddress',
    true,
    new Map([
      ['address', selectedAddress.toLocaleLowerCase('en-US')],
      ['chainId', '1'],
    ])
  ).subscribe(result => {
    let str = '';
    //@ts-ignore
    result.contractAddresses.forEach(contractAddress => (str += `${contractAddress}\r\n`));
    alert(str);
  });
};

const showExchangeRates = () => {
  exchangeRateRepo.refresh(true).subscribe();
  exchangeRateRepo.exchangeRateObservable.subscribe(result => {
    let str = '';
    result.forEach(res => (str += `${res.name} ${res.rate}\r\n`));
    alert(str);
  });
};

const showWBTCExchangeRate = () => {
  exchangeRateRepo.getCryptoExchangeRate(new CurrencyCode('WBTC')).subscribe(result => {
    alert(result.toString());
  });
};

const showKRWExchangeRate = () => {
  exchangeRateRepo.getFiatExchangeRate(new CurrencyCode('KRW')).subscribe(result => {
    alert(result.toString());
  });
};

export const DebugMenu = () => {
  return (
    <div className={parentContainer}>
      <div className={backgroundContainer}>
        <span className={circleBackground} />
      </div>
      <div className={container}>
        <img src={walletLogo} className={logo} />
        <h1 className={titleText}>Coinbase Wallet extension</h1>
        <h3 className={subtitleText}>Debug menu</h3>
        <p className={descriptionText}>
          Debug menu and unreleased features of the Coinbase wallet extension
        </p>
        <div
          className={debugButton}
          onClick={showCryptoCurrencies}
          data-testid="show-crypto-currencies-button"
        >
          Show Crypto Currencies
        </div>
        <p />
        <div
          className={debugButton}
          onClick={showCompoundTokens}
          data-testid="show-compound-tokens-button"
        >
          Show Compound Tokens
        </div>
        <p />
        <div
          className={debugButton}
          onClick={signInWithWalletlink}
          data-testid="sign-in-with-walletlink"
        >
          Signin With Walletlink
        </div>
        <p />
        <div
          className={debugButton}
          onClick={signInWithMnemonic}
          data-testid="sign-in-with-mnemonic"
        >
          Signin With Mnemonic
        </div>
        <p />
        <div
          className={debugButton}
          onClick={makeAuthenticatedApiCall}
          data-testid="make-authenticated-api-call"
        >
          Call Authenticated API (erc20)
        </div>
        <p />

        <div className={debugButton} onClick={showExchangeRates} data-testid="show-ex-rate-button">
          Show Exchange Rates
        </div>
        <p />

        <div
          className={debugButton}
          onClick={showWBTCExchangeRate}
          data-testid="show-wbtc-ex-rate-button"
        >
          Show wBTC Exchange Rate
        </div>
        <p />

        <div
          className={debugButton}
          onClick={showKRWExchangeRate}
          data-testid="show-krw-ex-rate-button"
        >
          Show KRW Exchange Rate
        </div>
      </div>
    </div>
  );
};
