// Wraps bugsnag notify (or any error reporting system) into a ReportError fn

import md5 from 'md5';

export type ReportedErrorInfo = {
  errorHash?: string;
};

export type ErrorData = {
  error: Error;
  errorInfo?: React.ErrorInfo;
  metadata?: Record<string, string | number | boolean | string[]>;
  context?: string;
  severity?: 'info' | 'warning' | 'error';
};

export type NotifyFunction = (errorData: ErrorData, errorHash: string) => void;
export type ReportErrorFn = (error: ErrorData) => ReportedErrorInfo;

const CONSOLE_LEVELS_PER_SEVERITY = {
  info: console.info,
  warning: console.warn,
  error: console.error,
};

export const CONSOLE_PREFIX = `%c Error Reporting:`;
export const CONSOLE_STYLE = `color:red;`;

export const consoleNotify: NotifyFunction = errorData => {
  const { severity } = errorData;
  const consoleLevel = CONSOLE_LEVELS_PER_SEVERITY[severity || 'error'] || console.error;

  consoleLevel(CONSOLE_PREFIX, CONSOLE_STYLE, errorData);
};

export const shouldReport = (error: Error | (Error & { reportError: false })) =>
  !('reportError' in error) || error.reportError !== false;

// Note: Don't export this. If you need an error hash, you'll get it by reporting the error.
const getErrorHash = (error: Error): string => {
  const hash: string = md5(error.message + error.stack);
  return hash.substring(0, 8);
};

export const createErrorReporter = (notify: NotifyFunction): ReportErrorFn => {
  return (passedErrorData: ErrorData): ReportedErrorInfo => {
    // Guard against errorData being the wrong type (catch statements have an "any")
    let errorData: ErrorData;
    if (passedErrorData instanceof Error) {
      errorData = { error: passedErrorData };
      // eslint-disable-next-line no-console
      console.warn(
        CONSOLE_PREFIX,
        CONSOLE_STYLE,
        'reportError called incorrectly (reportError(error) rather than reportError({error}))'
      );
    } else if (typeof passedErrorData === 'object') {
      errorData = passedErrorData;
      if (!(errorData.error instanceof Error)) {
        // eslint-disable-next-line no-console
        console.warn(
          CONSOLE_PREFIX,
          CONSOLE_STYLE,
          'reportError called incorrectly with something other than an error object',
          passedErrorData
        );

        errorData.metadata = {
          ...errorData.metadata,
          originalErrorData: JSON.stringify(errorData.error),
        };
        errorData.error = new Error('Unspecified error.');
      }
    } else {
      errorData = {
        error: new Error(`Unspecified error.`),
        metadata: {
          originalErrorData: JSON.stringify(passedErrorData),
        },
      };
      // eslint-disable-next-line no-console
      console.warn(CONSOLE_PREFIX, CONSOLE_STYLE, 'reportError called incorrectly');
    }

    const { error } = errorData;

    const errorHash = getErrorHash(error);

    if (shouldReport(error)) notify(errorData, errorHash);

    return { errorHash };
  };
};

let errorHandler = createErrorReporter(consoleNotify);
export const setErrorHandler = (newErrorHandler: ReportErrorFn) => {
  if (process.env.NODE_ENV === 'production') {
    errorHandler = newErrorHandler;
  } else {
    // eslint-disable-next-line no-console
    console.log(
      CONSOLE_PREFIX,
      CONSOLE_STYLE,
      "Custom error reporting handler registered, but ignored because this isn't production"
    );
  }
};

// Note: overwrite this on any platform when you configure bugsnag.
export const reportError: ReportErrorFn = args => {
  try {
    return errorHandler(args);
  } catch (error) {
    // eslint-disable-next-line no-console
    console.error(CONSOLE_PREFIX, CONSOLE_STYLE, 'Error reporting error', error);
    return {};
  }
};
