import { WalletLink, WalletLinkProvider } from 'walletlink';
import { WalletLinkExtensionUI } from './WalletLinkExtensionUI';
import {
  WalletLinkExtensionConnectionRequest,
  WalletLinkExtensionConnectionResponse,
} from '../contentscript/WalletLinkExtensionConnectionRequest';
import { BehaviorSubject, Observable, timer } from 'rxjs';
import { ScopedLocalStorage } from 'walletlink/build/npm/dist/lib/ScopedLocalStorage';
import { Session } from 'walletlink/build/npm/dist/relay/Session';
import { Pubsub, RequestEthereumAccounts, Connected } from '../util/Pubsub';
import { sendEvent } from '../init/analytics';
import { EVENTS } from '../init/events';
import { WALLET_USER_NAME_KEY } from 'walletlink/build/npm/dist/relay/WalletLinkRelayAbstract';
import log from '../util/log';
import { SelectedChain } from '../datalayer/ethereum/models/EthereumChain';
import { ServiceWorkerStorageKeys } from './ServiceWorkerStorageKeys';
import { WalletLinkProviderOptions } from 'walletlink/build/npm/dist/provider/WalletLinkProvider';

export class ParentWalletLinkManager {
  provider: WalletLinkProvider;
  private _ui: WalletLinkExtensionUI;
  private _storage: ScopedLocalStorage;
  private _childRequests = new Array<MessageEvent>();
  private readonly walletLinkURL = 'https://www.walletlink.org';

  constructor() {
    log('Constructing parent walletlink manager folks');

    this._storage = new ScopedLocalStorage(`-walletlink:${this.walletLinkURL}`);

    this._ui = this.makeUI(this.walletLinkURL);
    this.provider = this.makeProvider();

    chrome.runtime.onMessage.addListener(event => {
      if (
        event.type !== 'extensionUIRequest' ||
        typeof event.data === 'undefined' ||
        typeof event.data.dappInfo === 'undefined' ||
        typeof event.data.dappInfo.dappURL === 'undefined'
      ) {
        return;
      }

      this._childRequests.push(event);
    });
  }

  handleEvent(type: string, action: string, id: string) {
    log('In popup got event' + event);
    log(event);
    if (type !== 'extensionUIRequest') {
      log('Unknown event type in app ' + type);
      return;
    }

    switch (action) {
      case WalletLinkExtensionConnectionRequest.requestEthereumAccounts:
        chrome.runtime.sendMessage({ type: 'popupUIRequest', data: { id: id } }, response => {
          this.requestEthereumAccounts(response);
        });

        log('After requestEthereumAccounts in popup');
        break;
    }
  }

  get walletUsername(): string | null {
    return this._storage.getItem(WALLET_USER_NAME_KEY);
  }

  get selectedAddress(): string | undefined {
    return this.provider.selectedAddress;
  }

  public requestEthereumAccounts(event: any | null) {
    log('Got requestEthereumAccounts request in popup');

    this.provider.request!({ method: 'eth_requestAccounts' })
      .then(() => {
        //Let the service worker know that parent is currently connected
        chrome.storage.local.set({ [ServiceWorkerStorageKeys.parentConnected]: true });

        if (!event || typeof event.data.childSession.id === 'undefined') {
          log('No child session specified, skipping');
          Pubsub.next(new Connected());
          return;
        }
        sendEvent(EVENTS.CONNECTED);

        Pubsub.next(
          new RequestEthereumAccounts(
            event.data.id,
            event.data.dappInfo.dappName,
            event.data.dappInfo.dappLogoURL,
            event.data.dappInfo.dappLogoURLAlt,
            event.data.dappInfo.dappURL,
            event.data.childSession.id,
            event.data.childSession.secret
          )
        );
      })
      .catch(error => {
        //TODO - handle user canceling
        log('Got error requesting ethereum accounts' + error);
      });
  }

  public connectChild(opts: RequestEthereumAccounts): Promise<void> {
    return this.provider
      .childRequestEthereumAccounts(
        opts.sessionId,
        opts.sessionSecret,
        opts.dappName,
        opts.dappLogoURL,
        opts.dappURL
      )
      .then((response: unknown) => {
        log('Got connect child response ' + response);

        try {
          let dappHost = new URL(opts.dappURL).host;
          sendEvent(EVENTS.AUTHORIZED, {
            dapp: dappHost,
          });
        } catch (e) {
          log('Had error parsing url');
          sendEvent(EVENTS.AUTHORIZED);
        }

        chrome.runtime.sendMessage({
          type: 'extensionUIResponse',
          data: {
            id: opts.eventId,
            action: WalletLinkExtensionConnectionResponse.requestEthereumAccountsResponse,
            addresses: [this.provider.selectedAddress],
          },
        });
      });
  }

  public hasNextChildRequest(): boolean {
    return this._childRequests.length > 0;
  }
  public handleNextChildRequest() {
    const nextChildRequest = this._childRequests.pop();
    if (typeof nextChildRequest === 'undefined') {
      return;
    }

    Pubsub.next(
      new RequestEthereumAccounts(
        nextChildRequest.data.id,
        nextChildRequest.data.dappInfo.dappName,
        nextChildRequest.data.dappInfo.dappLogoURL,
        nextChildRequest.data.dappInfo.dappLogoURLAlt,
        nextChildRequest.data.dappInfo.dappURL,
        nextChildRequest.data.childSession.id,
        nextChildRequest.data.childSession.secret
      )
    );

    return;
  }

  public cancel(opts: RequestEthereumAccounts): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      chrome.runtime.sendMessage(
        {
          type: 'extensionUIResponse',
          data: {
            id: opts.eventId,
            action: WalletLinkExtensionConnectionResponse.requestEthereumAccountsCancel,
          },
        },
        () => resolve()
      );
    });
  }

  public disconnect() {
    this.provider.close();
  }

  public resetAndRequestEthereumAccounts() {
    this._ui = this.makeUI(this.walletLinkURL);
    this.provider = this.makeProvider();

    this.requestEthereumAccounts(null);
  }

  private makeUI(walletLinkURL: string): WalletLinkExtensionUI {
    const connected$: Observable<boolean> = new BehaviorSubject<boolean>(false).asObservable();

    return new WalletLinkExtensionUI({
      walletLinkUrl: walletLinkURL,
      version: WalletLink.VERSION,
      darkMode: false,
      session: Session.load(this._storage) || new Session(this._storage).save(),
      //@ts-ignore
      connected$: connected$,
    });
  }

  private makeProvider(): WalletLinkProvider {
    return new WalletLink({
      appName: 'Wallet extension',
      appLogoUrl: 'https://avatars.githubusercontent.com/u/18060234?s=100&v=4',
      walletLinkUIConstructor: options => this._ui,
    }).makeWeb3Provider(SelectedChain.nodeUrl, Number(SelectedChain.chainId));
  }
}
