import { Session } from 'walletlink/build/npm/dist/relay/Session';
import {
  WalletLinkExtensionConnectionRequest,
  WalletLinkExtensionConnectionResponse,
} from '../contentscript/WalletLinkExtensionConnectionRequest';
import log from '../util/log';

class WalletLinkServiceWorker {
  private callbackMap = new Map<string, number>();
  private popupUIResponseMap = new Map<string, any>();
  private _currPopupTab: number | null = null;
  private _currPopupWin: number | null = null;

  start() {
    if (typeof localStorage !== 'undefined') {
      /**
       * This service worker is running in a node environment.  This tells us that it was started
       * as part of starting an extension popup. The documentation on the chrome v3 architecture
       * is still pretty sparse but it looks like multiple service workers are started for your extension
       * 1. 1 for all the websites you visit to share
       * 2. 1 per chrome extension popup
       *
       * Since  our architecture relies on a single service worker, this is a hack to make sure we don't
       * handle requests multiple times.
       */
      return;
    }

    chrome.runtime.onMessage.addListener((message, sender, sendResponse) => {
      switch (message.type) {
        //From content script -> popup
        case 'extensionUIRequest':
          sendResponse();
          if (
            typeof sender.tab === 'undefined' ||
            typeof sender.tab.id === 'undefined' ||
            typeof sender.tab.url === 'undefined' ||
            typeof sender.tab.title === 'undefined'
          ) {
            log('Skipping message, sender tab not specified');
            return false;
          }

          let dappLogoURL = message.data.dappInfo.dappLogoURL;
          let dappLogoURLAlt = sender.tab.favIconUrl;
          return this.handleExtensionUIRequestEvent(
            message,
            sender.tab?.id,
            sender.tab.title,
            dappLogoURL,
            dappLogoURLAlt,
            sender.tab.url,
            message.data.hasMultipleProviders == true
          );
        // From popup -> content script
        case 'extensionUIResponse':
          sendResponse();
          return this.handlePopupResponseEvent(message);
        // From popup -> service worker
        case 'popupUIRequest':
          sendResponse(this.popupUIResponseMap.get(message.data.id));
          return true;
        default:
          log('Unknown message ' + message.type);
          sendResponse();
          return false;
      }
    });

    chrome.runtime.onConnect.addListener(port => {
      log('connected ', port);

      if (port.name === 'extensionUIRequest') {
        port.onMessage.addListener(log);
      }
    });

    chrome.runtime.onSuspend.addListener(() => {
      log('SUSPEND EVENT');
      // sendEvent(EVENTS.SUSPEND);
    });

    chrome.runtime.onInstalled.addListener(installType => {
      log('INSTALL EVENT');

      if (installType.reason !== 'install') {
        return;
      }

      const requestURL = `home.html`;
      chrome.tabs.create({
        url: chrome.runtime.getURL(requestURL),
      });
    });

    this.listenForWindowsClosingAndCancelEthereumRequest();
  }

  private handlePopupResponseEvent(event: any): boolean {
    log(event);

    if (typeof event.data.id === 'undefined') {
      log('Undefined id in service worker response handler');
      return false;
    }

    const originalTabId = this.callbackMap.get(event.data.id);

    if (typeof originalTabId === 'undefined') {
      log("Couldn't find original tab to send response to");
      return false;
    }

    this.callbackMap.delete(event.data.id);

    chrome.tabs.sendMessage(originalTabId, event);

    let eventsToDelete = new Array<string>();
    this.callbackMap.forEach((tabId, eventId, _) => {
      if (tabId === originalTabId) {
        eventsToDelete.push(eventId);
        //Also send back the same event for this eventId

        event.data.id = eventId;
        log('Also sending event ' + event + ' to tab ' + originalTabId);
        chrome.tabs.sendMessage(originalTabId, event);
      }
    });

    eventsToDelete.forEach(eventId => {
      this.callbackMap.delete(eventId);
    });

    return true;
  }

  private handleExtensionUIRequestEvent(
    event: any,
    tabId: number,
    dappName: string,
    dappLogoURL: string | undefined,
    dappLogoURLAlt: string | undefined,
    dappURL: string,
    hasMultipleProviders: boolean
  ): boolean {
    if (!event.data.id) {
      log('Undefined id in service worker request handler');
      return false;
    }

    if (dappURL.startsWith('chrome-extension://')) {
      throw Error(
        "The service worker is handling an extension ui request from the extension itself, shouldn't happen"
      );
    }

    let existingTabId: number | undefined = Array.from(this.callbackMap.values()).find(value => {
      return value === tabId;
    });

    this.callbackMap.set(event.data.id, tabId);

    if (typeof event.data.action === 'undefined') {
      log('Undefined action in service worker script');
      return false;
    }

    if (typeof existingTabId !== 'undefined') {
      //Already displaying a connection request for the current tab
      //Bring the window or tab to the foreground
      log('skipping starting UI since tab already exists');

      if (this._currPopupWin != null) {
        chrome.windows.update(this._currPopupWin, { focused: true });
      }

      if (this._currPopupTab != null) {
        chrome.tabs.update(this._currPopupTab, { highlighted: true });
      }

      return true;
    }

    switch (event.data.action) {
      case WalletLinkExtensionConnectionRequest.requestEthereumAccounts:
        this.requestEthereumAccounts(
          tabId,
          event.data.id,
          event.data.childSession,
          dappName,
          dappLogoURL,
          dappLogoURLAlt,
          dappURL,
          hasMultipleProviders
        );
        return true;
      default:
        log(`Got unknown action type ${event.data.action}`);
        return false;
    }
  }

  private requestEthereumAccounts(
    tabId: number,
    eventId: string,
    childSession: Session,
    dappName: string,
    dappLogoURL: string | undefined,
    dappLogoURLAlt: string | undefined,
    dappURL: string,
    hasMultipleProviders: boolean
  ) {
    chrome.windows.getCurrent(currentWin => {
      let leftPosition: number;
      let topPosition: number;
      if (
        typeof currentWin !== 'undefined' &&
        typeof currentWin.width !== 'undefined' &&
        typeof currentWin.left !== 'undefined' &&
        typeof currentWin.top !== 'undefined'
      ) {
        //Put the left corner of the popup aligned with the right of the window if there are multiple web3 providers
        //otherwise, align the right corner of the popup with the window
        let leftAdjustment = hasMultipleProviders ? 0 : -296;
        leftPosition = currentWin.width + currentWin.left + leftAdjustment;
        topPosition = currentWin.top;
      } else {
        leftPosition = 0;
        topPosition = 0;
      }

      const isFullScreen = currentWin.state === 'fullscreen';

      log('Checking if full screen ' + isFullScreen);

      const response = {
        type: 'extensionUIRequest',
        data: {
          action: WalletLinkExtensionConnectionRequest.requestEthereumAccounts,
          id: eventId,
          childSession: childSession,
          dappInfo: {
            dappName: dappName,
            dappLogoURL: dappLogoURL,
            dappLogoURLAlt: dappLogoURLAlt,
            dappURL: dappURL,
          },
          originTab: tabId,
        },
      };

      if (this._currPopupTab != null || this._currPopupWin != null) {
        chrome.runtime.sendMessage(response);

        if (this._currPopupWin != null) {
          chrome.windows.update(this._currPopupWin, { focused: true });
        }

        if (this._currPopupTab != null) {
          chrome.tabs.update(this._currPopupTab, { highlighted: true });
        }

        return;
      }

      this.popupUIResponseMap.set(eventId, response);

      const requestURL = `index.html?inPageRequest=true&type=extensionUIRequest&action=${WalletLinkExtensionConnectionRequest.requestEthereumAccounts}&id=${eventId}`;
      if (isFullScreen) {
        const fullScreenRequestURL = requestURL + `&fullScreen=true`;
        chrome.tabs.create(
          {
            url: chrome.runtime.getURL(fullScreenRequestURL),
          },
          tab => {
            if (typeof tab.id !== 'undefined') {
              this._currPopupTab = tab.id;
            }
          }
        );
      } else {
        chrome.windows.create(
          {
            url: chrome.runtime.getURL(requestURL),
            type: 'popup',
            left: leftPosition,
            top: topPosition,
            height: 550, //UI height from figma + toolbar height
            width: 296,
          },
          win => {
            if (typeof win?.id !== 'undefined') {
              this._currPopupWin = win.id;
            }
          }
        );
      }
    });
  }

  private listenForWindowsClosingAndCancelEthereumRequest() {
    chrome.tabs.onRemoved.addListener(tabId => {
      if (tabId !== this._currPopupTab) {
        return;
      }

      this._currPopupTab = null;

      this.callbackMap.forEach((originalTabId, eventId, _) => {
        chrome.tabs.sendMessage(originalTabId, {
          type: 'extensionUIResponse',
          data: {
            id: eventId,
            action: WalletLinkExtensionConnectionResponse.requestEthereumAccountsCancel,
          },
        });
      });

      this.callbackMap.clear();
    });

    chrome.windows.onRemoved.addListener(winId => {
      if (winId !== this._currPopupWin) {
        return;
      }

      this._currPopupWin = null;

      this.callbackMap.forEach((originalTabId, eventId, _) => {
        chrome.tabs.sendMessage(originalTabId, {
          type: 'extensionUIResponse',
          data: {
            id: eventId,
            action: WalletLinkExtensionConnectionResponse.requestEthereumAccountsCancel,
          },
        });
      });

      this.callbackMap.clear();
    });
  }
}

new WalletLinkServiceWorker().start();
