import { createMemoryHistory, MemoryHistory, Location } from 'history';
import log from './log';

interface NavigationHistoryEntity {
  entries: Location[];
  index: number;
}

class NavigationHistory {
  static readonly NAVIGATION_HISTORY_KEY = 'NAVIGATION_HISTORY';
  static readonly MAX_HISTORY_ENTRIES = 100;
  private history: MemoryHistory;

  constructor() {
    this.history = createMemoryHistory();
  }

  init(): Promise<NavigationHistory> {
    return this.fetchHistoryFromStorage().then(result => {
      if (!result || !result.entries) {
        return this.setHistoryInStorage(this.history).then(() => {
          log('NavigationHistory: created new navigation history');
          return this;
        });
      }

      log('NavigationHistory: fetched history from storage');
      const paths = this.getPathsFromEntries(result.entries);
      let trimmedPaths;
      let adjustedIndex = result.index;

      // If History gets too big, we cut it in half to save on memory in chrome.storage.local
      if (result.index > NavigationHistory.MAX_HISTORY_ENTRIES - 1) {
        trimmedPaths = paths.slice(Math.floor(result.index / 2), result.index + 1);
        adjustedIndex = trimmedPaths.length - 1;
      }

      this.history = createMemoryHistory({
        initialEntries: trimmedPaths || paths,
        initialIndex: adjustedIndex,
      });

      return this;
    });
  }

  getPathsFromEntries(entries?: Location[]): string[] {
    if (!entries || !entries.length) {
      // Router requires at least one default entry when History is manually passed in
      return ['/'];
    }

    return entries.map(entry => {
      return entry.pathname;
    });
  }

  getHistory(): MemoryHistory {
    return this.history;
  }

  fetchHistoryFromStorage(): Promise<NavigationHistoryEntity> {
    return new Promise((resolve, reject) => {
      chrome.storage.local.get([NavigationHistory.NAVIGATION_HISTORY_KEY], history => {
        const { [NavigationHistory.NAVIGATION_HISTORY_KEY]: navHistory = {} } = history;
        const { entries, index } = navHistory;

        if (chrome.runtime.lastError) {
          return reject('NavigationHistory: Error fetching History from storage');
        }

        resolve({
          entries,
          index,
        });
      });
    });
  }

  setHistoryInStorage(history: MemoryHistory): Promise<NavigationHistoryEntity> {
    const { entries, index } = history;

    return new Promise((resolve, reject) => {
      const input = { [NavigationHistory.NAVIGATION_HISTORY_KEY]: { entries, index } };

      chrome.storage.local.set(input, () => {
        if (chrome.runtime.lastError) {
          return reject('NavigationHistory: Error setting History in storage');
        }

        resolve({ entries, index });
      });
    });
  }
}

export default NavigationHistory;
