import { chrome } from 'jest-chrome';
import NavigationHistory from '../NavigationHistory';

describe('Navigation History', () => {
  // Mock chrome.storage.local as a Map
  const store = new Map();

  const storeGet = chrome.storage.local.get.mockImplementation((key, callback) => {
    // assume 1 key for this simple implementation
    // chrome storage seems to return empty object {} on not found
    if (Array.isArray(key)) {
      callback(store.get(key[0]) || {});
      return;
    }

    callback(store.get(key) || {});
  });

  const storeSet = chrome.storage.local.set.mockImplementation((object, callback) => {
    const keys = Object.keys(object);
    const values = Object.values(object);
    // assume 1 key/value for this simple implementation
    const key = keys[0];
    const value = {
      [key]: values[0],
    };

    store.set(key, value);

    if (callback) {
      callback();
    }
  });

  const mockEntries = [{
    pathname: '/',
    search: '',
    hash: '',
    state: undefined,
    key: 'aive3o'
  }, {
    pathname: '/path1',
    search: '',
    hash: '',
    state: undefined,
    key: '434ttg'
  }, {
    pathname: '/path2',
    search: '',
    hash: '',
    state: undefined,
    key: '15gght'
  }, {
    pathname: '/path3',
    search: '',
    hash: '',
    state: undefined,
    key: 'vzrg4h'
  }, {
    pathname: '/path4',
    search: '',
    hash: '',
    state: undefined,
    key: '54657g'
  }, {
    pathname: '/path5',
    search: '',
    hash: '',
    state: undefined,
    key: '65373s'
  }, {
    pathname: '/path6',
    search: '',
    hash: '',
    state: undefined,
    key: 'grthgr'
  }, {
    pathname: '/path7',
    search: '',
    hash: '',
    state: undefined,
    key: '26fafg'
  }, {
    pathname: '/path8',
    search: '',
    hash: '',
    state: undefined,
    key: '6357hj'
  }, {
    pathname: '/path9',
    search: '',
    hash: '',
    state: undefined,
    key: '256afg'
  }];

  beforeEach(() => {
    store.clear();
    jest.clearAllMocks();
  });

  it('Should initialize Navigation History from empty store', () => {
    const navigationHistory = new NavigationHistory();

    return navigationHistory.init().then(() => {
      expect(storeGet).toBeCalled();
      expect(storeSet).toBeCalled();

      const navHistory = navigationHistory.getHistory();
      const storeValues = store.get(NavigationHistory.NAVIGATION_HISTORY_KEY);

      expect(navHistory.entries[0].pathname).toEqual('/');
      expect(navHistory.index).toEqual(0);
      expect(storeValues[NavigationHistory.NAVIGATION_HISTORY_KEY].entries[0].pathname).toEqual('/');
      expect(storeValues[NavigationHistory.NAVIGATION_HISTORY_KEY].index).toEqual(0);
    });
  });

  it('Should initialize Navigation History from populated store', () => {
    const mockHistory = {
      [NavigationHistory.NAVIGATION_HISTORY_KEY]: {
        entries: mockEntries,
        index: 5,
      },
    };

    store.set(NavigationHistory.NAVIGATION_HISTORY_KEY, mockHistory);

    const navigationHistory = new NavigationHistory();

    return navigationHistory.init().then(() => {
      expect(storeGet).toBeCalled();
      expect(storeSet).not.toBeCalled();

      const navHistory = navigationHistory.getHistory();
      const storeValues = store.get(NavigationHistory.NAVIGATION_HISTORY_KEY);

      expect(navHistory.entries[4].pathname).toEqual('/path4');
      expect(navHistory.index).toEqual(5);
      expect(storeValues[NavigationHistory.NAVIGATION_HISTORY_KEY].entries[4].pathname).toEqual('/path4');
      expect(storeValues[NavigationHistory.NAVIGATION_HISTORY_KEY].index).toEqual(5);
    });
  });

  it('Should truncate entries when exceeding max', () => {
    const originalMaxHistory = NavigationHistory.MAX_HISTORY_ENTRIES;
    Object.defineProperty(NavigationHistory, 'MAX_HISTORY_ENTRIES', { value: 6 });

    const mockHistory = {
      [NavigationHistory.NAVIGATION_HISTORY_KEY]: {
        entries: mockEntries,
        index: 8,
      },
    };

    const originalPath = mockEntries[mockHistory[NavigationHistory.NAVIGATION_HISTORY_KEY].index].pathname;

    store.set(NavigationHistory.NAVIGATION_HISTORY_KEY, mockHistory);

    const navigationHistory = new NavigationHistory();

    return navigationHistory.init().then(() => {
      expect(storeGet).toBeCalled();
      expect(storeSet).not.toBeCalled();

      const navHistory = navigationHistory.getHistory();

      expect(originalPath).toEqual(navHistory.entries[navHistory.index].pathname);
      expect(navHistory.entries[0].pathname).toEqual('/path4');
      expect(navHistory.index).toEqual(4);
      expect(navHistory.length).toEqual(5);

      Object.defineProperty(NavigationHistory, 'MAX_HISTORY_ENTRIES', { value: originalMaxHistory });
    });
  });
});
