// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package main

import (
	"github.com/DataDog/datadog-go/statsd"
	"github.com/walletlink/walletlink/config"
	"github.com/walletlink/walletlink/internal/sweeper"
	"github.com/walletlink/walletlink/store"
	"github.com/walletlink/walletlink/util"
	"go.uber.org/zap"
	"gopkg.in/DataDog/dd-trace-go.v1/ddtrace/tracer"
)

// Sweeper archives old messages that are seen by subscribers because the messages never be reused
// This is to keep `store` table small to stabilize CPU usage and memory usage
func main() {
	var (
		logger       *zap.Logger
		statsdClient statsd.ClientInterface

		err error
	)

	if config.AppEnv == config.AppEnvDevelopment || config.AppEnv == config.AppEnvTest {
		statsdClient = &statsd.NoOpClient{}
		logger, _ = zap.NewDevelopment()
	} else {
		statsdClient, err = statsd.New(config.DataDogStatsdAddr)
		logger, _ = zap.NewProduction()

		tracer.Start(tracer.WithAnalytics(true), tracer.WithRuntimeMetrics())
		defer tracer.Stop()
	}

	defer logger.Sync() // nolint
	sugaredLogger := logger.Sugar()

	if err != nil {
		sugaredLogger.Fatal("could not connect to datadog statsd server")
	}

	if len(config.PostgresURL) == 0 {
		sugaredLogger.Fatal("DATABASE_URL is required")
	}

	st, err := store.NewPostgresStore(
		config.PostgresURL,
		"store",
		config.ServiceName,
		int(config.PGMaxIdelConns),
		int(config.PGMaxOpenConns),
		config.PGConnMaxLifetime,
	)

	if err != nil {
		sugaredLogger.Fatal(err)
	}

	st.SetArchiveTable("archived_store")

	redisClient, err := util.GetRedisClient(config.RedisURL)
	if err != nil {
		sugaredLogger.Fatalw("could not connect to redis", "err", err)
	}

	sw := sweeper.New(&sweeper.NewSweeperOptions{
		Store:        st,
		ServiceName:  config.ServiceName,
		StatsdClient: statsdClient,
		Logger:       logger,
		Interval:     config.SweeperInterval,
		RedisClient:  redisClient,
	})

	sugaredLogger.Infow("Started sweeper")
	sw.Start()
}
