// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package main

import (
	"log"
	"os"
	"path/filepath"

	"github.com/DataDog/datadog-go/statsd"
	"github.com/go-redis/redis/v8"
	"github.com/walletlink/walletlink/store"
	"github.com/walletlink/walletlink/util"
	"go.uber.org/zap"
	"gopkg.in/DataDog/dd-trace-go.v1/ddtrace/tracer"

	"github.com/walletlink/walletlink/webhook"

	"github.com/pkg/errors"
	"github.com/walletlink/walletlink/config"
	"github.com/walletlink/walletlink/server"
)

func main() {
	execFile, err := os.Executable()
	if err != nil {
		log.Fatalln(errors.Wrap(err, "could not determine executable path"))
	}
	execDir := filepath.Dir(execFile)
	webRoot := filepath.Join(execDir, "public")

	var st store.Store
	if len(config.PostgresURL) == 0 {
		st = store.NewMemoryStore()
	} else {
		var err error
		st, err = store.NewPostgresStore(
			config.PostgresURL,
			"store",
			config.ServiceName,
			int(config.PGMaxIdelConns),
			int(config.PGMaxOpenConns),
			config.PGConnMaxLifetime,
		)

		if err != nil {
			log.Panicln(err)
		}
	}

	var (
		logger       *zap.Logger
		statsdClient statsd.ClientInterface
	)

	if config.AppEnv == "development" || config.AppEnv == "test" {
		statsdClient = &statsd.NoOpClient{}
		logger, _ = zap.NewDevelopment()
	} else {
		statsdClient, err = statsd.New(config.DataDogStatsdAddr)
		logger, _ = zap.NewProduction()

		tracer.Start(tracer.WithAnalytics(true), tracer.WithRuntimeMetrics())
		defer tracer.Stop()
	}

	defer logger.Sync() // nolint
	sugaredLogger := logger.Sugar()

	if err != nil {
		sugaredLogger.Error("could not connect to datadog statsd server")
	}

	wh := webhook.NewWebhook(config.ServerURL)
	wh.SetLogger(logger)

	var redisClient *redis.Client
	if config.UseRedisPubSub {
		redisClient, err = util.GetRedisClient(config.RedisURL)
		if err != nil {
			sugaredLogger.Fatalw("could not connect to redis", "err", err)
		}

		sugaredLogger.Info("Using redis pubsub")
	} else {
		sugaredLogger.Info("Using in-memory pubsub")
	}

	srv := server.NewServer(&server.NewServerOptions{
		Store:          st,
		WebRoot:        webRoot,
		AllowedOrigins: config.AllowedOrigins,
		Webhook:        wh,
		ServerURL:      config.ServerURL,
		ForceSSL:       config.ForceSSL,
		ReadDeadline:   config.ReadDeadline,
		ServiceName:    config.ServiceName,
		StatsdClient:   statsdClient,
		Logger:         logger,
		RedisClient:    redisClient,
	})

	sugaredLogger.Infof(
		"walletlinkd %s-%s listening on port %d...\n"+
			"Serving static assets in %s...\n",
		config.Version,
		config.GitCommit,
		config.Port,
		webRoot,
	)
	srv.Start(uint(config.Port))
}
