package sweeper

import (
	"context"
	"time"

	"github.com/DataDog/datadog-go/statsd"
	"github.com/go-redis/redis/v8"
	"github.com/walletlink/walletlink/store"
	"github.com/walletlink/walletlink/util"
	"go.uber.org/zap"
)

// Sweeper - sweeper
type Sweeper struct {
	store        *store.PostgresStore
	logger       *zap.SugaredLogger
	statsdClient statsd.ClientInterface
	serviceName  string
	interval     time.Duration
	redisClient  *redis.Client
}

// NewSweeperOptions - options for New function
type NewSweeperOptions struct {
	Store        *store.PostgresStore
	Logger       *zap.Logger
	StatsdClient statsd.ClientInterface
	ServiceName  string
	Interval     time.Duration
	RedisClient  *redis.Client
}

const (
	fromQueue  = "from-queue"
	toQueue    = "to-queue"
	queueValue = "a"
)

// New - construct a Sweeper
func New(options *NewSweeperOptions) *Sweeper {
	s := &Sweeper{
		store:        options.Store,
		serviceName:  options.ServiceName,
		statsdClient: options.StatsdClient,
		logger:       options.Logger.Sugar(),
		interval:     options.Interval,
		redisClient:  options.RedisClient,
	}

	return s
}

// Start - start the sweeper
func (s *Sweeper) Start() {
	month := 30 * 24 * time.Hour
	size := 10000

	// Make sure single item is enqueued
	if s.checkIfQueueIsEmpty() {
		s.enqueue()
	}

	for {
		time.Sleep(s.interval)

		_, err := s.redisClient.Do(
			context.Background(),
			util.RedisCommandBLMOVE,
			fromQueue,
			toQueue,
			util.RedisParamRight,
			util.RedisParamLeft,
			util.RedisFormatSec(time.Second),
		).Result()

		if err == redis.Nil {
			continue
		}

		s.dequeue()

		startTime := time.Now()
		n, err := s.store.ArchiveKeys(s.logger, size, month)
		duration := time.Since(startTime) * time.Second

		if err != nil {
			s.logger.Warnw("failed to archive table", "err", err, "num", n, "duration", duration)
		}
		s.logger.Infow("archived messages", "num", n, "duration", duration)

		s.enqueue()
	}
}

func (s *Sweeper) checkIfQueueIsEmpty() bool {
	v, err := s.redisClient.LLen(context.Background(), fromQueue).Result()
	if err != nil {
		s.logger.Errorw("failed to push a value to redis", "err", err)
	}

	return v <= 0
}

func (s *Sweeper) enqueue() {
	_, err := s.redisClient.RPush(context.Background(), fromQueue, queueValue).Result()
	if err != nil {
		s.logger.Errorw("failed to push a value to redis", "err", err)
	}
}

func (s *Sweeper) dequeue() {
	_, err := s.redisClient.LRem(context.Background(), toQueue, 1, queueValue).Result()
	if err != nil {
		s.logger.Errorw("failed to remove an item from redis", "err", err)
	}
}
