// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package server

import (
	"crypto/subtle"
	"net/http"
	"net/http/pprof"
	"time"

	"github.com/gorilla/websocket"
	"github.com/walletlink/walletlink/config"
	"github.com/walletlink/walletlink/server/rpc"
)

const (
	websocketReadLimit = 1024 * 1024
	handshaketimeout   = time.Second * 30
)

func (srv *Server) rpcHandler(w http.ResponseWriter, r *http.Request) {
	upgrader := &websocket.Upgrader{HandshakeTimeout: handshaketimeout}

	upgrader.CheckOrigin = func(r *http.Request) bool {
		origin := r.Header.Get("Origin")
		if len(origin) == 0 || len(srv.allowedOrigins) == 0 || srv.allowedOrigins.Contains("*") {
			return true
		}
		return srv.allowedOrigins.Contains(origin)
	}

	ws, err := upgrader.Upgrade(w, r, nil)

	if err != nil {
		srv.logger.Errorw("websocket upgrade failed", "err", err)
		return
	}
	defer ws.Close()
	ws.SetReadLimit(websocketReadLimit)

	sendCh := make(chan interface{})
	defer close(sendCh)

	go func() {
		isClosed := false

		for {
			res, ok := <-sendCh
			if !ok {
				return
			}

			if isClosed {
				continue
			}

			if v, ok := res.(rune); ok && v == 'h' {
				if err := ws.WriteMessage(websocket.TextMessage, []byte("h")); err != nil {
					srv.logger.Errorw("websocket write failed", "err", err)
				}
				continue
			}

			var err error
			if v, ok := res.(string); ok {
				err = ws.WriteMessage(websocket.TextMessage, []byte(v))
			} else {
				err = ws.WriteJSON(res)
			}

			if err != nil {
				srv.logger.Errorw("websocket write failed", "err", err)
				if err := ws.Close(); err != nil {
					srv.logger.Errorw("websocket close failed", "err", err)
				}
				isClosed = true
			}
		}
	}()

	handler, err := rpc.NewMessageHandler(
		sendCh,
		srv.store,
		srv.pubSub,
		srv.webhook,
	)

	handler.SetStatsdClient(srv.statsdClient)

	if err != nil {
		srv.logger.Errorw("message handler creation failed", "err", err)
		return
	}

	defer handler.Close()

	for {
		if srv.readDeadline > 0 {
			if err := ws.SetReadDeadline(time.Now().Add(srv.readDeadline)); err != nil {
				srv.logger.Errorw("websocket set read deadline failed", "err", err)
				break
			}
		}

		msgType, msgData, err := ws.ReadMessage()
		if err != nil {
			if !websocket.IsCloseError(err) &&
				!websocket.IsUnexpectedCloseError(err) {
				srv.logger.Errorw("websocket read failed", "err", err)
			}
			break
		}

		if msgType != websocket.TextMessage && msgType != websocket.BinaryMessage {
			continue
		}

		if err := handler.HandleRawMessage(msgData); err != nil {
			srv.logger.Errorw("failed to handle a raw message", "err", err)
			break
		}
	}
}

func (srv *Server) goroutineHandler(w http.ResponseWriter, r *http.Request) {
	user, pass, ok := r.BasicAuth()

	if !ok || subtle.ConstantTimeCompare([]byte(user), []byte("wallet")) != 1 || subtle.ConstantTimeCompare([]byte(pass), []byte(config.BasicAuthPassword)) != 1 {
		w.Header().Set("WWW-Authenticate", `Basic realm="Provide user name and password"`)
		w.WriteHeader(401)
		if _, err := w.Write([]byte("Unauthorised.\n")); err != nil {
			srv.logger.Errorw("failed to render error page", "err", err)
		}
		return
	}

	pprof.Handler("goroutine").ServeHTTP(w, r)
}

func (srv *Server) heapHandler(w http.ResponseWriter, r *http.Request) {
	user, pass, ok := r.BasicAuth()

	if !ok || subtle.ConstantTimeCompare([]byte(user), []byte("wallet")) != 1 || subtle.ConstantTimeCompare([]byte(pass), []byte(config.BasicAuthPassword)) != 1 {
		w.Header().Set("WWW-Authenticate", `Basic realm="Provide user name and password"`)
		w.WriteHeader(401)
		if _, err := w.Write([]byte("Unauthorised.\n")); err != nil {
			srv.logger.Errorw("failed to render error page", "err", err)
		}
		return
	}

	pprof.Handler("heap").ServeHTTP(w, r)
}
