// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package rpc

import (
	"fmt"
	"log"

	"github.com/DataDog/datadog-go/statsd"
	"github.com/walletlink/walletlink/webhook"

	"github.com/pkg/errors"
	"github.com/walletlink/walletlink/store"
	"github.com/walletlink/walletlink/store/models"
	"github.com/walletlink/walletlink/util"
)

// MessageHandler - handles RPC messages
type MessageHandler struct {
	authedSessions util.StringSet // IDs of authenticated sessions
	isHost         bool

	sendCh  chan<- interface{}
	store   store.Store
	pubSub  IPubSub
	webhook webhook.Caller

	statsdClient statsd.ClientInterface
}

// NewMessageHandler - construct a MessageHandler
func NewMessageHandler(
	sendCh chan<- interface{},
	sto store.Store,
	pubSub IPubSub,
	webhook webhook.Caller,
) (*MessageHandler, error) {
	if sendCh == nil {
		return nil, errors.Errorf("sendCh must not be nil")
	}
	if sto == nil {
		return nil, errors.Errorf("store must not be nil")
	}
	if pubSub == nil {
		return nil, errors.Errorf("pubSub must not be nil")
	}
	return &MessageHandler{
		authedSessions: util.NewStringSet(),
		isHost:         false,
		sendCh:         sendCh,
		store:          sto,
		pubSub:         pubSub,
		webhook:        webhook,
	}, nil
}

// SetStatsdClient - sets statsd client
func (c *MessageHandler) SetStatsdClient(statsdClient statsd.ClientInterface) {
	c.statsdClient = statsdClient
}

// HandleRawMessage - handle a raw client message
func (c *MessageHandler) HandleRawMessage(data []byte) error {
	if len(data) == 1 && data[0] == 'h' {
		c.sendCh <- 'h'
		return nil
	}

	msg, msgType, err := unmarshalClientMessage(data)
	if err != nil {
		return err
	}

	var res serverMessage
	messageType := "unknown_message"
	var webhookID *string

	switch msg := msg.(type) {
	case *clientMessageHostSession:
		res = c.handleHostSession(msg)
		messageType = "host_session"
	case *clientMessageJoinSession:
		res = c.handleJoinSession(msg)
		messageType = "join_session"
	case *clientMessageIsLinked:
		res = c.handleIsLinked(msg)
		messageType = "is_linked"
	case *clientMessageSetSessionConfig:
		res = c.handleSetSessionConfig(msg)
		messageType = "set_session_config"
		webhookID = msg.WebhookID
	case *clientMessageGetSessionConfig:
		res = c.handleGetSessionConfig(msg)
		messageType = "get_session_config"
	case *clientMessagePublishEvent:
		res = c.handlePublishEvent(msg)
		messageType = "publish_event"
	default:
		errMsg := fmt.Sprintf("unsupported message type: %s", msgType)
		res = newServerMessageFail(0, "", errMsg)
	}

	if c.statsdClient != nil {
		if err := c.statsdClient.Count("message_count", 1, []string{"type:" + messageType}, 1); err != nil {
			log.Printf("failed to send metrics: %v", err)
		}

		if err := c.statsdClient.Gauge("message_size", float64(len(data)), []string{"type:" + messageType}, 1); err != nil {
			log.Printf("failed to send metrics: %v", err)
		}

		if webhookID != nil {
			if err := c.statsdClient.Set("webhook_id", *webhookID, []string{"type:" + messageType}, 1); err != nil {
				log.Printf("failed to send metrics: %v", err)
			}
		}
	}

	if res != nil {
		c.sendCh <- res
	}

	return nil
}

// Close - clean up
func (c *MessageHandler) Close() {
	c.pubSub.UnsubscribeAll(c.sendCh)
}

func (c *MessageHandler) handleHostSession(
	msg *clientMessageHostSession,
) serverMessage {
	if c.isHost {
		return newServerMessageFail(
			msg.ID, msg.SessionID, "cannot host more than one session",
		)
	}

	session, err := c.findSessionWithIDAndKey(msg.SessionID, msg.SessionKey)
	if err != nil {
		return newServerMessageFail(msg.ID, msg.SessionID, err.Error())
	}

	if session == nil {
		// there isn't an existing session; persist the new session
		session = &models.Session{ID: msg.SessionID, Key: msg.SessionKey}
		if err := session.Save(c.store); err != nil {
			return newServerMessageFail(msg.ID, msg.SessionID, "internal error")
		}
	}

	c.isHost = true
	c.authedSessions.Add(msg.SessionID)
	c.pubSub.Subscribe(hostPubSubID(msg.SessionID), c.sendCh)

	return newServerMessageOK(msg.ID, msg.SessionID)
}

func (c *MessageHandler) handleJoinSession(
	msg *clientMessageJoinSession,
) serverMessage {
	if c.isHost {
		return newServerMessageFail(
			msg.ID, msg.SessionID, "host cannot join a session",
		)
	}

	session, err := c.findSessionWithIDAndKey(msg.SessionID, msg.SessionKey)
	if err != nil {
		return newServerMessageFail(msg.ID, msg.SessionID, err.Error())
	}

	if session == nil {
		// there isn't an existing session; fail
		errMsg := fmt.Sprintf("no such session: %s", msg.SessionID)
		return newServerMessageFail(msg.ID, msg.SessionID, errMsg)
	}

	session.Linked = true
	if err := session.Save(c.store); err != nil {
		fmt.Println(err)
		return newServerMessageFail(msg.ID, msg.SessionID, "internal error")
	}

	c.authedSessions.Add(msg.SessionID)
	gpsID := guestPubSubID(msg.SessionID)

	c.pubSub.Subscribe(gpsID, c.sendCh)
	onlineGuests := c.pubSub.Len(gpsID)

	// send Linked message to host
	subID := hostPubSubID(msg.SessionID)
	joinedMsg := newServerMessageLinked(msg.SessionID, onlineGuests)
	c.pubSub.Publish(subID, joinedMsg)

	return newServerMessageOK(msg.ID, msg.SessionID)
}

func (c *MessageHandler) handleIsLinked(
	msg *clientMessageIsLinked,
) serverMessage {
	if !c.isHost {
		return newServerMessageFail(
			msg.ID, msg.SessionID, "only hosts are allowed",
		)
	}

	session, err := c.findAuthedSessionWithID(msg.SessionID)
	if err != nil {
		return newServerMessageFail(msg.ID, msg.SessionID, err.Error())
	}

	onlineGuests := c.pubSub.Len(guestPubSubID(msg.SessionID))

	return newServerMessageIsLinkedOK(
		msg.ID, msg.SessionID, session.Linked, onlineGuests,
	)
}

func (c *MessageHandler) handleSetSessionConfig(
	msg *clientMessageSetSessionConfig,
) serverMessage {
	if valid, invalidReason := models.IsValidSessionConfig(
		msg.WebhookID,
		msg.WebhookURL,
		msg.Metadata,
	); !valid {
		return newServerMessageFail(msg.ID, msg.SessionID, invalidReason)
	}

	session, err := c.findAuthedSessionWithID(msg.SessionID)
	if err != nil {
		return newServerMessageFail(msg.ID, msg.SessionID, err.Error())
	}

	if msg.WebhookID != nil {
		session.WebhookID = *msg.WebhookID
	}

	if msg.WebhookURL != nil {
		session.WebhookURL = *msg.WebhookURL
	}

	if session.Metadata == nil {
		session.Metadata = map[string]string{}
	}

	for k, v := range msg.Metadata {
		if v != nil {
			// disallow having more than 50 entries
			if _, exists := session.Metadata[k]; !exists && len(session.Metadata) >= 50 {
				continue
			}
			session.Metadata[k] = *v
		} else {
			delete(session.Metadata, k)
		}
	}

	if err := session.Save(c.store); err != nil {
		fmt.Println(err)
		return newServerMessageFail(msg.ID, msg.SessionID, "internal error")
	}

	// send SessionConfigUpdated message
	updatedMsg := newServerMessageSessionConfigUpdated(
		msg.SessionID, session.WebhookID, session.WebhookURL, session.Metadata,
	)
	for _, subID := range []string{
		hostPubSubID(msg.SessionID),
		guestPubSubID(msg.SessionID),
	} {
		c.pubSub.Publish(subID, updatedMsg)
	}

	return newServerMessageOK(msg.ID, msg.SessionID)
}

func (c *MessageHandler) handleGetSessionConfig(
	msg *clientMessageGetSessionConfig,
) serverMessage {
	session, err := c.findAuthedSessionWithID(msg.SessionID)
	if err != nil {
		return newServerMessageFail(msg.ID, msg.SessionID, err.Error())
	}

	return newServerMessageGetSessionConfigOK(
		msg.ID,
		msg.SessionID,
		session.WebhookID,
		session.WebhookURL,
		session.Metadata,
	)
}

func (c *MessageHandler) handlePublishEvent(
	msg *clientMessagePublishEvent,
) serverMessage {
	lenEvent := len(msg.Event)
	if lenEvent == 0 || lenEvent > 100 {
		return newServerMessageFail(msg.ID, msg.SessionID, "invalid event name")
	}

	session, err := c.findAuthedSessionWithID(msg.SessionID)
	if err != nil {
		return newServerMessageFail(msg.ID, msg.SessionID, err.Error())
	}

	eventID, err := util.RandomHex(4)
	if err != nil {
		fmt.Println(errors.Wrap(err, "failed to generate eventID"))
		return newServerMessageFail(msg.ID, msg.SessionID, "internal error")
	}

	event := &models.Event{
		ID:    eventID,
		Event: msg.Event,
		Data:  msg.Data,
	}
	if err := event.Save(c.store, msg.SessionID); err != nil {
		fmt.Println(err)
		return newServerMessageFail(msg.ID, msg.SessionID, "internal error")
	}

	var subID string
	if c.isHost {
		// if host, publish to guests
		subID = guestPubSubID(msg.SessionID)
	} else {
		// if guest, publish to host
		subID = hostPubSubID(msg.SessionID)
	}

	eventMsg := newServerMessageEvent(msg.SessionID, eventID, msg.Event, msg.Data)
	c.pubSub.Publish(subID, eventMsg)

	if c.isHost && msg.CallWebhook {
		go c.callWebhook(eventID, session.ID, session.WebhookID, session.WebhookURL)
	}

	return newServerMessagePublishEventOK(msg.ID, msg.SessionID, eventID)
}

func (c *MessageHandler) findSessionWithIDAndKey(
	sessionID string,
	sessionKey string,
) (*models.Session, error) {
	if !models.IsValidSessionID(sessionID) {
		return nil, errors.New("invalid session id")
	}
	if !models.IsValidSessionKey(sessionKey) {
		return nil, errors.New("invalid session key")
	}

	session, err := models.LoadSession(c.store, sessionID)
	if err != nil {
		fmt.Println(err)
		return nil, errors.New("internal error")
	}
	if session == nil {
		return nil, nil
	}

	// there is an existing session; check that session key matches
	if session.Key != sessionKey {
		return nil, errors.New("incorrect session key")
	}

	return session, nil
}

func (c *MessageHandler) findAuthedSessionWithID(
	sessionID string,
) (*models.Session, error) {
	if !c.authedSessions.Contains(sessionID) {
		return nil, errors.Errorf("not authenticated to session: %s", sessionID)
	}

	session, err := models.LoadSession(c.store, sessionID)
	if err != nil {
		fmt.Println(err)
		return nil, errors.New("internal error")
	}
	if session == nil {
		return nil, errors.New("session is gone somehow")
	}

	return session, nil
}

func (c *MessageHandler) callWebhook(eventID, sessionID, webhookID, webhookURL string) {
	if c.webhook == nil || len(webhookURL) == 0 || len(webhookID) == 0 {
		return
	}

	err := c.webhook.Call(eventID, sessionID, webhookID, webhookURL)
	if err != nil {
		fmt.Println(errors.Wrap(err, "failed to call webhook"))
	}

}

func hostPubSubID(sessionID string) string {
	return "h." + sessionID
}

func guestPubSubID(sessionID string) string {
	return "g." + sessionID
}
