// Copyright (c) 2018-2021 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2021 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package rpc

import (
	"context"
	"encoding/json"
	"sync"

	"github.com/go-redis/redis/v8"
	"go.uber.org/zap"
)

// RedisPubSub - pub/sub interface for message senders
type RedisPubSub struct {
	redisClient *redis.Client
	lock        *sync.Mutex
	pubSubs     map[string]map[Subscriber]*redis.PubSub
	logger      *zap.SugaredLogger
}

// NewRedisPubSub - construct a RedisPubSub
func NewRedisPubSub(redisClient *redis.Client, logger *zap.Logger) *RedisPubSub {
	return &RedisPubSub{
		redisClient: redisClient,
		lock:        &sync.Mutex{},
		pubSubs:     map[string]map[Subscriber]*redis.PubSub{},
		logger:      logger.Sugar().Named("pubsub"),
	}
}

// Unsubscribe - unsubscribes a Subscriber from an id
func (r *RedisPubSub) Unsubscribe(id string, subscriber Subscriber) {
	if id == "" || subscriber == nil {
		return
	}

	r.lock.Lock()
	defer r.lock.Unlock()

	pubSubs, ok := r.pubSubs[id]
	if !ok {
		return
	}

	pubSub, ok := pubSubs[subscriber]
	if !ok {
		return
	}

	if err := pubSub.Unsubscribe(context.Background(), id); err != nil {
		r.logger.Warnw("failed to unsubscribe", "err", err)
	}
}

// UnsubscribeAll - unsubscribes subscriber from every id and returns the number
// of unsubscriptions done
func (r *RedisPubSub) UnsubscribeAll(subscriber Subscriber) int {
	result := 0

	if subscriber == nil {
		return result
	}

	r.lock.Lock()
	defer r.lock.Unlock()

	for id, pubSubs := range r.pubSubs {
		pubSub, ok := pubSubs[subscriber]
		if !ok {
			continue
		}

		if err := pubSub.Unsubscribe(context.Background(), id); err != nil {
			r.logger.Warnw("failed to unsubscribe", "err", err)
		}

		if err := pubSub.Close(); err != nil {
			r.logger.Warnw("failed to close pubsub", "err", err)
		}

		result++
	}

	return result
}

// Len - returns the number of subscribers for a given id
func (r *RedisPubSub) Len(id string) int {
	n, err := r.redisClient.PubSubNumSub(context.Background(), id).Result()
	if err != nil {
		r.logger.Warnw("failed to get a number of subscribers", "err", err)
		return 0
	}

	return int(n[id])
}

// Publish - publishes a message to all subscribers of an id and returns
// the number of subscribers messaged
func (r *RedisPubSub) Publish(id string, msg interface{}) int {
	v, err := json.Marshal(msg)
	if err != nil {
		r.logger.Warnw("failed to marshal json", "err", err)
		return 0
	}

	if _, err := r.redisClient.Publish(context.Background(), id, v).Result(); err != nil {
		r.logger.Warnw("failed to publish a message", "err", err)
		return 0
	}

	return r.Len(id)
}

// Subscribe - subscribes a Subscriber to an id
func (r *RedisPubSub) Subscribe(id string, subscriber Subscriber) {
	if id == "" || subscriber == nil {
		return
	}
	r.lock.Lock()
	defer r.lock.Unlock()

	pubSub, ok := r.pubSubs[id][subscriber]
	if !ok {
		pubSub = r.redisClient.Subscribe(context.Background(), id)
		if r.pubSubs[id] == nil {
			r.pubSubs[id] = map[Subscriber]*redis.PubSub{}
		}

		r.pubSubs[id][subscriber] = pubSub
	}

	ch := pubSub.Channel()
	go func() {
		for msg := range ch {
			// When the message is wrapped with double quotes
			if string(msg.Payload[0]) == "\"" {
				var result string
				if err := json.Unmarshal([]byte(msg.Payload), &result); err != nil {
					r.logger.Warnw("failed to parse a message", "err", err)
				} else {
					subscriber <- result
				}
				continue
			}

			subscriber <- msg.Payload
		}
	}()
}
