// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package server

import (
	"fmt"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"net/url"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"github.com/DataDog/datadog-go/statsd"
	"github.com/go-redis/redis/v8"
	"github.com/walletlink/walletlink/config"
	"github.com/walletlink/walletlink/server/rpc"
	"github.com/walletlink/walletlink/store"
	"github.com/walletlink/walletlink/util"
	"github.com/walletlink/walletlink/webhook"
	"go.uber.org/zap"
	"gopkg.in/DataDog/dd-trace-go.v1/contrib/gorilla/mux"
	"gopkg.in/DataDog/dd-trace-go.v1/ddtrace/tracer"
)

// Server - server
type Server struct {
	router         *mux.Router
	store          store.Store
	pubSub         rpc.IPubSub
	allowedOrigins util.StringSet
	webhook        webhook.Caller
	serverURL      string
	readDeadline   time.Duration

	serviceName  string
	statsdClient statsd.ClientInterface
	memoryCache  *util.MemoryCache
	cacheTTL     time.Duration

	logger *zap.SugaredLogger
}

// NewServerOptions - options for NewServer function
type NewServerOptions struct {
	Store          store.Store
	WebRoot        string
	AllowedOrigins util.StringSet
	Webhook        webhook.Caller
	ServerURL      string
	ForceSSL       bool
	ReadDeadline   time.Duration
	ServiceName    string
	StatsdClient   statsd.ClientInterface
	Logger         *zap.Logger
	RedisClient    *redis.Client
}

const (
	clientTypeKey         = "client-type"
	clientTypeAndroid     = "android"
	clientTypeIOS         = "ios"
	clientTypeWeb         = "web"
	clientTypeHealthCheck = "healthcheck"
	clientTypeUnknown     = "unknown"

	statsdSessionIDSetKey   = "session_id"
	statsdOpenFileMetricKey = "open_files"
	statsdIPSetKey          = "ip"

	activeConnsThreshold = 50000
)

// NewServer - construct a Server
func NewServer(options *NewServerOptions) *Server {
	if options == nil {
		options = &NewServerOptions{}
	}

	st := options.Store
	if st == nil {
		st = store.NewMemoryStore()
	}

	logger := options.Logger
	if options.Logger == nil {
		logger = zap.NewNop()
	}

	var pubSub rpc.IPubSub
	if options.RedisClient != nil {
		pubSub = rpc.NewRedisPubSub(options.RedisClient, logger)
	} else {
		pubSub = rpc.NewPubSub()
	}

	router := mux.NewRouter(mux.WithServiceName(options.ServiceName))

	srv := &Server{
		router:         router,
		store:          st,
		pubSub:         pubSub,
		allowedOrigins: options.AllowedOrigins,
		webhook:        options.Webhook,
		serverURL:      options.ServerURL,
		readDeadline:   options.ReadDeadline,

		serviceName:  options.ServiceName,
		statsdClient: options.StatsdClient,

		memoryCache: util.NewMemoryCache(),
		cacheTTL:    config.MemoryCacheTTL,
		logger:      logger.Sugar(),
	}

	if options.ForceSSL {
		router.Use(srv.forceSSL)
	}

	router.Use(srv.ddTagOSType)
	router.Use(srv.ddTrackSessionID)
	router.Use(srv.logRequest)

	router.HandleFunc("/rpc", srv.rpcHandler).Methods("GET")
	router.HandleFunc("/events", srv.getEventsSinceHandler).Methods("GET")
	router.HandleFunc("/events/{id}", srv.getEventByIDHandler).Methods("GET")
	router.HandleFunc("/events/{id}/seen", srv.markEventSeenHandler).Methods("POST")

	if len(config.BasicAuthPassword) > 0 {
		router.HandleFunc("/debug/pprof/goroutine", srv.goroutineHandler).Methods("GET")
		router.HandleFunc("/debug/pprof/heap", srv.heapHandler).Methods("GET")
	}

	if len(options.WebRoot) > 0 {
		router.PathPrefix("/").Methods("GET").Handler(
			http.FileServer(http.Dir(options.WebRoot)),
		)
	}

	return srv
}

// Start - start the server
func (s *Server) Start(port uint) {
	httpServer := &http.Server{
		Addr: fmt.Sprintf("0.0.0.0:%d", port),
		// Good practice to set timeouts to avoid Slowloris attacks.
		WriteTimeout: time.Second * 30,
		ReadTimeout:  time.Second * 30,
		IdleTimeout:  time.Second * 60,
		Handler:      s.router,
	}

	go func() {
		ticker := time.NewTicker(3 * time.Minute)

		for {
			for range ticker.C {
				files := map[string]time.Time{}

				fdpath := filepath.Join("/proc", strconv.Itoa(os.Getpid()), "fd")
				ffiles, err := ioutil.ReadDir(fdpath)
				if err != nil {
					s.logger.Warnw("Failed to read /proc", "err", err)
					continue
				}

				for _, f := range ffiles {
					if f.Name() == "0" || f.Name() == "1" || f.Name() == "2" {
						continue
					}

					fpath, err := os.Readlink(filepath.Join(fdpath, f.Name()))
					if err != nil {
						continue
					}

					files[fpath] = f.ModTime()
				}

				numOfOpenFiles := float64(len(files))
				if err := s.statsdClient.Gauge(statsdOpenFileMetricKey, numOfOpenFiles, nil, 1); err != nil {
					s.logger.Warnw("failed to update metric for", "name", statsdOpenFileMetricKey, "err", err)
				}

				// Restart when there are too many connections open
				if numOfOpenFiles > activeConnsThreshold {
					s.logger.Error("restarting due to too many connections")
					os.Exit(1)
				}
			}
		}
	}()

	go func() {
		ticker := time.NewTicker(5 * time.Minute)
		for {
			for range ticker.C {
				s.memoryCache.HouseKeeping()
			}
		}
	}()

	log.Fatal(httpServer.ListenAndServe())
}

func (s *Server) forceSSL(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if r.URL.Scheme != "https" && r.Header.Get("X-Forwarded-Proto") != "https" {
			if r.Method != http.MethodGet {
				w.WriteHeader(403)
				return
			}

			u, _ := url.Parse(s.serverURL)
			u.Scheme = "https"
			u.Path = r.URL.Path
			u.RawQuery = r.URL.RawQuery

			http.Redirect(w, r, u.String(), 301)
			return
		}
		next.ServeHTTP(w, r)
	})
}

func (s *Server) ddTagOSType(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		span, ok := tracer.SpanFromContext(r.Context())
		if ok {
			span.SetTag(clientTypeKey, getClientType(r))
		}

		next.ServeHTTP(w, r)
	})
}

func (s *Server) ddTrackSessionID(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		sessionID, _, ok := r.BasicAuth()

		if ok && s.statsdClient != nil {
			if err := s.statsdClient.Set(s.serviceName+"."+statsdSessionIDSetKey, sessionID, nil, 1); err != nil {
				s.logger.Warnw("failed to send metrics", "err", err)
			}

			if err := s.statsdClient.Set(s.serviceName+"."+statsdIPSetKey, clientIP(r), nil, 1); err != nil {
				s.logger.Warnw("failed to send metrics", "err", err)
			}
		}

		next.ServeHTTP(w, r)
	})
}

func (s *Server) logRequest(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		start := time.Now()
		path := r.URL.Path
		method := r.Method
		raw := r.URL.RawQuery

		next.ServeHTTP(w, r)

		clientType := getClientType(r)
		if clientType == clientTypeUnknown {
			clientType = strings.TrimSpace(r.Header.Get("User-Agent"))
			if len(clientType) > 20 {
				clientType = clientType[0:20]
			}
		}

		referer := r.Header.Get("Referer")

		s.logger.Infow(
			"REQ",
			"duration",
			time.Since(start)*time.Second,
			"clientType",
			clientType,
			"network.client.ip",
			clientIP(r),
			"http.method",
			method,
			"path",
			path,
			"query",
			raw,
			"referer",
			referer,
		)
	})
}

// getClientType - returns client type from UserAgent header
func getClientType(req *http.Request) string {
	userAgent := strings.TrimSpace(req.Header.Get("User-Agent"))
	tag := clientTypeUnknown
	if strings.HasPrefix(userAgent, "okhttp") {
		tag = clientTypeAndroid
	} else if strings.HasPrefix(userAgent, "Mozilla") {
		tag = clientTypeWeb
	} else if strings.HasPrefix(userAgent, "ELB-HealthChecker") {
		tag = clientTypeHealthCheck
	} else if strings.HasPrefix(userAgent, "Coinbase%20Wallet") {
		tag = clientTypeIOS
	}

	return tag
}

// clientIP - returns client IP from header
func clientIP(req *http.Request) string {
	clientIP := req.Header.Get("X-Forwarded-For")
	clientIP = strings.TrimSpace(strings.Split(clientIP, ",")[0])
	if clientIP != "" {
		return clientIP
	}

	clientIP = strings.TrimSpace(req.Header.Get("X-Real-Ip"))
	if clientIP != "" {
		return clientIP
	}

	if ip, _, err := net.SplitHostPort(strings.TrimSpace(req.RemoteAddr)); err == nil {
		return ip
	}

	return ""
}
