// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package store

import (
	"context"
	"database/sql"
	"encoding/json"
	"fmt"
	"reflect"
	"time"

	"github.com/lib/pq" // register postgres adapter
	"github.com/pkg/errors"
	"go.uber.org/zap"
	sqltrace "gopkg.in/DataDog/dd-trace-go.v1/contrib/database/sql"
)

// PostgresStore - persistent postgres store
type PostgresStore struct {
	db               *sql.DB
	tableName        string
	archiveTableName string
}

var _ Store = (*PostgresStore)(nil)

// NewPostgresStore - make a connection to a PostgreSQL instance, and return
// a PostgresStore
func NewPostgresStore(
	url, tableName, serviceName string,
	maxIdleConns, maxOpenConns int,
	maxConnMaxLifetime time.Duration,
) (*PostgresStore, error) {
	sqltrace.Register("postgres", &pq.Driver{}, sqltrace.WithServiceName(serviceName))

	db, err := sqltrace.Open("postgres", url)
	if err != nil {
		return nil, errors.Wrap(err, "unable to open a postgres connection")
	}

	db.SetMaxIdleConns(maxIdleConns)
	db.SetMaxOpenConns(maxOpenConns)
	db.SetConnMaxLifetime(maxConnMaxLifetime)

	return &PostgresStore{
		db:        db,
		tableName: tableName,
	}, nil
}

// Close - close PostgreSQL connection
func (ps *PostgresStore) Close() error {
	return ps.db.Close()
}

// Set - save the value to the postgres database under a given key
func (ps *PostgresStore) Set(key string, value interface{}) error {
	j, err := json.Marshal(value)
	if err != nil {
		return errors.Wrap(err, "could not serialize value")
	}

	_, err = ps.db.Exec(
		fmt.Sprintf(
			`INSERT INTO %s (key, value) VALUES ($1, $2)
			ON CONFLICT (key)
			DO UPDATE SET value = EXCLUDED.value, updated_at = now()`,
			pq.QuoteIdentifier(ps.tableName),
		),
		key,
		string(j),
	)
	if err != nil {
		return errors.Wrap(err, "failed to insert value into postgres store")
	}

	return nil
}

// SetWithParentID - save the value to the postgres database under a given key with a session ID
func (ps *PostgresStore) SetWithParentID(parentID, key string, value interface{}) error {
	j, err := json.Marshal(value)
	if err != nil {
		return errors.Wrap(err, "could not serialize value")
	}

	_, err = ps.db.Exec(
		fmt.Sprintf(
			`INSERT INTO %s (parent_id, key, value) VALUES ($1, $2, $3)
			ON CONFLICT (key)
			DO UPDATE SET parent_id = EXCLUDED.parent_id, value = EXCLUDED.value, updated_at = now()`,
			pq.QuoteIdentifier(ps.tableName),
		),
		parentID,
		key,
		string(j),
	)
	if err != nil {
		return errors.Wrap(err, "failed to insert value into postgres store")
	}

	return nil
}

// Get - load data for a given key. value passed must be a reference.
// (false, nil) is returned if key does not exist.
func (ps *PostgresStore) Get(key string, value interface{}) (bool, error) {
	var j string
	err := ps.db.QueryRow(
		fmt.Sprintf("SELECT value FROM %s WHERE key = $1", pq.QuoteIdentifier(ps.tableName)),
		key,
	).Scan(&j)
	if err != nil {
		if err == sql.ErrNoRows {
			return false, nil
		}
		return false, errors.Wrap(err, "failed to load value from postgres store")
	}

	if err = json.Unmarshal([]byte(j), value); err != nil {
		return false, errors.Wrap(err, "could not deserialize value")
	}

	return true, nil
}

// FindByPrefix - load all values whose key starts with the given prefix that
// were updated after since. If unseen is true, only return values that have not
// been marked seen.
func (ps *PostgresStore) FindByPrefix(
	prefix string,
	since int64,
	unseen bool,
	values interface{},
) error {
	valuesValue := reflect.ValueOf(values)
	if valuesValue.Kind() != reflect.Ptr ||
		valuesValue.Elem().Kind() != reflect.Slice {
		return errors.New("values must be a pointer to a slice")
	}

	q := fmt.Sprintf(
		`SELECT value
		FROM %s
		WHERE key LIKE $1 || '%%' AND updated_at > $2`,
		pq.QuoteIdentifier(ps.tableName),
	)
	if unseen {
		q += " AND seen_at IS NULL"
	}
	q += " ORDER BY updated_at DESC"

	rows, err := ps.db.Query(q, prefix, time.Unix(since, 0))
	if err != nil {
		return errors.Wrap(err, "failed to load values from postgres store")
	}
	defer rows.Close()

	sliceType := reflect.TypeOf(values).Elem()
	elemType := sliceType.Elem()

	vs := reflect.New(sliceType).Elem()

	for rows.Next() {
		var j string
		if err = rows.Scan(&j); err != nil {
			return errors.Wrap(err, "failed to scan row")
		}

		value := reflect.New(elemType).Interface()
		if err = json.Unmarshal([]byte(j), value); err != nil {
			continue
		}

		vs = reflect.Append(vs, reflect.ValueOf(value).Elem())
	}

	reflect.ValueOf(values).Elem().Set(vs)

	return nil
}

// FindByParentID - load all values by parent ID that were updated after since.
// If unseen is true, only return values that have not been marked seen.
func (ps *PostgresStore) FindByParentID(
	parentID string,
	since int64,
	unseen bool,
	values interface{},
) error {
	valuesValue := reflect.ValueOf(values)
	if valuesValue.Kind() != reflect.Ptr ||
		valuesValue.Elem().Kind() != reflect.Slice {
		return errors.New("values must be a pointer to a slice")
	}

	q := fmt.Sprintf(
		`SELECT value FROM %s WHERE parent_id = $1 AND updated_at > $2`,
		pq.QuoteIdentifier(ps.tableName),
	)
	if unseen {
		q += " AND seen_at IS NULL"
	}
	q += " ORDER BY updated_at DESC"

	rows, err := ps.db.Query(q, parentID, time.Unix(since, 0))
	if err != nil {
		return errors.Wrap(err, "failed to load values from postgres store")
	}
	defer rows.Close()

	sliceType := reflect.TypeOf(values).Elem()
	elemType := sliceType.Elem()

	vs := reflect.New(sliceType).Elem()

	for rows.Next() {
		var j string
		if err = rows.Scan(&j); err != nil {
			return errors.Wrap(err, "failed to scan row")
		}

		value := reflect.New(elemType).Interface()
		if err = json.Unmarshal([]byte(j), value); err != nil {
			continue
		}

		vs = reflect.Append(vs, reflect.ValueOf(value).Elem())
	}

	reflect.ValueOf(values).Elem().Set(vs)

	return nil
}

// MarkSeen - mark the value with the given key as seen
func (ps *PostgresStore) MarkSeen(key string) (updated bool, err error) {
	result, err := ps.db.Exec(
		fmt.Sprintf(
			"UPDATE %s SET seen_at = now(), updated_at = now() WHERE key = $1",
			pq.QuoteIdentifier(ps.tableName),
		),
		key,
	)
	if err != nil {
		return false, errors.Wrap(err, "failed to update value")
	}
	num, _ := result.RowsAffected()
	return num > 0, nil
}

// Remove - remove a key. does not return an error if key does not exist
func (ps *PostgresStore) Remove(key string) error {
	_, err := ps.db.Exec(
		fmt.Sprintf("DELETE FROM %s WHERE key = $1", pq.QuoteIdentifier(ps.tableName)),
		key,
	)
	if err != nil && err != sql.ErrNoRows {
		return errors.Wrap(err, "failed to delete value")
	}

	return nil
}

// SetArchiveTable - set the table name for archive data
func (ps *PostgresStore) SetArchiveTable(archiveTableName string) {
	ps.archiveTableName = archiveTableName
}

// ArchiveKeys - archive keys
func (ps *PostgresStore) ArchiveKeys(
	logger *zap.SugaredLogger,
	limit int,
	expiry time.Duration,
) (int, error) {
	numOfArchived := 0

	if len(ps.archiveTableName) == 0 {
		return numOfArchived, errors.New("archive table name is not set")
	}

	q := fmt.Sprintf(
		`SELECT key, value, parent_id, created_at, updated_at, seen_at
		 FROM %s WHERE seen_at IS NOT NULL AND seen_at < $1 LIMIT $2`,
		pq.QuoteIdentifier(ps.tableName),
	)

	rows, err := ps.db.Query(q, time.Now().Add(-expiry), limit)
	if err != nil {
		return numOfArchived, err
	}

	defer rows.Close()

	for rows.Next() {
		var (
			key       string
			value     string
			parentID  sql.NullString
			createdAt time.Time
			updatedAt time.Time
			seenAt    sql.NullTime
		)

		if err = rows.Scan(&key, &value, &parentID, &createdAt, &updatedAt, &seenAt); err != nil {
			return numOfArchived, err
		}

		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Minute)
		defer cancel()

		tx, err := ps.db.BeginTx(ctx, &sql.TxOptions{Isolation: sql.LevelDefault})
		if err != nil {
			return numOfArchived, err
		}
		defer tx.Rollback() // nolint

		if _, err := tx.Exec(
			fmt.Sprintf(
				`INSERT INTO %s (key, value, parent_id, created_at, updated_at, seen_at)
				VALUES ($1, $2, $3, $4, $5, $6)
				ON CONFLICT (key) DO NOTHING`,
				pq.QuoteIdentifier(ps.archiveTableName),
			),
			key, value, parentID, createdAt, updatedAt, seenAt,
		); err != nil {
			return numOfArchived, err
		}

		if _, err := tx.Exec(
			fmt.Sprintf(`DELETE FROM %s WHERE key = $1`, pq.QuoteIdentifier(ps.tableName)),
			key,
		); err != nil {
			return numOfArchived, err
		}

		if err := tx.Commit(); err != nil {
			return numOfArchived, err
		}

		numOfArchived++
	}

	return numOfArchived, nil
}
