// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package util

import (
	"context"
	"log"
	"net"
	"net/url"
)

type IPCheckerFunc interface {
	IsInPrivateNetwork(urlStr string) (bool, error)
	IsIPv6(urlStr string) (bool, error)
	ResolveIPs(urStr string) ([]net.IP, error)
}

type IPChecker struct {
	DNSResolver *net.Resolver
}

var privateIPBlocks []*net.IPNet

func init() {
	for _, cidr := range []string{
		"127.0.0.0/8",    // IPv4 loopback
		"10.0.0.0/8",     // RFC1918
		"172.16.0.0/12",  // RFC1918
		"192.168.0.0/16", // RFC1918
		"169.254.0.0/16", // RFC3927 link-local
		"::1/128",        // IPv6 loopback
		"fe80::/10",      // IPv6 link-local
		"fc00::/7",       // IPv6 unique local addr
	} {
		_, block, err := net.ParseCIDR(cidr)
		if err != nil {
			log.Panicf("failed to parse %q: %v", cidr, err)
		}
		privateIPBlocks = append(privateIPBlocks, block)
	}
}

func NewIPChecker() *IPChecker {
	return &IPChecker{}
}

// IsInPrivateNetwork - check if a given url is in private network
func (ic *IPChecker) IsInPrivateNetwork(urlStr string) (bool, error) {
	ips, err := ic.ResolveIPs(urlStr)
	if err != nil {
		return false, err
	}

	for _, ip := range ips {
		if ip.IsLoopback() || ip.IsLinkLocalUnicast() || ip.IsLinkLocalMulticast() {
			return true, nil
		}

		for _, block := range privateIPBlocks {
			if block.Contains(ip) {
				return true, nil
			}
		}
	}

	return false, nil
}

// IsIPv6 - check if the hostname is the IPv6 hostname
func (ic *IPChecker) IsIPv6(urlStr string) (bool, error) {
	ips, err := ic.ResolveIPs(urlStr)
	if err != nil {
		return false, err
	}

	for _, ip := range ips {
		if ip.To4() != nil {
			return false, nil
		}
	}

	return true, nil
}

// ResolveIPs - resolve IPs by url
func (ic *IPChecker) ResolveIPs(urlStr string) ([]net.IP, error) {
	u, err := url.Parse(urlStr)
	if err != nil {
		return nil, err
	}

	addrs, err := ic.DNSResolver.LookupIPAddr(context.Background(), u.Hostname())
	if err != nil {
		return nil, err
	}

	ips := make([]net.IP, len(addrs))
	for i, ia := range addrs {
		ips[i] = ia.IP
	}

	return ips, nil
}
