// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package util

import (
	"context"
	"fmt"
	"net"
	"os"
	"testing"

	"github.com/google/gopacket"
	layers "github.com/google/gopacket/layers"
	"github.com/stretchr/testify/require"
)

const (
	dnsResolverHost = "127.0.0.1"
	dnsResolverPort = 25600
)

func TestMain(m *testing.M) {
	errCh := make(chan error)
	go startDNS(errCh)
	<-errCh

	os.Exit(m.Run())
}

func startDNS(ready chan error) {
	records := map[string]map[string]string{
		"www.google.com": map[string]string{
			"address": "216.58.196.142",
			"version": "4",
		},
		"www.ipv6.com": map[string]string{
			"address": "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
			"version": "6",
		},
	}

	// Listen on UDP Port
	u, err := net.ListenUDP("udp", &net.UDPAddr{
		Port: dnsResolverPort,
		IP:   net.ParseIP(dnsResolverHost),
	})

	ready <- err

	// Wait to get request on that port
	for {
		tmp := make([]byte, 1024)
		_, addr, _ := u.ReadFrom(tmp)
		clientAddr := addr
		packet := gopacket.NewPacket(tmp, layers.LayerTypeDNS, gopacket.Default)
		dnsPacket := packet.Layer(layers.LayerTypeDNS)
		replyMess, _ := dnsPacket.(*layers.DNS)

		var dnsAnswer layers.DNSResourceRecord
		dnsAnswer.Type = layers.DNSTypeA
		ip, ok := records[string(replyMess.Questions[0].Name)]
		if ok {
			if ip["version"] == "4" {
				a, _, _ := net.ParseCIDR(ip["address"] + "/24")
				dnsAnswer.Type = layers.DNSTypeA
				dnsAnswer.IP = a
				dnsAnswer.Name = []byte(replyMess.Questions[0].Name)
				dnsAnswer.Class = layers.DNSClassIN
			} else {
				a, _, _ := net.ParseCIDR(ip["address"] + "/32")
				dnsAnswer.Type = layers.DNSTypeAAAA
				dnsAnswer.IP = a
				dnsAnswer.Name = []byte(replyMess.Questions[0].Name)
				dnsAnswer.Class = layers.DNSClassIN
			}

			replyMess.QR = true
			replyMess.ANCount = 1
			replyMess.OpCode = layers.DNSOpCodeNotify
			replyMess.AA = true
			replyMess.Answers = append(replyMess.Answers, dnsAnswer)
			replyMess.ResponseCode = layers.DNSResponseCodeNoErr
		} else {
			dnsAnswer.Type = layers.DNSTypeA
			dnsAnswer.Name = []byte(replyMess.Questions[0].Name)
			dnsAnswer.Class = layers.DNSClassIN
			replyMess.QR = true
			replyMess.ANCount = 1
			replyMess.OpCode = layers.DNSOpCodeNotify
			replyMess.AA = true
			replyMess.Answers = append(replyMess.Answers, dnsAnswer)
			replyMess.ResponseCode = layers.DNSResponseCodeNXDomain
		}

		buf := gopacket.NewSerializeBuffer()
		opts := gopacket.SerializeOptions{}
		replyMess.SerializeTo(buf, opts)   //nolint
		u.WriteTo(buf.Bytes(), clientAddr) //nolint
	}
}

func TestIsInPrivateNetwork(t *testing.T) {
	ic := NewIPChecker()
	ic.DNSResolver = &net.Resolver{
		PreferGo: true,
		Dial: func(ctx context.Context, network, address string) (net.Conn, error) {
			return net.Dial("udp", fmt.Sprintf("%s:%d", dnsResolverHost, dnsResolverPort))
		},
	}

	v, err := ic.IsInPrivateNetwork("http://localhost:3000")
	require.Nil(t, err)
	require.Equal(t, v, true)

	v, err = ic.IsInPrivateNetwork("http://127.0.0.1:3000")
	require.Nil(t, err)
	require.Equal(t, v, true)

	v, err = ic.IsInPrivateNetwork("https://www.google.com")
	require.Nil(t, err)
	require.Equal(t, v, false)

	v, err = ic.IsInPrivateNetwork("https://www.facebook.com")
	require.NotNil(t, err)
	require.Equal(t, v, false)

	v, err = ic.IsInPrivateNetwork("http://216.58.196.142")
	require.Nil(t, err)
	require.Equal(t, v, false)
}

func TestIsIPv6(t *testing.T) {
	ic := NewIPChecker()
	ic.DNSResolver = &net.Resolver{
		PreferGo: true,
		Dial: func(ctx context.Context, network, address string) (net.Conn, error) {
			return net.Dial("udp", fmt.Sprintf("%s:%d", dnsResolverHost, dnsResolverPort))
		},
	}

	v, err := ic.IsIPv6("http://localhost:3000")
	require.Nil(t, err)
	require.Equal(t, v, false)

	v, err = ic.IsIPv6("https://www.google.com")
	require.Nil(t, err)
	require.Equal(t, v, false)

	v, err = ic.IsIPv6("https://www.ipv6.com")
	require.Nil(t, err)
	require.Equal(t, v, true)
}
