package util

import (
	"sync"
	"time"
)

type Item struct {
	Data       interface{}
	Expiration *time.Time
}

type MemoryCache struct {
	storage map[string]*Item
	mu      sync.RWMutex
}

// NewMemoryCache - construct a MemoryCache
func NewMemoryCache() *MemoryCache {
	return &MemoryCache{
		storage: make(map[string]*Item),
	}
}

// Set - cache data under a given key
func (mc *MemoryCache) Set(key string, value interface{}, expiry ...time.Duration) {
	mc.mu.Lock()
	defer mc.mu.Unlock()

	if len(expiry) > 0 {
		mc.storage[key] = &Item{
			Data:       value,
			Expiration: timeToPtr(time.Now().Add(expiry[0])),
		}
	} else {
		mc.storage[key] = &Item{
			Data: value,
		}
	}
}

// Get - load data for a given key
func (mc *MemoryCache) Get(key string) interface{} {
	mc.mu.RLock()
	defer mc.mu.RUnlock()

	item := mc.storage[key]
	if item != nil && !item.IsExpired() {
		return mc.storage[key].Data
	}

	return nil
}

// Clean - clean expired data
func (mc *MemoryCache) HouseKeeping() {
	mc.mu.RLock()
	defer mc.mu.RUnlock()

	for key, item := range mc.storage {
		if item != nil && item.IsExpired() {
			delete(mc.storage, key)
		}
	}
}

// IsExpired - returns whether if data is expired
func (i *Item) IsExpired() bool {
	return i.Expiration != nil && i.Expiration.Before(time.Now())
}

// Len - returns number of cached items
func (mc *MemoryCache) Len() int {
	mc.mu.RLock()
	defer mc.mu.RUnlock()

	return len(mc.storage)
}

func timeToPtr(t time.Time) *time.Time {
	return &t
}
