// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package webhook

import (
	"encoding/json"
	"io"
	"net/http"
	"net/http/httptrace"

	"github.com/pkg/errors"
	"github.com/walletlink/walletlink/config"
	"github.com/walletlink/walletlink/util"
	"go.uber.org/zap"
)

// Caller - Interface for calling webhooks
type Caller interface {
	Call(eventID, sessionID, webhookID, webhookURL string) error
}

var _ Caller = (*Webhook)(nil)

// Webhook - used to broadcast push notifications
type Webhook struct {
	serverURL string
	client    *http.Client
	ipChecker util.IPCheckerFunc
	logger    *zap.SugaredLogger
}

type callParams struct {
	EventID   string `json:"eventId"`
	SessionID string `json:"sessionId"`
	WebhookID string `json:"webhookId"`
	ServerURL string `json:"serverUrl"`
}

// NewWebhook - returns a new Webhook instance
func NewWebhook(serverURL string) *Webhook {
	wh := &Webhook{
		serverURL: serverURL,
		client: &http.Client{
			Transport: &http.Transport{
				ResponseHeaderTimeout: config.WebhookTimeout,
			},
			CheckRedirect: func(req *http.Request, via []*http.Request) error {
				return http.ErrUseLastResponse
			},
			Timeout: config.WebhookTimeout,
		},
		ipChecker: util.NewIPChecker(),
	}

	wh.logger = zap.NewNop().Sugar()

	return wh
}

// SetLogger - configure a logger
func (a *Webhook) SetLogger(logger *zap.Logger) {
	a.logger = logger.Sugar().Named("webhook")
}

// Call - call webhook to notify about a new event
func (a *Webhook) Call(eventID, sessionID, webhookID, webhookURL string) error {
	if config.DisallowLocalWebhookURL {
		isPrivate, err := a.ipChecker.IsInPrivateNetwork(webhookURL)
		if err != nil {
			return errors.Wrap(err, "failed to call webhook")
		}

		if isPrivate {
			return errors.New("private network address is not allowed to call webhook")
		}
	}

	if config.DisallowIPv6WebhookURL {
		isIPv6, err := a.ipChecker.IsIPv6(webhookURL)
		if err != nil {
			return errors.Wrap(err, "failed to call webhook")
		}

		if isIPv6 {
			return errors.New("ipv6 address is not allowed to call webhook")
		}
	}

	expectedIPs, err := a.ipChecker.ResolveIPs(webhookURL)
	if err != nil {
		return err
	}

	notification := callParams{
		EventID:   eventID,
		SessionID: sessionID,
		WebhookID: webhookID,
		ServerURL: a.serverURL,
	}

	pr, pw := io.Pipe()

	go func() {
		defer pw.Close()
		if err := json.NewEncoder(pw).Encode(notification); err != nil {
			a.logger.Warnw("failed to send a notification", "err", err)
		}
	}()

	req, err := http.NewRequest("POST", webhookURL, pr)
	if err != nil {
		return err
	}

	trace := &httptrace.ClientTrace{
		DNSDone: func(dnsInfo httptrace.DNSDoneInfo) {
			expectedIPMap := map[string]struct{}{}
			for _, expectedIP := range expectedIPs {
				expectedIPMap[expectedIP.String()] = struct{}{}
			}

			if len(expectedIPMap) != len(dnsInfo.Addrs) {
				a.logger.Warnw(
					"number of resolved addresses are not the same",
					"expected", expectedIPs,
					"actual", dnsInfo.Addrs,
				)
			}

			for _, addr := range dnsInfo.Addrs {
				if _, ok := expectedIPMap[addr.IP.String()]; !ok {
					a.logger.Warnw("unexpected resolved address", "ip", addr.IP.String())
				}
			}
		},
	}
	req = req.WithContext(httptrace.WithClientTrace(req.Context(), trace))

	req.Header.Set("Content-Type", "application/json")
	res, err := a.client.Do(req)
	if err != nil {
		return errors.Wrap(err, "failed to call webhook")
	}

	if res.StatusCode != 200 {
		return errors.Errorf(
			"sending notification failed with status code: %d",
			res.StatusCode,
		)
	}

	return nil
}
