// Copyright (c) 2018-2020 WalletLink.org <https://www.walletlink.org/>
// Copyright (c) 2018-2020 Coinbase, Inc. <https://www.coinbase.com/>
// Licensed under the Apache License, version 2.0

package webhook

import (
	"encoding/json"
	"net"
	"net/http"
	"net/http/httptest"
	"testing"

	"github.com/stretchr/testify/require"
	"github.com/walletlink/walletlink/config"
)

var (
	webhookID = "1234"
	eventID   = "5678"
	sessionID = "abcd"
	serverURL = "https://example.com"
)

type FakeIPChecker struct {
	IsPrivate bool
	IPv6      bool
	IPs       []net.IP
	Err       error
}

func (fic *FakeIPChecker) IsInPrivateNetwork(urlStr string) (bool, error) {
	return fic.IsPrivate, fic.Err
}

func (fic *FakeIPChecker) IsIPv6(urlStr string) (bool, error) {
	return fic.IPv6, fic.Err
}

func (fic *FakeIPChecker) ResolveIPs(urlStr string) ([]net.IP, error) {
	return fic.IPs, fic.Err
}

func TestSuccessfulWebhookSend(t *testing.T) {
	testServer := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		w.WriteHeader(200)

		defer r.Body.Close()

		requestBody := callParams{}

		if err := json.NewDecoder(r.Body).Decode(&requestBody); err != nil {
			require.FailNow(t, "unable to decode request body")
		}

		require.Equal(t, webhookID, requestBody.WebhookID)
		require.Equal(t, eventID, requestBody.EventID)
		require.Equal(t, sessionID, requestBody.SessionID)
		require.Equal(t, serverURL, requestBody.ServerURL)
	}))

	defer testServer.Close()

	wh := NewWebhook(serverURL)
	err := wh.Call(eventID, sessionID, webhookID, testServer.URL)
	require.Nil(t, err)
}

func TestFailedWebhookSend(t *testing.T) {
	testServer := httptest.NewServer(http.HandlerFunc(func(
		w http.ResponseWriter, r *http.Request,
	) {
		w.WriteHeader(400)
	}))

	defer testServer.Close()

	wh := NewWebhook(serverURL)
	err := wh.Call(eventID, sessionID, webhookID, testServer.URL)
	require.EqualError(
		t, err, "sending notification failed with status code: 400",
	)
}

func TestIPv6WebhookURL(t *testing.T) {
	config.DisallowLocalWebhookURL = true

	wh := NewWebhook(serverURL)
	wh.ipChecker = &FakeIPChecker{IsPrivate: false, IPv6: true, Err: nil}
	err := wh.Call(eventID, sessionID, webhookID, "http://ipv6.com")
	require.EqualError(
		t, err, "ipv6 address is not allowed to call webhook",
	)
}

func TestLoopbackWebhookURL(t *testing.T) {
	config.DisallowLocalWebhookURL = true

	wh := NewWebhook(serverURL)
	wh.ipChecker = &FakeIPChecker{IsPrivate: true, IPv6: false, Err: nil}
	err := wh.Call(eventID, sessionID, webhookID, "http://private.address:8080")
	require.EqualError(
		t, err, "private network address is not allowed to call webhook",
	)
}
