const webpack = require("webpack");
const path = require("path");
const CopyPlugin = require("copy-webpack-plugin");
const babelConfig = require('./.babelrc');
const ReactRefreshWebpackPlugin = require('@pmmmwh/react-refresh-webpack-plugin');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const DotenvWebpack = require('dotenv-webpack');
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin;

function allowHttpmatches(buffer) {
  // copy-webpack-plugin passes a buffer
  var manifest = JSON.parse(buffer.toString());

  // make any modifications you like, such as
  manifest.content_scripts[0].matches = ["https://*/*", "http://*/*"];
  manifest.web_accessible_resources[0].matches = ["*://*/*"];

  return JSON.stringify(manifest, null, 2);
}
module.exports = (env, { mode }) => {
  const webpackConfig = {
    mode: 'development',
    stats: 'errors-warnings',
    devtool: 'inline-source-map',
    entry: {
      popup: path.join(__dirname, "src/index.tsx"),
      home: path.join(__dirname, "src/home.tsx"),
      debug: path.join(__dirname, "src/debug/debug.tsx"),
      content: path.join(__dirname, "src/contentscript/content.ts"),
      service: path.join(__dirname, "src/serviceworker/service.ts"),
      requestProvider: path.join(__dirname, "src/contentscript/requestProvider.ts"),
      requestRelay: path.join(__dirname, "src/contentscript/requestRelay.ts"),
    },
    output: {
      path: path.join(__dirname, "dist"),
      filename: "[name].js",
      publicPath: '/',
      clean: true
    },
    module: {
      rules: [
        {
          test: /\.(js|jsx)$/,
          use: "babel-loader",
          exclude: [/node_modules/, /walletLink/],
        },
        {
          test: /\.ts$/,
          loader: "ts-loader",
          exclude: [/node_modules/, /walletLink/],
        },
        {
          test: /\.tsx$/,
          use: [
            {
              loader: "babel-loader",
              options: {
                ...babelConfig,
                plugins: babelConfig.plugins.concat(['react-refresh/babel']),
              },
            },
            {
              loader: 'linaria/loader',
              options: { 
                displayName: true,
                babelOptions: 
                { 
                presets: [
                  "@babel/preset-react",
                  '@babel/preset-typescript'
                ]
                }
              },
            }
          ], 
          exclude: [/node_modules/, /walletLink/],
        },
        {
          test: /\.html/,
          use: [{ loader: 'html-loader' }],
        },
        {
          test: /\.css$/,
          use:  [
            {
              loader: MiniCssExtractPlugin.loader,
              options: {
               // Enables es module syntax for better bundle optimization
                esModule: true,
                publicPath: ''
              },
            },
            'css-loader',
          ],
        },
        {
          test: /\.(svg)$/,
          use: "file-loader",
        },
        {
          test: /\.png$/,
          use: [
            {
              loader: "file-loader",
            },
            {
              loader: "url-loader",
              options: {
                mimetype: "image/png",
              },
            },
          ],
        },
        {
          test: /\.(woff2)$/,
          use: [
            {
              loader: 'file-loader',
              options: {
                name: '[name].[ext]',
                outputPath: 'fonts/'
              }
            }
          ]
        },
      ],
    },
    resolve: {
      extensions: [".js", ".jsx", ".tsx", ".ts"],
      alias: {
        "react-dom": "@hot-loader/react-dom",
        "stream": require.resolve("stream-browserify"),
        "crypto": require.resolve("crypto-browserify"),
        "process": "process/browser",
        "Buffer": require.resolve("buffer"),
        "@wallets": path.resolve(__dirname, "src/datalayer/wallets/"),
        "@http": path.resolve(__dirname, "src/datalayer/http/"),
        "@blockchains": path.resolve(__dirname, "src/datalayer/blockchains/"),
        "@ethereum": path.resolve(__dirname, "src/datalayer/ethereum/"),
        "@compoundfinance": path.resolve(__dirname, "src/datalayer/compoundfinance/"),
        "@store": path.resolve(__dirname, "src/datalayer/store/"),
        "@core": path.resolve(__dirname, "src/datalayer/core/"),
        "@user": path.resolve(__dirname, "src/datalayer/user/")
      },
      fallback: {
        "crypto": require.resolve("crypto-browserify"),
        "stream": require.resolve("stream-browserify"),
        "buffer": require.resolve("buffer")
      }
    },
    devServer: {
      contentBase: "./dist",
    },
    plugins: [
      new HtmlWebpackPlugin({
        templateContent: `
          <html>
            <head>
              <meta charset="UTF-8">
              <script src="ga.js"></script>
              <title>Coinbase Wallet</title>
            </head>
            <body>
                <div id="root"></div>
            </body>
          </html>
        `,
        chunks: ["popup"]
      }),
      new HtmlWebpackPlugin({
        filename: "home.html",
        templateContent: `
          <html>
            <head>
              <meta charset="UTF-8">
              <script src="ga.js"></script>
              <title>Wallet Extension</title>
            </head>
            <body>
                <div id="root"></div>
            </body>
          </html>
        `,
        chunks: ["home"]
      }),
      new HtmlWebpackPlugin({
        filename: "debug.html",
        templateContent: `
          <html>
            <head>
              <title>Wallet Extension</title>
            </head>
            <body>
                <div id="root"></div>
            </body>
          </html>
        `,
        chunks: ["debug"]
      }),
      new CopyPlugin({
        patterns: [
          { from: "public", to: "." },
          {
            from: "public/manifest.json",
            to: "./manifest.json",
            transform(content,path) {
              return allowHttpmatches(content)
            },
            force: true
          }
        ],
      }),
      new webpack.ProvidePlugin({
        process: 'process/browser',
      }),
      new webpack.HotModuleReplacementPlugin(),
      new BundleAnalyzerPlugin({
        analyzerMode: 'disabled',
        generateStatFile: true
      }),
      new ReactRefreshWebpackPlugin(),
      new MiniCssExtractPlugin({
        filename: '[name].css',
      }),
      new webpack.ProvidePlugin({
        process: 'process/browser',
        Buffer: ['buffer', 'Buffer']
      }),
      new DotenvWebpack({
        path: path.join(__dirname,'.env.local'),
        safe: true
      })
    ],
  }

  return webpackConfig;
};
