const webpack = require("webpack");
const path = require("path");
const CopyPlugin = require("copy-webpack-plugin");
const babelConfig = require('./.babelrc');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const HtmlWebpackPlugin = require('html-webpack-plugin');
const DotenvWebpack = require('dotenv-webpack');
const TerserPlugin = require('terser-webpack-plugin');

module.exports = () => {
  const webpackConfig = {
    mode: "production",
    entry: {
      popup: path.join(__dirname, "src/index.tsx"),
      home: path.join(__dirname, "src/home.tsx"),
      content: path.join(__dirname, "src/contentscript/content.ts"),
      service: path.join(__dirname, "src/serviceworker/service.ts"),
      requestProvider: path.join(__dirname, "src/contentscript/requestProvider.ts"),
      requestRelay: path.join(__dirname, "src/contentscript/requestRelay.ts"),
    },
    output: {
      path: path.join(__dirname, "dist"),
      filename: "[name].js",
      clean: true
    },
    optimization: {
      minimize: true,
      minimizer: [
        new TerserPlugin({
          terserOptions: {
            output: {
              // Turned on because emoji and regex is not minified properly using default
              // https://github.com/facebook/create-react-app/issues/2488
              ascii_only: true,
            },
          },
        }),
      ]
    },
    module: {
      rules: [
        {
          test: /\.(js|jsx)$/,
          use: "babel-loader",
          exclude: [/node_modules/, /walletLink/],
        },
        {
          test: /\.ts$/,
          loader: "ts-loader",
          exclude: [/node_modules/, /walletLink/],
        },
        {
          test: /\.tsx$/,
          use: [
            {
              loader: "babel-loader",
              options: {
                ...babelConfig
              },
            },
            {
              loader: 'linaria/loader',
              options: { 
                displayName: false,
                babelOptions: 
                { 
                presets: [
                  "@babel/preset-react",
                  '@babel/preset-typescript'
                ]
                }
              },
            }
          ],
          exclude: [/node_modules/, /walletLink/],
        },
        {
          test: /\.html/,
          use: [{ loader: 'html-loader' }],
        },
        {
          test: /\.css$/,
          use:  [
            {
              loader: MiniCssExtractPlugin.loader,
              options: {
               // Enables es module syntax for better bundle optimization
                esModule: true,
                publicPath: ''
              },
            },
            'css-loader',
          ],
        },
        {
          test: /\.(svg)$/,
          use: "file-loader",
        },
        {
          test: /\.png$/,
          use: [
            {
              loader: "file-loader",
            },
            {
              loader: "url-loader",
              options: {
                mimetype: "image/png",
              },
            },
          ],
        },
        {
          test: /\.(woff2)$/,
          use: [
            {
              loader: 'file-loader',
              options: {
                name: '[name].[ext]',
                outputPath: 'fonts/'
              }
            }
          ]
        },
      ],
    },
    resolve: {
      extensions: [".js", ".jsx", ".tsx", ".ts"],
      alias: {
        "stream": require.resolve("stream-browserify"),
        "crypto": require.resolve("crypto-browserify"),
        "process": "process/browser",
        "Buffer":require.resolve("buffer"),
        "@wallets": path.resolve(__dirname, "src/datalayer/wallets/"),
        "@http": path.resolve(__dirname, "src/datalayer/http/"),
        "@blockchains": path.resolve(__dirname, "src/datalayer/blockchains/"),
        "@ethereum": path.resolve(__dirname, "src/datalayer/ethereum/"),
        "@compoundfinance": path.resolve(__dirname, "src/datalayer/compoundfinance/"),
        "@store": path.resolve(__dirname, "src/datalayer/store/"),
        "@core": path.resolve(__dirname, "src/datalayer/core/"),
        "@user": path.resolve(__dirname, "src/datalayer/user/")
      },
      fallback: {
        "crypto": require.resolve("crypto-browserify"),
        "stream": require.resolve("stream-browserify"),
        "buffer": require.resolve("buffer")
      }
    },
    plugins: [
      new HtmlWebpackPlugin({
        templateContent: `
          <html>
            <head>
              <meta charset="UTF-8">
              <script src="ga.js"></script>
              <title>Coinbase Wallet</title>
            </head>
            <body>
                <div id="root"></div>
            </body>
          </html>
        `,
        chunks: ["popup"]
      }),
      new HtmlWebpackPlugin({
        filename: "home.html",
        templateContent: `
          <html>
            <head>
              <meta charset="UTF-8">
              <script src="ga.js"></script>
              <title>Wallet Extension</title>
            </head>
            <body>
                <div id="root"></div>
            </body>
          </html>
        `,
        chunks: ["home"]
      }),
      new CopyPlugin({
        patterns: [{ from: "public", to: "." }],
      }),
      new webpack.ProvidePlugin({
        process: 'process/browser',
      }),
      new MiniCssExtractPlugin({
        filename: '[name].css',
      }),
      new webpack.ProvidePlugin({
        process: 'process/browser',
        Buffer: ['buffer', 'Buffer']
      }),
      new DotenvWebpack({
        path: path.join(__dirname,'.env.production'),
        safe: true
      }),
    ],
  }

  return webpackConfig;
};
